#!/bin/bash

# This script is called from Java based Job Server
# Job server calls this script with some important information like (IPO)
# Earlier we relied on this IPO information provided over the command line to this script.
# But presence of characters like single quote in user program names caused trouble.
# We are now shifting to env variables based solution, where java based program has already set our required variables.
# Still this script receives ipo on commandline and we haven't phased it out yet.
# This program prepares necessary Fifo etc and calls user program

set -x
set -o pipefail

start=$(date +%s)





# Some helpful debugging info
echo "Current program full path: $(pwd)/$0"
echo "Host name is: $(hostname)"
echo "Current user is: $(whoami)"
echo "No of command line args: $#"
echo "Printing command line args: $*"



inargs=()
pargs=()
outargs=()


# Lets do some sanity checking about env variables required
function envSanityChecking {

    if [[ -z "${__plumb_no_of_inputs}" ]]
    then
        echo "__plumb_no_of_inputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_inputs}");
    do
        x="__plumb_i$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 1
        fi
        inargs+=("${!x}")

    done;



    if [[ -z "${__plumb_no_of_outputs}" ]]
    then
        echo "__plumb_no_of_outputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_outputs}");
    do
        x="__plumb_o$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 2
        fi
        outargs+=("${!x}")

    done;


    if [[ -z "${__plumb_prog_name}" ]]
    then
        echo "__plumb_prog_name not set."
        exit 3
    fi

    if [[ -z "${__plumb_prog_args}" ]]
    then
        echo "__plumb_prog_args not set."
    fi

}



function checkBinaryAvailability {


        programPath="${__plumb_prog_name}"
        programName=$(basename "$programPath")
        restOfCmd="${__plumb_prog_args}"
        newCmd="./$programName  $restOfCmd"

        /usr/local/bin/hdfs -copyToLocal "$programPath"  ./
        retCode=$?
        if [ $retCode == 0 ]
        then
                chmod +x ./"$programName"
                return $retCode
        fi


        echo "Copying binary from HDFS to local dir for file: $programPath failed with error code $retCode"

        if [ -e "$programPath" ]
        then
                ln -s -f "$programPath"  "$programName"
                return $?
        fi

        echo "$programPath not found on this server. Exiting."
        exit 4

}


function LinkRegExScript {
    if [[ -z "${pathToRegEx}" ]]  #  YARNApplicationMaster should have set it up
    then
        echo "Env variable: pathToRegEx not set."
        return 1
    fi


    ln -s "${pathToRegEx}"  ./regex.py

}


function LinkHLEScript {
    if [[ -z "${pathToHLEBASHAPI}" ]]  #  YARNApplicationMaster should have set it up
    then
        echo "Env variable: pathToHLEBASHAPI not set."
        return 1
    fi


    ln -s "${pathToHLEBASHAPI}"  ./call_lander_api_function_via_rpc.sh

}



eval envSanityChecking
eval checkBinaryAvailability
# above func sets newCmd
pargs+=("$newCmd")

eval LinkRegExScript
eval LinkHLEScript



# Some more debugging information
echo -n "Input: "
echo "${inargs[*]}"

echo -n "Program: "
echo "${pargs[*]}"

echo -n "Output: "
echo "${outargs[*]}"



inFifoList=()
outFifoList=()


function finish {
  echo " Cleaning up Fifos... "
  for inF in ${inFifoList[*]}
  do
    [ -e "$inF" ] && rm -f "$inF"
  done

  for outF in ${outFifoList[*]}
  do
    [ -e "$outF" ] && rm -f "$outF"
  done



  date_epochs=$(date +%s)
  duration=$((date_epochs-start))
  echo "duration = $duration"


}

# Ask shell to run finish function on program exit.
trap finish EXIT


#################################################################
function die {
    echo "$1"
    echo "Existing ..."
    exit 1

}

#################################################################
# Some exports of env variables
PLUMB_INPUTS=$(IFS=,; echo "${inargs[*]}")
export PLUMB_INPUTS
# following for case-insensitive search
shopt -s nocasematch
if [[ " ${outargs[*]} " =~ " None " ]]; then
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=true

else
    PLUMB_OUTPUTS=$(IFS=,; echo "${outargs[*]}")
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=false
fi
shopt -u nocasematch

printenv



# User program command line
uCmd="${pargs[*]}"




echo "User command line is: $uCmd"

# Run user program
eval "$uCmd"
retCode=$?


# Terminate any stuck child-processes
# That can happen if $uCmd was faulty and child-processing are stuck waiting

if [ $retCode != 0 ]
then
	pkill -P $$
fi

for child in "${children[@]}"
do
    echo "Waiting on child with id: $child"
    wait "$child"

done

exit $retCode
