#!/bin/bash


set -x
set -o pipefail

start=$(date +%s)





# Some helpful debugging info
echo "Current program full path: $(pwd)/$0"
echo "Host name is: $(hostname)"
echo "Current user is: $(whoami)"
echo "No of command line args: $#"
echo "Printing command line args: $*"


ls -lah

inargs=()
pargs=()
outargs=()


# Lets do some sanity checking about env variables required
function envSanityChecking {

    if [[ -z "${__plumb_no_of_inputs}" ]]
    then
        echo "__plumb_no_of_inputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_inputs}");
    do
        x="__plumb_i$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 1
        fi
        inargs+=("${!x}")

    done;



    if [[ -z "${__plumb_no_of_outputs}" ]]
    then
        echo "__plumb_no_of_outputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_outputs}");
    do
        x="__plumb_o$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 2
        fi
        outargs+=("${!x}")

    done;


    if [[ -z "${__plumb_prog_name}" ]]
    then
        echo "__plumb_prog_name not set."
        exit 3
    fi

    if [[ -z "${__plumb_prog_args}" ]]
    then
        echo "__plumb_prog_args not set."
    fi

}



function checkBinaryAvailability {


        programPath="${__plumb_prog_name}"
        programName=$(basename "$programPath")
        restOfCmd="${__plumb_prog_args}"
        newCmd="./$programName  $restOfCmd"

        /usr/local/bin/hdfs -copyToLocal "$programPath"  ./
        retCode=$?
        if [ $retCode == 0 ]
        then
                chmod +x ./"$programName"
                return $retCode
        fi


        echo "Copying binary from HDFS to local dir for file: $programPath failed with error code $retCode"

        if [ -e "$programPath" ]
        then
                ln -s -f "$programPath"  "$programName"
                return $?
        fi

        echo "$programPath not found on this server. Exiting."
        exit 4

}

# Following should be called as: SymLinkScripts "name_of_env_variable" "targe_path"
function SymLinkScripts {
    if [[ $# != 2 ]]
    then
      echo "SymLinkScripts function needs two parameters"
      return 2
    fi
    env_variable_name="$1"
    target_path="$2"

    if [[ -z "${!env_variable_name}" ]]  #  YARNApplicationMaster should have set it up
    then
        echo "Env variable: $env_variable_name not set."
        return 1
    fi
    ln -s "${!env_variable_name}"  "$target_path"
}

# Following called with one arg of name
function checkYamlFiles {
    if [[ $# != 1 ]]
    then
      echo "checkYamlFiles function needs one parameter"
      return 3
    fi
    if [[ -z "${!$1}" ]]  #  YARNApplicationMaster should have set it up
    then
        echo "Env variable: $1 not set."
        return 4
    fi
}


eval envSanityChecking
eval checkBinaryAvailability
# above func sets newCmd
pargs+=("$newCmd")


# Now link to helping python scripts for the end user
retCode=0
eval SymLinkScripts "pathToRegEx" "./regex.py"
retCode=$((retCode+$?))
eval SymLinkScripts "pathToYamlUtil" "./yamlutil.py"
retCode=$((retCode+$?))
if [[ $retCode != 0 ]]
then
  echo "Linking to regex or yamlutils failed."
  exit 100
fi

# Check yaml env variables present
retCode=0
eval checkYamlFiles "inyaml"
retCode=$((retCode+$?))
eval checkYamlFiles "outyaml"
retCode=$((retCode+$?))
eval checkYamlFiles "erryaml"
retCode=$((retCode+$?))
if [[ $retCode != 0 ]]
then
  echo "in, out or err yaml evn variables not found in yarn_window_job_wrapper.sh."
  exit 101
fi



# Some more debugging information
echo -n "Input: "
echo "${inargs[*]}"

echo -n "Program: "
echo "${pargs[*]}"

echo -n "Output: "
echo "${outargs[*]}"


inFifoList=()
outFifoList=()


function finish {
  echo " Cleaning up Fifos... "
  for inF in ${inFifoList[*]}
  do
    [ -e "$inF" ] && rm -f "$inF"
  done

  for outF in ${outFifoList[*]}
  do
    [ -e "$outF" ] && rm -f "$outF"
  done



  date_epochs=$(date +%s)
  duration=$((date_epochs-start))
  echo "duration = $duration"


}

# Ask shell to run finish function on program exit.
trap finish EXIT

#################################################################
# Some exports of env variables
PLUMB_INPUTS=$(IFS=,; echo "${inargs[*]}")
export PLUMB_INPUTS
# following for case-insensitive search
shopt -s nocasematch
if [[ " ${outargs[*]} " =~ " None " ]]; then
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=true

else
    PLUMB_OUTPUTS=$(IFS=,; echo "${outargs[*]}")
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=false
fi
shopt -u nocasematch

printenv



# User program command line
uCmd="${pargs[*]}"





userProg="$cell_userProg"
echo "$userProg"


eval "${pargs[*]}" < ./inyaml > ./outyaml



