package edu.isi.ant.JobAPI;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.logging.Logger;

public class HDFSHelper

  {
      public static String hdfs = "/usr/local/bin/hdfs";

    public static int makeDir(String dirPath, Logger logger)
    {
      logger.info("Making HDFS directory " + dirPath);
      int result = 0;

      try {


        ProcessBuilder pb = new ProcessBuilder(hdfs, "-mkdir", "-p", dirPath)
                                            .inheritIO();
        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
      }

      return result;
    }



      public static int makeDir(String dirPath, String owner, Logger logger)
      {
          int result = 0;
          result = makeDir(dirPath, logger);
          if(result != 0) return result;


          logger.info("Setting owner on HDFS dir " + dirPath);


          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-chown", owner, dirPath)
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }



    public static int copyFromLocal(String localSrc, String hdfsDst, Logger logger)
    {

      logger.info("Copying " + localSrc + " to " + hdfsDst);
      int result = 0;

      try {


        ProcessBuilder pb = new ProcessBuilder(hdfs, "-copyFromLocal", "-f", localSrc, hdfsDst)
                                            .inheritIO();
        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
      }

      return result;
    }



      public static int copyToLocal(String hdfsSrc, String localDst, Logger logger)
      {

          logger.info("Copying " + hdfsSrc + " to " + localDst);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-copyToLocal", hdfsSrc, localDst)
                                                    .inheritIO(); // necessery to do, else can hang for ever
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }



      public static int copyToLocalWithLocalReplacement(String hdfsSrc, String localDst, Logger logger)
      {

          logger.info("Copying " + hdfsSrc + " to " + localDst);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-copyToLocal -f ", hdfsSrc, localDst)
                      .inheritIO(); // necessery to do, else can hang for ever
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }



      public static int mvFromLocal(String localSrc, String hdfsDst, Logger logger)
      {

          logger.info("Moving all files from " + localSrc + " to " + hdfsDst);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-moveFromLocal", "-f", localSrc+"/*", hdfsDst)
                                            .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }


      public static int mvOneFileFromLocal(String localSrc, String hdfsDst, Logger logger)
      {

          logger.info("Moving  " + localSrc + " to " + hdfsDst);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-moveFromLocal", "-f", localSrc, hdfsDst)
                                              .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }



      public static ArrayList<String> listFiles(String hdfsPath, Logger logger)
      {

          logger.info("Listing files for HDFS path:" + hdfsPath);
          ArrayList<String> files = new ArrayList<>();


          try {




              ProcessBuilder pb = new ProcessBuilder(ServerConfig.getString("hdfsFileListerScript"),
                                                        hdfsPath
              );


              Process p = pb.start();

              BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));

              String line = null;
              while ( (line = reader.readLine()) != null) {
                  files.add(line);
                  //System.out.println(line);
              }

              reader.close();




              p.waitFor();
              p.exitValue();

              p.destroy();
              pb = null;



          }catch(Exception e)
          {

              e.printStackTrace();
          }

          return files;


      }


      public static int mvOneFile(String hdfsSrc, String hdfsDst, Logger logger)
      {
          //Src = hdfs file name
          //Dst = hdfs dir
          String fName = MyUtils.getBaseName(hdfsSrc);
          String dFName = String.format("%s/%s", hdfsDst, fName);

          // Handle a special case before moving forward
          String s = hdfsSrc.replaceAll("/+", "/");
          String d = dFName.replaceAll("/+", "/");
          if(s.equals(d)) return 0;

          // Since hdfs mv command does not support -f switch, I am doing the following
          if (0 == HDFSHelper.doesXExists(dFName, false))
          {
              HDFSHelper.delOneFile(dFName,logger);
          }

          logger.info("Moving  " + hdfsSrc + " to " + hdfsDst);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-mv", hdfsSrc, hdfsDst)
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }

      public static int delOneFile(String hdfsSrc, Logger logger)
      {


          logger.info("Deleting old  " + hdfsSrc);
          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-rm", hdfsSrc)
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }



      public static int doesXExists(String dirPath, boolean isDir)
      {

          int result = 0;

          try {

              String testWhat = (isDir?"-d":"-f");


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-test", testWhat, dirPath)
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }


      public static int chmod(String path, String rights)
      {

          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-chmod", rights, path)
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }


      public static int empytDir(String dirPath)
      {

          int result = 0;

          try {


              ProcessBuilder pb = new ProcessBuilder(hdfs, "-rm ", dirPath+"/*")
                      .inheritIO();
              Process p = pb.start();

              p.waitFor();
              result = p.exitValue();
              p.destroy();
              pb = null;

          }catch(Exception e)
          {
              result = 1;
          }

          return result;
      }


  }//end of class

