package edu.isi.ant.JobAPI;

import edu.isi.ant.yamlseranddeser.Stage;
import edu.isi.ant.yamlseranddeser.Window;
import io.grpc.lander.queue.api.ReplyWindow;
import io.grpc.lander.queue.api.WindowReply;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.util.ClassUtil;
import org.apache.hadoop.yarn.api.records.LocalResource;
import org.apache.hadoop.yarn.conf.YarnConfiguration;


import javax.net.ssl.SSLException;
import java.io.*;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;


public class LumpinessAwareFastDispatcher implements Runnable {

    private Map<String, Integer> dupTable;
    private Map<String, Stage> dispTable;
    private Map<String, Window> allWindows;
    //private static final String dupDispStateSavePathInHDFS = "/user/aqadeer/jobserver/dup_disp_table/";
    //public static final String dispPath = String.format("%s%s", dupDispStateSavePathInHDFS, "dispTable");
    //public static final String dupPath =  String.format("%s%s", dupDispStateSavePathInHDFS, "dupTable");

    private Logger logger;
    private ArrayList<StageControllerCell> stageControllerCells = new ArrayList<>();
    private ArrayList<StageControllerCell> stageControllerCellsForWindows = new ArrayList<>();
    private Map<String, StageControllerCell> stageControllerCellsForStateful = new HashMap<>();


    //private ArrayList<YARNJobSubmitter> Yarnjobs = new ArrayList();
    private ArrayList<JobState> yarnJobs = new ArrayList<>();

    WindowClient windowClient;

    {
        try {
            windowClient = new WindowClient(ServerConfig.getString("QueueServerHostName"),
                        ServerConfig.getInt("QueueServerListeningPort"));
        } catch (SSLException e) {
            e.printStackTrace();
        }
    }


    private int maxAllowedWorkersGlobally = MyUtils.getTotalClusterCores();
    private int currentlyActiveWorkers = 0;
    // Following two are here to facilitate experimentation. They don't affect normal run.
    private final int workToFinish = 1393 * 3;//896*3; //300; //300;
    private boolean experimentmode = false;
    private transient boolean needsStopping = false;


    public LumpinessAwareFastDispatcher(Map<String, Integer> dupTable,
                                                  Map<String, Stage> dispTable,
                                                  Map<String, Window> allWindows,
                                                  Logger logger) {

        this.copyResourcesForApplicationMaster(); // one time only


        //...................................................
        this.dupTable = dupTable;
        this.dispTable = dispTable;
        this.allWindows = allWindows;
        this.logger = logger;
        logger.info("LAD: main master starting...");
        logger.info("LAD: setting no of workers:" + maxAllowedWorkersGlobally);



        MyUtils.seralizeObjectInHDFS(this.dispTable, ServerConfig.getString("dispPath"), ServerConfig.getString("dispPathInHDFS"), logger);
        MyUtils.seralizeObjectInHDFS(this.dupTable, ServerConfig.getString("dupPath"), ServerConfig.getString("dupPathInHDFS"), logger);


        //MyUtils.seralizeObjectInLocal(this.dispTable, Constants.dispPath, logger);
        //MyUtils.seralizeObjectInLocal(this.dupTable, Constants.dupPath, logger);




        for (Map.Entry<String, Stage> entry : this.dispTable.entrySet()) {
            StageControllerCell cell = new StageControllerCell();
            cell.stageName = entry.getKey();
            cell.stage = entry.getValue();
            switch (cell.stage.getAbstraction()) {
                case BLOCK_STREAMING:
                    this.stageControllerCells.add(cell);
                    break;

                case WINDOWED_STREAMING:
                    this.stageControllerCellsForWindows.add(cell);
                    break;

                case STATEFUL_STREAMING:
                    this.stageControllerCellsForStateful.put(cell.stageName, cell);
                    break;

                default:
                    logger.warning("Stage abstraction name (" + cell.stage.getAbstraction().name() + ") unknown.");
                    break;

            }
        }




/*
        // old. To remove
        for (Map.Entry<String, Stage> entry : this.dispTable.entrySet()) {
            StageControllerCell cell = new StageControllerCell();
            cell.stageName = entry.getKey();
            cell.stage = entry.getValue();
            //If this stage has a window, exclude it.  We will handle them separately.
            if(!cell.stage.isThisStageHasWindow())
            {
                this.stageControllerCells.add(cell);
            }
            else
            {
                this.stageControllerCellsForWindows.add(cell);
            }

        }

*/
        //ToDo: Following failure thread should be mutually exclusive with the code that finds next windows
        //      to schedule.
        //ToDo: Following failure handling thread needs to change for windowed case

        // start a thread to take care of failures
        Thread t = new Thread() {
            public void run() {
                while (true) {
                    try {
                        for (String q : dupTable.keySet()) {
                            int val = dupTable.get(q);
                            if(val > 1)
                            {
                                for (int i = 1; i <= val; i++) {
                                    String derivedQ = String.format("%s_%s", q, i);
                                    logger.info("ErrorRecover on stage with name: " +  derivedQ);
                                    QueueHelper.doErrorRecovery(derivedQ, allWindows.containsKey(derivedQ), logger);

                                }

                            }else
                            {
                                logger.info("ErrorRecover on stage with name: " +  q);
                                QueueHelper.doErrorRecovery(q, allWindows.containsKey(q), logger);

                            }


                        }

                        TimeUnit.MINUTES.sleep(ServerConfig.getInt("errorRecoveryInterval"));
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }

                }
            }
        };
        t.start();



    }





    // ToDo: Can an exception (like some issues in gRPC call) kill the following dispatch thread?  If yes that will
    // stop all serving of files.  Need to verify and needs a solution.
    @Override
    public void run() {
        long millis = System.currentTimeMillis();
        this.logger.info("Starting time in unix epochs: " + millis / 1000);


        long t1 = System.nanoTime();
        long roundNo = 1;


        while (this.isDone() != true) {



            this.logger.info("Starting round no: " + roundNo);
            this.logger.info("Finding max allowed workers...");
            maxAllowedWorkersGlobally = MyUtils.getTotalClusterCores();
            this.logger.info("Max allowed workers are: " + maxAllowedWorkersGlobally);


            // Get stages's backlog and get allowed YARN workers quota per stage
            this.logger.info("Finding worker distribution...");
            this.findWorkerDistributionPerStageBacklog();




            // Assign YARN workers to stages
            this.logger.info("Assigning workers to pending work...");
            for (StageControllerCell cell : this.stageControllerCells) {
                this.assignYARNWorkers(cell);
            }

            // Assign YARN Workers to windowed stages
            this.logger.info("Finding any windows ready to run and assigning workers to them...");
            for (StageControllerCell cell : this.stageControllerCellsForWindows) {
                this.assignYARNWorkersToWindows(cell);
            }

            // If one worker is not running for stateful-streaming, re-run
            this.logger.info("Finding any stateful_streaming stages and making sure one instance is running...");
            this.logger.info("No of stateful_streaming stages are: " + this.stageControllerCellsForStateful.keySet().size());
            for(String stageName : this.stageControllerCellsForStateful.keySet())
            {
                this.assignYARNWorkersToStatefulStreaming(this.stageControllerCellsForStateful.get(stageName));
            }





            // Sleep some time before checking statuses and starting next round
            this.logger.info("Sleep...");
            try {
                TimeUnit.MINUTES.sleep(3);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }



            this.logger.info("Wake up and check woker status...");
            this.checkAlreadySpawnedWorkerStatusNonBlocking();



            roundNo++;
        }

        // wait on any un-finished jobs
        this.waitForUnFinishedJobs();

        long t2 = System.nanoTime();
        long elapsedTime = (t2-t1);

        this.logger.info("Dispather thread exiting...");

        logger.info("****@@@@****** =>>>>> elapsedTime = " + elapsedTime + " nano seconds" +
                " and Hours: " + ((float)elapsedTime/((float)1000000000 * 3600)));

        //signal that App is done
        JobAPIService.makeFile("./p_done/done");

        System.exit(0);

    }

    private void waitForUnFinishedJobs()
    {

        long t1 = System.nanoTime();
        long t2 = System.nanoTime();
        long minutesElapsed = (t2-t1)/(1000000000 * 60);


        while(this.currentlyActiveWorkers != 0 && minutesElapsed < ServerConfig.getInt("maxYARNJobRunTime") )
        {
            // Sleep some time before checking statuses and starting next round
            this.logger.info("Sleep...");
            try {
                TimeUnit.MINUTES.sleep(3);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }



            this.logger.info("Wake up and check woker status...");
            this.checkAlreadySpawnedWorkerStatusNonBlocking();

            t2 = System.nanoTime();
            minutesElapsed = (t2-t1)/(1000000000 * 60);

        }

    }


    synchronized public void writeStopSignal()
    {
        this.needsStopping = true;
    }

    synchronized private boolean readStopSignal()
    {
        return this.needsStopping;
    }





    private void copyResourcesForApplicationMaster() {
        // ToDo: Use MyUtil's function and remove code redundance
        // Do one time jar copy for later Yarn submission here
        // setup jar for AM
        String jar = ClassUtil.findContainingJar(YARNJobSubmitter.class);
        YarnConfiguration yarnConf = YARNJobSubmitter.getConf();
        FileSystem fs = null;
        try {
            fs = FileSystem.get(yarnConf);
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Failed to get YARN configuration...");
            System.exit(1);
        }
        ///Path src = new Path(jar);
        ///Path dest = new Path(fs.getHomeDirectory(), src.getName());
        ///System.out.format("Copying JAR from %s to %s%n", src, dest);
        try {
            LocalResource shellScript = MyUtils.makeLocalResource(jar,
                    yarnConf, true);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Failed to copy job jar to HDFS.  Will exit");
            System.exit(2);
        }

        // copy shell script for later Yarn submission as well
        ///src = new Path("./scripts/yarn_job_wrapper.sh");
        ///dest = new Path(fs.getHomeDirectory(), src.getName());
        ///System.out.format("Copying yarn script wrapper from %s to %s%n", src, dest);
        try {
            LocalResource shellScript = MyUtils.makeLocalResource("./scripts/yarn_job_wrapper.sh",
                    yarnConf, true);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Failed to copy shell script to HDFS.  Will exit");
            System.exit(3);
        }


        // copy jar containing Stage
        jar = ClassUtil.findContainingJar(Stage.class);

        ///src = new Path(jar);
        ///dest = new Path(fs.getHomeDirectory(), src.getName());
        ///System.out.format("Copying JAR from %s to %s%n", src, dest);

        try {
            LocalResource shellScript = MyUtils.makeLocalResource(jar,
                    yarnConf, true);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Failed to copy job jar to HDFS.  Will exit");
            System.exit(2);
        }

        // copy YARN wrapper for windowed jobs
        try {
            LocalResource shellScript = MyUtils.makeLocalResource("./scripts/yarn_window_job_wrapper.sh",
                    yarnConf, true);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Failed to copy /scripts/yarn_window_job_wrapper.sh to HDFS.  Will exit");
            System.exit(3);
        }

        // copy stateful-streaming YARN wrapper
        try {
            LocalResource shellScript = MyUtils.makeLocalResource("./scripts/yarn_job_wrapper_for_stateful_streaming.sh",
                    yarnConf, true);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Failed to copy yarn_job_wrapper_for_stateful_streaming.sh to HDFS.  Will exit");
            System.exit(4);
        }

        // copy config file (aqadeer copied following snippet from plain dispatcher that was changed by yuri probably)
        Path src = new Path(ServerConfig.CONFIGURATION_FILE);
        Path dest = new Path(ServerConfig.getString("HDFSScratchSpace"), src.getName());
        System.out.format("Copying configuration file from %s to %s%n", src, dest);
        try {
            fs.copyFromLocalFile(src, dest);
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Failed to copy server config to to HDFS.  Will exit");
            System.exit(1);
        }



    }


    // Fetch lengths of queues attached to stages
    // Return value is total backlog across all stages.
    private int findWorkerDistributionPerStageBacklog() {

        //int qLengths[] = new int[this.dispTable.size()];
        int totalQLengths = 0;


        for (StageControllerCell cell : this.stageControllerCells) {
            // rate-limit a stage if user asked so
            if (cell.stage.getMaxConcurrentJobs() != 0)
            {
                cell.backlog = Math.min(this.getInputQueueSize(cell), cell.stage.getMaxConcurrentJobs());
            }
            else
            {
                cell.backlog = this.getInputQueueSize(cell);
            }


            totalQLengths += cell.backlog;
        }


        float multiplier = (totalQLengths != 0) ? ((float) this.maxAllowedWorkersGlobally / (float) totalQLengths) : 0;

        for (StageControllerCell cell : this.stageControllerCells) {
            cell.maxAllowedWorkers = (int) (Math.ceil(cell.backlog * multiplier));


        }


        return totalQLengths;

    }


    private int getInputQueueSize(StageControllerCell cell) {
        // Getting min of all input queues to a stage taken here

        List<String> inQueues = cell.stage.getInput();
        int min = Integer.MAX_VALUE;

        for (String qName : inQueues) {
            QueueHelper.QResult result = QueueHelper.getQueueSizeRepeatOnError(qName, logger);
            min = (result.qSize < min) ? result.qSize : min;

            this.logger.info("LAD: " + qName + " has length: " + result.qSize);

        }

        return min;
    }


    private boolean isDone() {
        if(this.experimentmode) return isDoneForExperiments();



        return this.readStopSignal();

    }

    private boolean isDoneForExperiments() {
        int total = 0;

        int a = QueueHelper.getQueueSizeRepeatOnError("rsscint", logger).qSize;
        int b = QueueHelper.getQueueSizeRepeatOnError("message_question.fsdb.xz", logger).qSize;
        int c = QueueHelper.getQueueSizeRepeatOnError("pcap.xz", logger).qSize;

        total = a + b + c;

        return ((total == this.workToFinish) ? true : false);

    }








    private void spawnWorker(StageControllerCell cell) {

        List<String> args = new ArrayList<String>();
        args.add(cell.stageName);
        ExecutionReport execResult = this.executeNonBlocking(args);

        //ToDo: What to do if above call to execute fails?  Retry?

        if (execResult.isSuccess)
        {
            this.currentlyActiveWorkers++;
            this.yarnJobs.add(new JobState(cell, execResult));

        }
    }


    private void spawnWindowWorker(StageControllerCell cell, ReplyWindow replyWindow) {

        //List<String> args = new ArrayList<String>();
        //args.add(cell.stageName);
        ExecutionReport execResult = this.executeWindowNonBlocking(cell, replyWindow);

        //ToDo: What to do if above call to execute fails?  Retry?

        if (execResult.isSuccess)
        {
            this.currentlyActiveWorkers++;
            this.yarnJobs.add(new JobState(cell, execResult));

        }
    }


    private void spawnStatefulStreamingWorker(StageControllerCell cell) {
        System.out.println("####################### step 3");
        List<String> args = new ArrayList<String>();
        args.add(cell.stageName);
        ExecutionReport execResult = this.executeStatefulStreamingNonBlocking(args);

        //ToDo: What to do if above call to execute fails?  Retry?

        if (execResult.isSuccess)
        {
            JobState jState = new JobState(cell, execResult);
            this.currentlyActiveWorkers++;
            this.yarnJobs.add(jState);
            cell.statefulStreamingWorker = jState;

        }
    }





    private void checkAlreadySpawnedWorkerStatus() {
        int freeWorkerFound = 0;

        while (freeWorkerFound == 0) {
            // Run through all running workers to check status
            // Sleep 30 seconds if nothing found
            List<JobState> finished = new ArrayList();
            for (JobState jobState:this.yarnJobs) {
                if (jobState.execReport.submitter.isJobFinished()) {
                    finished.add(jobState);
                    freeWorkerFound++;

                }
            }
            this.yarnJobs.removeAll(finished);



            if (freeWorkerFound == 0) {
                try {
                    this.logger.info("Sleep inside checkAlreadySpawnedWorkerStatus");
                    TimeUnit.SECONDS.sleep(10);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        } // end of while

        this.currentlyActiveWorkers -= freeWorkerFound;

    }

    private void checkAlreadySpawnedWorkerStatusNonBlocking() {
        int freeWorkerFound = 0;


        List<JobState> finished = new ArrayList();
        for (JobState jobState : this.yarnJobs) {
            if (jobState.execReport.submitter.isJobFinished()) {
                finished.add(jobState);
                freeWorkerFound++;

            }
        }
        this.yarnJobs.removeAll(finished);


        this.currentlyActiveWorkers -= freeWorkerFound;

    }

    private void assignYARNWorkers(StageControllerCell cell) {
        int howManyTimes = Math.min(cell.backlog, cell.maxAllowedWorkers);
        //ToDo: Following for testing only.
        System.out.println("For Stage:" + cell.stageName + " workers used are:" + howManyTimes);
        while (howManyTimes != 0) {
            if (this.currentlyActiveWorkers < this.maxAllowedWorkersGlobally) {
                this.spawnWorker(cell);
                cell.backlog--;
                howManyTimes--;
            } else {
                this.checkAlreadySpawnedWorkerStatus();
            }
        }

    }

    private void assignYARNWorkersToWindows(StageControllerCell cell) {
        if(this.windowClient == null) return;
        // fetch windows ready for scheduling for this stage.
        // ToDo: For now just one input.  For multi input stages, needs fixing
        String windowedInput = cell.stage.getInput().get(0);
        WindowReply reply = this.windowClient.getWindowsToSchedule(windowedInput);
        if(reply == null || reply.getResult() == false || reply.getReadyWindowCount() < 1) return;

        int howManyTimes = reply.getReadyWindowCount();
        //ToDo: Following for testing only.
        System.out.println("For Stage:" + cell.stageName + " workers used are:" + howManyTimes);
        if (true) return;
        int i = 0;
        while (howManyTimes != 0) {
            if (this.currentlyActiveWorkers < this.
                    maxAllowedWorkersGlobally) {
                this.spawnWindowWorker(cell, reply.getReplyWindows(i));
                i++;
                howManyTimes--;
            } else {
                this.checkAlreadySpawnedWorkerStatus();
            }
        }

    }

    private void assignYARNWorkersToStatefulStreaming(StageControllerCell cell) {

        System.out.println("####################### step 1");
        if (cell.statefulStreamingWorker == null || cell.statefulStreamingWorker.execReport.submitter.isJobFinished())
        {
            System.out.println("####################### step 2");
            int howManyTimes = 1;
            while (howManyTimes != 0) {
                if (this.currentlyActiveWorkers < this.maxAllowedWorkersGlobally) {
                    this.spawnStatefulStreamingWorker(cell);
                    howManyTimes--;
                } else {
                    this.checkAlreadySpawnedWorkerStatus();
                }
            }

        }


    }




    private ExecutionReport executeNonBlocking(List<String> cmdLine) {
        logger.info("Making a new YARNJobSubmitter...");
        YARNJobSubmitter submitter = new YARNJobSubmitter();
        logger.info("submitting job....");
        boolean isSuccess = submitter.submitJobNonBlocking(cmdLine);
        return new ExecutionReport(isSuccess, submitter);
    }


    private ExecutionReport executeWindowNonBlocking(StageControllerCell cell, ReplyWindow replyWindow) {
        logger.info("Making a new YARNJobSubmitterForWindows...");
        YARNJobSubmitterForWindows submitter = new YARNJobSubmitterForWindows();
        logger.info("submitting window job....");
        boolean isSuccess = submitter.submitJobNonBlocking(cell, replyWindow);
        return new ExecutionReport(isSuccess, submitter);
    }

    private ExecutionReport executeStatefulStreamingNonBlocking(List<String> cmdLine) {
        logger.info("Making a new Stateful-Streaming Worker...");
        YARNJobSubmitterForStatefulStreaming submitter = new YARNJobSubmitterForStatefulStreaming();
        logger.info("submitting stateful-streaming job....");
        //ToDo: Following to fix
        boolean isSuccess = submitter.submitJobNonBlocking(cmdLine);
        return new ExecutionReport(isSuccess, submitter);
    }








    static class StageControllerCell {
        public String stageName;
        public Stage stage;
        public int backlog;
        public int maxAllowedWorkers;
        public JobState statefulStreamingWorker;

    }


    static class TezQRep {

        public TezQRep(String qName, String hdfsURI) {
            this.qName = qName;
            this.hdfsURI = hdfsURI;
        }

        public String qName;
        public String hdfsURI;
    }


    static class JobState {


        public ExecutionReport execReport;
        public StageControllerCell cell;

        public JobState(StageControllerCell cell, ExecutionReport execReport )
        {
            this.cell = cell;
            this.execReport = execReport;


        }
    }


    static class ExecutionReport
    {
        public boolean isSuccess;
        public JobSubmitter submitter;


        public ExecutionReport(boolean isSuccess, JobSubmitter submitter)
        {
            this.isSuccess = isSuccess;
            this.submitter =submitter;
        }
    }









}


