package edu.isi.ant.JobAPI;

import edu.isi.ant.yamlseranddeser.InputQueueParsingFailedException;
import edu.isi.ant.yamlseranddeser.WindowParsingFailedException;
import edu.isi.ant.yamlseranddeser.YAMLSerAndDeSer;
import org.apache.commons.io.FileUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.*;
import org.apache.hadoop.hdfs.DFSClient;
import org.apache.hadoop.mapreduce.ClusterMetrics;
import org.apache.hadoop.yarn.api.records.LocalResource;
import org.apache.hadoop.yarn.api.records.LocalResourceType;
import org.apache.hadoop.yarn.api.records.LocalResourceVisibility;
import org.apache.hadoop.yarn.api.records.YarnClusterMetrics;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.util.ConverterUtils;
import org.apache.hadoop.yarn.util.Records;
import org.apache.htrace.fasterxml.jackson.databind.ObjectMapper;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MyUtils {

    public static boolean seralizeObject(Object obj, String fullyQualifiedPath) {
        boolean result = true;
        //Saving of object in a file
        FileOutputStream file = null;
        try {
            file = new FileOutputStream(fullyQualifiedPath);

            ObjectOutputStream out = new ObjectOutputStream(file);

            // Method for serialization of object
            out.writeObject(obj);

            out.close();
            file.close();
        } catch (IOException e) {
            e.printStackTrace();
            result = false;
        }

        System.out.println("Object has been serialized");
        return result;
    }


    public static Object deSeralizeObject(String fullyQualifiedPath) {
        Object obj = null;
        // Deserialization
        try {
            // Reading the object from a file
            FileInputStream file = new FileInputStream(fullyQualifiedPath);
            ObjectInputStream in = new ObjectInputStream(file);

            // Method for deserialization of object
            obj = in.readObject();

            in.close();
            file.close();


        } catch (IOException | ClassNotFoundException ex) {
            ex.printStackTrace();
        }

        return obj;
    }


    public static boolean seralizeObjectInHDFS(Object obj, String localFileName, String hdfsFileName, Logger logger) {
        boolean result = true;

        //Saving of object in a file
        FileOutputStream file = null;
        try {
            file = new FileOutputStream(localFileName);

            ObjectOutputStream out = new ObjectOutputStream(file);

            // Method for serialization of object
            out.writeObject(obj);

            out.close();
            file.close();
        } catch (IOException e) {
            e.printStackTrace();
            result = false;
        }

        if (result == false) {
            System.out.println("Serialization to local failed.");
            return result;
        }

        System.out.println("Object has been serialized locally...");


        YarnConfiguration conf = YARNJobSubmitter.getConf();
        try {
            FileSystem fs = FileSystem.get(conf);
            Path src = new Path(localFileName);
            //did't use current user's hdfs home directory because now e.g.
            // it might be aqadeer but on fetch time, it might be hlander.
            Path dst = new Path(hdfsFileName);
            fs.moveFromLocalFile(src, dst);
        } catch (IOException e) {
            result = false;
            System.out.println("Failed to copy from local to HDFS.");
            e.printStackTrace();
        }


        return result;
    }


    public static Object deSeralizeObjectFromHDFS(String localFileName, String hdfsFileName, Logger logger) {
        Object obj = null;

        YarnConfiguration conf = YARNJobSubmitter.getConf();
        try {
            FileSystem fs = FileSystem.get(conf);
            Path dst = new Path(localFileName);
            Path src = new Path(hdfsFileName);
            fs.copyToLocalFile(src, dst);
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Failed to copy from HDFS to local.");
            return null;
        }


        // Deserialization
        try {
            // Reading the object from a file
            File file = new File(localFileName);
            FileInputStream fis = new FileInputStream(localFileName);
            ObjectInputStream in = new ObjectInputStream(fis);

            // Method for deserialization of object
            obj = in.readObject();

            in.close();
            file.delete();


        } catch (IOException | ClassNotFoundException ex) {
            ex.printStackTrace();
        }


        return obj;
    }

    public static boolean seralizeObjectInHDFS_v1(Object obj, String fullyQualifiedPath, Logger logger
    ) {
        boolean result = true;
        String localFileName = "./__x__";
        //Saving of object in a file
        FileOutputStream file = null;
        try {
            file = new FileOutputStream(localFileName);

            ObjectOutputStream out = new ObjectOutputStream(file);

            // Method for serialization of object
            out.writeObject(obj);

            out.close();
            file.close();
        } catch (IOException e) {
            e.printStackTrace();
            result = false;
        }

        if (result == false) {
            System.out.println("Serialization to local failed.");
            return result;
        }

        System.out.println("Object has been serialized locally...");

        int retCode = HDFSHelper.mvOneFileFromLocal(localFileName, fullyQualifiedPath, logger);


        return (retCode == 0) ? true : false;
    }


    public static Object deSeralizeObjectFromHDFS_v1(String fullyQualifiedPath, Logger logger) {
        Object obj = null;
        String localFileName = "./__x__";
        int retCode = HDFSHelper.copyToLocal(fullyQualifiedPath, localFileName, logger);
        if (retCode != 0) {
            System.out.println("Failed to copy from HDFS to local.");
            return null;
        }
        // Deserialization
        try {
            // Reading the object from a file
            File file = new File(localFileName);
            FileInputStream fis = new FileInputStream(localFileName);
            ObjectInputStream in = new ObjectInputStream(fis);

            // Method for deserialization of object
            obj = in.readObject();

            in.close();
            file.delete();


        } catch (IOException | ClassNotFoundException ex) {
            ex.printStackTrace();
        }


        return obj;
    }


    public static boolean seralizeObjectInLocal(Object obj, String localFileName, Logger logger
    ) {
        boolean result = true;

        //Saving of object in a file
        FileOutputStream file = null;
        try {
            file = new FileOutputStream(localFileName);

            ObjectOutputStream out = new ObjectOutputStream(file);

            // Method for serialization of object
            out.writeObject(obj);

            out.close();
            file.close();
        } catch (IOException e) {
            e.printStackTrace();
            result = false;
        }

        if (result == false) {
            System.out.println("Serialization to local failed.");
            return result;
        }

        System.out.println("Object has been serialized locally...");

        return result;

    }


    public static Object deSeralizeObjectFromLocal(String localFileName, Logger logger) {
        Object obj = null;

        // Deserialization
        try {
            // Reading the object from a file
            File file = new File(localFileName);
            FileInputStream fis = new FileInputStream(localFileName);
            ObjectInputStream in = new ObjectInputStream(fis);

            // Method for deserialization of object
            obj = in.readObject();

            in.close();
            file.delete();


        } catch (IOException | ClassNotFoundException ex) {
            ex.printStackTrace();
        }


        return obj;
    }

    public static boolean fetchQueuingSystemCode(Logger logger) {
        boolean result = false;
        int retCode = HDFSHelper.copyToLocalWithLocalReplacement(ServerConfig.getString("queuingSystem"),
                "./", logger);

        result = retCode == 0 ? true : false;

        if (retCode != 0) {

            logger.info("Failed to fetch queuing system out of hdfs to local");
        }

        System.out.println("Copy from HDFS to local sussessful...");

        return result;
    }


    public static boolean fetchQueuingSystemCode_v2(Logger logger) {

        boolean result = true;
        int retCode = 0;

        Configuration conf = new Configuration();
        conf.set(ServerConfig.getString("yarn_resourcemanager_hostname_key"),
                 ServerConfig.getString("yarn_resourcemanager_hostname_val"));
        conf.set(ServerConfig.getString("fs_default_name_key"),
                 ServerConfig.getString("fs_default_name_val"));
        try {
            FileSystem fs = FileSystem.get(conf);
            org.apache.hadoop.fs.Path src = new org.apache.hadoop.fs.Path(
                ServerConfig.getString(ServerConfig.Param.QUEUINGSYSTEMHDFSPATH));

            org.apache.hadoop.fs.Path dst = new org.apache.hadoop.fs.Path("./");
            RemoteIterator<LocatedFileStatus> sourceFiles = fs.listFiles(src, true);
            if (sourceFiles != null) {
                while (sourceFiles.hasNext()) {
                    //FileUtil.copy(fs, sourceFiles.next().getPath(), fs, dst, false, conf);
                    fs.copyToLocalFile(sourceFiles.next().getPath(), dst);
                }
            }


        } catch (IOException e) {
            e.printStackTrace();
            retCode = 1;
        }


        if (retCode != 0 && !(new File(ServerConfig.getString("localQueueSysName"))).exists()) {
            result = false;
            logger.info("Failed to fetch queuing system out of hdfs to local");
        }

        System.out.println("Copy from HDFS to local sussessful...");

        return result;
    }

    public static boolean fetchQueuingSystemCode_v1(Logger logger) {

        boolean result = true;


        int retCode = HDFSHelper.copyToLocal(ServerConfig.getString("queuingSystem"), "./", logger);

        if (retCode != 0 && !(new File(ServerConfig.getString("localQueueSysName"))).exists()) {
            result = false;
            logger.info("Failed to fetch queuing system out of hdfs to local");
        }

        System.out.println("Copy from HDFS to local sussessful...");

        return result;
    }


    public static boolean processUserJobFileWithOutpus(YAMLSerAndDeSer yamlProcessor, String YAMLFile, Logger logger) {

        String YAMLOut = String.format("%s_%s", YAMLFile, "server.yaml");
        String YAMLDot = String.format("%s_%s", YAMLFile, "server.dot");
        String YAMLGraphViz = String.format("%s_%s", YAMLFile, "server.png");

        String message = String.format("out = %s \n dot = %s \n graph = %s \n", YAMLOut, YAMLDot, YAMLGraphViz);
        logger.info(message);

        boolean result = true;

        try {
            result = yamlProcessor.readYAMLFile(YAMLFile);
            if (!result) return result;

            result = yamlProcessor.writeYAMLFile(YAMLOut);
            if (!result) return result;

            result = yamlProcessor.makeDotGraph("Visualize", YAMLDot, YAMLGraphViz);

        } catch (IOException | WindowParsingFailedException | InputQueueParsingFailedException e) {
            result = false;
        }

        return result;
    }

    public static boolean processUserJobFile(String YAMLFile) {

        boolean result1 = true;
        boolean result2 = true;
        YAMLSerAndDeSer yamlProcessor = new YAMLSerAndDeSer();


        try {
            String file = "/tmp/_tmp_.yaml";
            result1 = yamlProcessor.readYAMLFile(YAMLFile);
            // If user yaml does not have all of IPO, then following fails.
            // It is a stop-gap check here until a better checking strategy.
            result2 = yamlProcessor.writeYAMLFile(file);
            File f = new File(file);
            if(f.exists()) f.delete();



        } catch (IOException | WindowParsingFailedException | InputQueueParsingFailedException e) {
            result1 = false;
            result2 = false;
        }

        return (result1 && result2);
    }

    public static boolean processUserJobFile(YAMLSerAndDeSer yamlProcessor, String YAMLFile) {

        boolean result = true;


        try {
            result = yamlProcessor.readYAMLFile(YAMLFile);
            if (!result) return result;


        } catch (IOException | WindowParsingFailedException | InputQueueParsingFailedException e) {
            result = false;
        }

        return result;
    }

    public static boolean delFile(String aFile) {
        boolean result = false;

        File f = new File(aFile);
        if (f.exists()) {
            result = f.delete();
        }
        return result;
    }

    public static boolean concatenate(String from, String to) {
        System.out.println("from = " + from + " and to = " + to);

        FileReader fr = null;
        FileWriter fw = null;
        boolean result = true;


        try {
            fr = new FileReader(from);
            fw = new FileWriter(to, true);

            int c = fr.read();
            fw.write("\n\n");
            while (c != -1) {
                fw.write(c);
                c = fr.read();
            }


        } catch (IOException e) {

            result = false;
            e.printStackTrace();

        } finally {

            try {

                if (fw != null)
                    fw.close();

                if (fr != null)
                    fr.close();

            } catch (IOException ex) {

                ex.printStackTrace();

            }
        }


        return result;

    }

    public static String getBaseName(String path) {
        File theFile = new File(path);
        return theFile.getName();
    }

    public static String getDirName(String path) {
        File theFile = new File(path);
        return theFile.getParent();
    }

    public static String getClientSubmitDir(String clientDir) {
        return String.format("%s/%s", clientDir, ServerConfig.getString("submitted"));
    }

    public static String getClientRetractDir(String clientDir) {
        return String.format("%s/%s", clientDir, ServerConfig.getString("retracted"));
    }

    public static String getClientAcceptDir(String clientDir) {
        return String.format("%s/%s", clientDir, ServerConfig.getString("accepted"));
    }

    public static String getClientRejectDir(String clientDir) {
        return String.format("%s/%s", clientDir, ServerConfig.getString("rejected"));
    }

    public static void emptyScratch(String scratchSpace)
    {
        try {

            File scratch = new File(scratchSpace);
            if(scratch.exists())
            {
                FileUtils.cleanDirectory(scratch);
            }


        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    public static LocalResource makeLocalResource(String source, YarnConfiguration conf, boolean needsCopy) throws Exception
    {
        Path src = new Path(source);
        FileSystem fs = FileSystem.get(conf);
        String dst_string = String.format("%s/%s", ServerConfig.getString("HDFSScratchSpace"), src.getName());
        //Path dest = new Path(fs.getHomeDirectory(), src.getName());
        Path dest = new Path(dst_string);

        if (needsCopy)
        {
            System.out.format("Copying JAR from %s to %s%n", src, dest);
            fs.copyFromLocalFile(src, dest);
        } else if (!fs.exists(dest))
        {
            String msg = String.format("Didn't find file: %s on HDFS in dir: %s", src, dest);
            System.err.println(msg);
            throw new Exception(msg);
        }



        FileStatus fileStat = null;
        try {
            fileStat = FileSystem.get(conf).getFileStatus(dest);
        } catch (IOException e) {
            e.printStackTrace();
        }

        LocalResource aFile = Records.newRecord(LocalResource.class);
        aFile.setResource(ConverterUtils.getYarnUrlFromPath(dest));
        aFile.setSize(fileStat.getLen());
        aFile.setTimestamp(fileStat.getModificationTime());
        aFile.setType(LocalResourceType.FILE);
        aFile.setVisibility(LocalResourceVisibility.APPLICATION);

        return aFile;

    }

    public static String getYARNApplicationID(String containerID)
    {
        if (containerID == null) return "containerID_was_null.";
        //Expected kind of containerID: container_1591306856963_4207_01_000001
        Pattern pattern = Pattern.compile("([a-zA-Z]+)_(\\d+)_(\\d+)_(.)+", Pattern.CASE_INSENSITIVE);

        String result = "Failed to parse container id.";

        Matcher matcher = pattern.matcher(containerID);
        if(matcher.matches())
        {
            result = String.format("application_%s_%s", matcher.group(2), matcher.group(3));
        }
        return result;
    }

    public static int getTotalClusterCores()
    {
        int totalCores = ServerConfig.getInt("maxAllowedWorkersGlobally");
        if (ServerConfig.getInt("FindTotalVirtualCoresOnClusterDynamically") == 0) return totalCores;

        try {
            URL url = new URL(ServerConfig.getString("clusterMetricsURL"));
            HttpURLConnection connection = (HttpURLConnection) url.openConnection();
            connection.setRequestProperty("accept", "application/json");
            InputStream responseStream = connection.getInputStream();
            ObjectMapper mapper = new ObjectMapper();
            HashMap<String,HashMap<String,Integer>> map = mapper.readValue(responseStream, HashMap.class);
            System.out.println("Map size: " + map.size());
            HashMap<String,Integer> clusterMetrics = map.get("clusterMetrics");
            System.out.println("Total Virtual Cores: " + clusterMetrics.get("totalVirtualCores"));
            totalCores = clusterMetrics.get("totalVirtualCores");

        } catch (MalformedURLException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return totalCores;
    }

    public static String getNewFileName(String oldFullURL, String newQName) {
        //old == hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap_1/20151005-021001-00083337.lax.pcap.xz
        String fileName = oldFullURL.substring(oldFullURL.lastIndexOf("/")+1); // +1 because we want to keep last /
        System.out.println(fileName);

        // This parsing functionality is duplicated here (already done in regex.py of Queue)
        // See next ToDo  This function is a stop-gap solution until we are able to reuse queue's regex.py
        Pattern namePattern = Pattern.compile("(\\d{8})-(\\d{6})-(\\d+)((-([^.]+))?)\\.([^.]+)\\.(.+)", Pattern.CASE_INSENSITIVE);
        Matcher matcher = namePattern.matcher(fileName);
        String oldQName = "";
        if(!matcher.matches())
        {
            System.err.println("Unable to extract queue name from file name string.");
            System.exit(100);
        }
        else
        {
            oldQName = matcher.group(8);
        }

        System.out.println("oldFullURL = " + oldFullURL);
        System.out.println("oldQName = " + oldQName);
        System.out.println("newQName = " + newQName);
        int startIndex = oldFullURL.lastIndexOf(oldQName) - 1;
        if ((startIndex <= 0)) throw new AssertionError();

        String firstPart = oldFullURL.substring(0, startIndex+1); // Java substring is from startIndex to (endIndex -1)

        return String.format("%s%s", firstPart, newQName); // . already be there
    }

}
