package edu.isi.ant.JobAPI;

import io.grpc.ManagedChannel;
import io.grpc.lander.queue.api.*;
import io.grpc.netty.GrpcSslContexts;
import io.grpc.netty.NegotiationType;
import io.grpc.netty.NettyChannelBuilder;
import io.netty.handler.ssl.SslContext;

import javax.net.ssl.SSLException;
import java.io.File;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.util.Map;

public class NamingServices {

    private static final Logger logger = Logger.getLogger(JobAPIClient.class.getName());
    private final ManagedChannel channel;
    private final LanderQueueAPIImplementationGrpc.LanderQueueAPIImplementationBlockingStub blockingStub;
    private final String whoami = "Admin"; //System.getProperty("user.name");


    public NamingServices()throws SSLException
    {
        this(ServerConfig.getString("QueueServerHostName"), ServerConfig.getInt("QueueServerListeningPort"));
    }


    public NamingServices(String host, int port) throws SSLException {



        String clientKey = String.format("%s/%s", ServerConfig.getString("adminClientPrivateKeyDir"),
                ServerConfig.getString("adminClientPrivateKeyFileName"));
        if(!ServerConfig.doesLocalFileExists(clientKey))
        {
            System.err.println("Failed to find private key for user: " + whoami + " key looked for = " + clientKey);
            System.exit(1);
        }

        String clientCertificateChain = String.format("%s/%s", ServerConfig.getString("clientCertChainDir"),
                ServerConfig.getString("clientCertChainFileName"));
        if(!ServerConfig.doesLocalFileExists(clientCertificateChain))
        {
            System.err.println("Failed to find certificate chain for user: " + whoami + " chain looked for = " + clientCertificateChain);
            System.exit(2);
        }

        File roots_pem = new File(ServerConfig.getString("rootsCertChain"));
        File client_key = new File(clientKey); //new File(Constants.clientPrivateKey);
        File client_certificate_chain = new File(clientCertificateChain);//new File(Constants.clientCertChain);

        SslContext sslContext = GrpcSslContexts.forClient()
                .trustManager(roots_pem).keyManager(client_certificate_chain, client_key)
                .build();


        channel = NettyChannelBuilder.forAddress(host, port)
                .negotiationType(NegotiationType.TLS)
                .sslContext(sslContext)
                .build();

        blockingStub = LanderQueueAPIImplementationGrpc.newBlockingStub(channel);

    }

    public void shutdown() throws InterruptedException {
        channel.shutdown().awaitTermination(5, TimeUnit.SECONDS);
    }

    @Override
    public void finalize() {
        try {
            this.shutdown();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }


    public RegExReply getQueueName(String fileName)
    {
        RegExRequest.Builder requestBuilder = RegExRequest.newBuilder().setFileName(fileName);
        RegExRequest request = requestBuilder.build();
        RegExReply response;

        return this.blockingStub.getFileNameParts(request);
    }
}

class FileNameParts
{
    public String event_time_in_utc;
    public String sequence_number;
    public String additional_info;
    public String queue_sequence_number_instance;
    public String queue_name;
}
