package edu.isi.ant.JobAPI;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.logging.Logger;
import java.io.File;

public class QueueHelper
  {



    static final private ServerConfig CONF = ServerConfig.getInstance();
    static final private String path = CONF.getString("localQueueSysName") + "/shell/";
    static final private String shellScript = CONF.getString("localQueueSysName") + "/shell/call_lander_api_function_via_rpc.sh";
    static final private String user = System.getProperty("user.name");
    
    //ToDo: What is a right number of retries?
    //Earlier it was set to unlimited and due to any problem, we were stuck in functions
    //like releaseFile etc.
    private static final int MAXTRIES = 25;
    
    //....................................
    public static int createQueue(String qName, Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");
      logger.info("Current working dir: = " + currentDir);

      try {

        String cmd1 = String.format("source " + shellScript);
	String cmd2 = String.format("lander_queue_create_hle \"%s\"  \"raw\"  \"%s\"", qName, user);

	String cmd = String.format("%s && %s", cmd1, cmd2);
	//System.out.println(cmd);
	logger.info(cmd);
	
	File cwd = new File(path);

	
	ProcessBuilder pb = new ProcessBuilder("/bin/bash", "-c", cmd)
	                        .inheritIO()
				.directory(cwd);

        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      } catch(Exception e) {
        result = 1;
	e.printStackTrace();
      }

      return result;
    }

    //..................................................................
    public static int createQueueRepeatOnError(String qName, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = QueueHelper.createQueue(qName, logger);
        if (retCode != 0) {
          logger.info("createQueue " + qName + " retCode = " + retCode);
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }





    //................................................................

    public static int connectDerivedQueue(String from, String to, Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_connect_derived_queue_hle \"%s\"  \"%s\"   \"%s\"", from, to, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash", "-c", cmd)
                                .inheritIO()
                                .directory(cwd);

        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      } catch(Exception e) {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }

    //..................................................................
    public static int connectDerivedQueueRepeatOnError(String from, String to, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = QueueHelper.connectDerivedQueue(from, to, logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }

    //.................................................................
    public static int activateQueue(String qName, Logger logger)
    { 
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_activate_hle \"%s\"  \"%s\"", qName, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                                               "-c",
                                               cmd
                                              )
                                .inheritIO()
                                .directory(cwd);


        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      { 
        result = 1;
        e.printStackTrace();
      }

      return result;
    }

    //..................................................................
    public static int activateQueueRepeatOnError(String qName, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = QueueHelper.activateQueue(qName, logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }

    //..........................................................................

    public static QResult getQueueSize(String qName, Logger logger)
    {

      String currentDir = System.getProperty("user.dir");
      QResult result = new QResult();

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_size_hle \"%s\"  \"%s\"", qName, user);

        String cmd = String.format("echo XXX $PATH >/tmp/xxx; %s && %s", cmd1, cmd2);
        ///System.out.println(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                //.inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();
	

        BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
        StringBuilder builder = new StringBuilder();
        String line = null;
        while ( (line = reader.readLine()) != null) {
	  //logger.info("Reader read: " + line);
          builder.append(line);
          //builder.append(System.getProperty("line.separator"));
        }
        String qSizeString = builder.toString();
        System.out.println("Size = " + qSizeString);



        p.waitFor();
        result.retCode = p.exitValue();
        System.out.println("retCode = " + result.retCode);
        //ToFix: At some calls getQSize from db fails.  Need to see why. Is it out of db connections?
        ///logger.info(String.format("Queue %s has size %s", qName, qSizeString));
        if (result.retCode == 0) result.qSize = Integer.parseInt(qSizeString);
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result.retCode = 1;
        e.printStackTrace();
      }

      //System.out.println("*******************************************");

      return result;
    }

    //..................................................................
    public static QResult getQueueSizeRepeatOnError(String qName, Logger logger)
    {
      QueueHelper.QResult result;
      int localRetryBudget = MAXTRIES;
      do
      {
        result = QueueHelper.getQueueSize(qName, logger);
        if(result.retCode !=0 ) {
          try {
            Thread.sleep(20000L);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return result;
          }
        }
        localRetryBudget--;

      }while(result.retCode !=0 && localRetryBudget > 0);

      return result;
    }
    //...............................................................................



    public static QResult reserveTopFile(String qName, Logger logger, int threadId)
    {

      String currentDir = System.getProperty("user.dir");
      QResult result = new QResult();

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_reserve_top_without_local_copy_hle \"%s\"  \"%s\"", qName, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                //.inheritIO()
                .directory(cwd);


        pb.redirectError(ProcessBuilder.Redirect.INHERIT);

        hack(pb);
        Process p = pb.start();

        BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
        StringBuilder builder = new StringBuilder();
        String line = null;
        while ( (line = reader.readLine()) != null) {
          builder.append(line);
          //builder.append(System.getProperty("line.separator"));
        }
        String reservedFile = builder.toString();



        p.waitFor();
        result.retCode = p.exitValue();
        if (result.retCode == 0)
        {
            logger.info(String.format("[Thread No: %d] Queue %s has reserved file %s", threadId, qName, reservedFile));
            result.reservedHDFSFile = reservedFile;

        }

        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result.retCode = 1;
        e.printStackTrace();
      }

      return result;
    }

    //...............................................................................
    public static QResult reserveTopFileRepeatOnError(String qName, Logger logger, int threadId)
    {
      QueueHelper.QResult result;
      int localRetryBudget = MAXTRIES;

      do
      {
        result = reserveTopFile(qName, logger, threadId);
        if(result.retCode !=0 ) {
          result.qSize = QueueHelper.getQueueSizeRepeatOnError(qName, logger).qSize;
          if (result.qSize == 0) break;
          try {
            Thread.sleep(20000L);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return result;
          }
        }
        else
        {
          result.qSize = 1; // expectation is that, there was one file in
        }
        localRetryBudget--;

      }while(result.retCode !=0 && localRetryBudget > 0);

      return result;
    }
    //...............................................................................

    public static QResult getEnqueuURI(String qName, String file, Logger logger)
    {

      String currentDir = System.getProperty("user.dir");
      QResult result = new QResult();

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_getEnqueueURI_hle \"%s\"  \"%s\" \"%s\"", qName, file, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                //.inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();

        BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
        StringBuilder builder = new StringBuilder();
        String line = null;
        while ( (line = reader.readLine()) != null) {
          builder.append(line);
          //builder.append(System.getProperty("line.separator"));
        }
        String reservedFile = builder.toString();
        logger.info(String.format("Queue %s has reserved file %s", qName, reservedFile));
        result.reservedHDFSFile = reservedFile;


        p.waitFor();
        result.retCode = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result.retCode = 1;
        e.printStackTrace();
      }

      return result;
    }

    //...............................................................................
    public static QResult getEnqueuURIRepeatOnError(String qName, String file, Logger logger)
    {
      QueueHelper.QResult result;
      int localRetryBudget = MAXTRIES;
      do
      {
        result = QueueHelper.getEnqueuURI(qName, file, logger);
        if(result.retCode !=0 || result.reservedHDFSFile.isEmpty()) {
          try {
            Thread.sleep(20000L);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return result;
          }
        }
        localRetryBudget--;

      }while((result.retCode !=0 || result.reservedHDFSFile.isEmpty()) && localRetryBudget > 0);

      return result;
    }
    //...............................................................................

    public static QResult getDistributeURI(String qName, String file, Logger logger)
    {
      String currentDir = System.getProperty("user.dir");
      QResult result = new QResult();

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_getDistributeURI_hle \"%s\"  \"%s\" \"%s\"", qName, file, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                //.inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();

        BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
        StringBuilder builder = new StringBuilder();
        String line = null;
        while ( (line = reader.readLine()) != null) {
          builder.append(line);
          //builder.append(System.getProperty("line.separator"));
        }
        String reservedFile = builder.toString();
        logger.info(String.format("Queue %s has reserved file %s", qName, reservedFile));
        result.reservedHDFSFile = reservedFile;


        p.waitFor();
        result.retCode = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result.retCode = 1;
        e.printStackTrace();
      }

      return result;
    }

    //...............................................................................
    public static QResult getDistributeURIRepeatOnError(String qName, String file, Logger logger)
    {
      QueueHelper.QResult result;
      int localRetryBudget = MAXTRIES;
      do
      {
        result = getDistributeURI(qName, file, logger);
        if(result.retCode !=0 || result.reservedHDFSFile.isEmpty()) {
          try {
            Thread.sleep(20000L);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return result;
          }
        }
        localRetryBudget--;

      }while((result.retCode !=0 || result.reservedHDFSFile.isEmpty()) && localRetryBudget > 0);

      return result;
    }
    //...............................................................................

    public static int releaseFile(String qName, String fileURI, Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_release_hle \"%s\"  \"%s\"  \"%s\"", qName, fileURI, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                .inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }
    //...............................................................................
    public static int releaseFileRepeatOnError(String qName, String fileURI, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = releaseFile(qName, fileURI, logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }

    //................................................................
    public static int enqueueFile(String qName, String fileURI, Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_enqueue_remote_hle \"%s\"  \"%s\"  \"%s\"", qName, fileURI, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash", "-c", cmd)
                .inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }

    //...............................................................................
    public static int enqueueFileRepeatOnError(String qName, String fileURI, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = QueueHelper.enqueueFile(qName, fileURI, logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }

    //................................................................
    public static int distributeFile(String qName, String fileURI, Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_distribute_hle \"%s\"  \"%s\"  \"%s\"", qName, fileURI, user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash", "-c", cmd)
                .inheritIO()
                .directory(cwd);

        hack(pb);
        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }
    //...............................................................................
    public static int distributeFileRepeatOnError(String qName, String fileURI, Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = distributeFile(qName, fileURI, logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }





    //................................................................



    static class QResult
    {
      public int retCode;
      public int qSize;
      public String reservedHDFSFile;
    }



    // Follwong function was used in some debugging session and is helpful

    public void pwd(Logger logger)
    {


      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_reserve_top_without_local_copy_hle \"%s\"  \"%s\"", "abc", user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                "pwd"
        )
                //.inheritIO()
                .directory(cwd);


        Process p = pb.start();

        BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
        StringBuilder builder = new StringBuilder();
        String line = null;
        while ( (line = reader.readLine()) != null) {
          builder.append(line);
          //builder.append(System.getProperty("line.separator"));
        }
        String output = builder.toString();
        logger.info("pwd = " + output);



        p.waitFor();
        p.exitValue();

        p.destroy();
        pb = null;

      }catch(Exception e)
      {

        e.printStackTrace();
      }


    }

    public static int doErrorRecovery(String qName, boolean isWindow, Logger logger)
    {
      int howOld = CONF.getInt("errorRecoveryInterval"); //this is in minutes
      int maxFaultCount = CONF.getInt("maxRetires"); // 2 retries allowed
      int result = 0;
      String currentDir = System.getProperty("user.dir");
      logger.info("Current working dir: = " + currentDir);

      try {
        String isThisQueueHasWindow = isWindow?"yes":"no";
        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_do_Error_Recovery_hle \"%s\"  \"%d\" \"%d\"  \"%s\" \"%s\"",
                                                                    qName, howOld, maxFaultCount, isThisQueueHasWindow,
                                                                    user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                .inheritIO()
                .directory(cwd);




        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }

    //.................................................................
    public int deactivateAllQueues(Logger logger)
    {
      int result = 0;
      String currentDir = System.getProperty("user.dir");

      try {

        String cmd1 = String.format("source " + shellScript);
        String cmd2 = String.format("lander_queue_deactivate_all_queues_hle  \"%s\"", user);

        String cmd = String.format("%s && %s", cmd1, cmd2);
        //System.out.println(cmd);
        logger.info(cmd);

        File cwd = new File(path);


        ProcessBuilder pb = new ProcessBuilder("/bin/bash",
                "-c",
                cmd
        )
                .inheritIO()
                .directory(cwd);


        Process p = pb.start();

        p.waitFor();
        result = p.exitValue();
        p.destroy();
        pb = null;

      }catch(Exception e)
      {
        result = 1;
        e.printStackTrace();
      }

      return result;
    }

    //..................................................................
    public int deactivateAllQueuesRepeatOnError(Logger logger)
    {
      int retCode = 0;
      int localRetryBudget = MAXTRIES;
      do
      {
        retCode = deactivateAllQueues(logger);
        if (retCode != 0) {
          try {
            Thread.sleep(2000);
          } catch (InterruptedException e) {
            e.printStackTrace();
            return retCode;
          }
        }
        localRetryBudget--;

      }while(retCode != 0 && localRetryBudget > 0);

      return retCode;
    }

    //..........................................................................



    public static void hack(ProcessBuilder pb)
    {
      // hack disabled by aqadeer on June 26, 2019
      //pb.environment().put("HOME", "/nfs/jade/hadoop_u");
      //pb.environment().put("USER", "hadoop_u");

    }



  }//end of class

