package edu.isi.ant.JobAPI;

import java.io.File;
import java.nio.file.Files;
import java.util.logging.Logger;

import org.apache.commons.configuration2.builder.fluent.Configurations;
import org.apache.commons.configuration2.Configuration;
import org.apache.commons.configuration2.ex.ConfigurationException;

//TODO add mapping enum to strings
public class ServerConfig {
    public static final String CONFIGURATION_FILE = "server.config";
    //Singleton
    private static ServerConfig _instance = null;
    private Configuration conf = null;

    public enum Param {
        QUEUINGSYSTEMHDFSPATH("queuingSystemHDFSPath"),
        //TODO add enums for all strings and use them in code so renaming is easy
        //(probably do initial, hierarchical renaming first)
        TODOMORE("todoMORE");

        public final String name;
       
        private Param(String name) { this.name = name; }
        public String toString() { return this.name; }
        public String n() { return this.name; }
    }

    private ServerConfig() {
        Configurations configs = new Configurations();
        String fileName = CONFIGURATION_FILE;
        try {
            conf = configs.properties(new File(fileName));
            //conf.setProperty("scratchSpace", System.getProperty("user.home") + "/" + conf.getString("scratchSpace"));
        } catch (ConfigurationException cex) {
            System.err.println("Can't read config file: " + fileName);
            System.err.println(cex);
            System.exit(1);
        }
    }

    public static ServerConfig getInstance() {
        if (_instance == null) {
            _instance = new ServerConfig();
        }
        return _instance;
    }
    public static Configuration getConfiguration() {
        return getInstance().conf;
    }

    public static String getString(String key) {
        return getConfiguration().getString(key);
    }
    public static String getString(Param p) {
        return getConfiguration().getString(p.n());
    }
    public static String getString(String key, String defaultValue) {
        return getConfiguration().getString(key, defaultValue);
    }
    public static String getString(Param p, String defaultValue) {
        return getConfiguration().getString(p.n(), defaultValue);
    }
    public static int getInt(String key) {
        return getConfiguration().getInt(key);
    }
    public static int getInt(String key, int defaultValue) {
        return getConfiguration().getInt(key, defaultValue);
    }

    public static void verify(Logger logger) {
        ServerConfig c = getInstance();

        // vital
        if (!doesHDFSDirExists(c.getString(Param.QUEUINGSYSTEMHDFSPATH))) {
            System.err.println("queuingSystemHDFSPath not present");
            System.exit(1);
        }
        if (!doesLocalDirExists(c.getString("localQueueSysName"))) {
            System.err.println("localQueueSysName not present");
            System.exit(2);
        }
        if (!doesLocalFileExists(c.getString("serverPrivateKey")))
        {
            System.err.println("serverPrivateKey not present");
            System.exit(3);
        }
        if (!doesLocalFileExists(c.getString("serverCertChain"))) {
            System.err.println("serverCertChain not present.");
            System.exit(4);
        }
        if (!doesLocalFileExists(c.getString("rootsCertChain"))) {
            System.err.println("rootsCertChain not present.");
            System.exit(5);
        }
        if (!doesLocalDirExists(c.getString("internalScriptsPath"))) {
            System.err.println("internalScriptsPath not present");
            System.exit(6);
        }
        if (!makeLocalDirIfDoesntExist(c.getString("scratchSpace"))) {
            System.err.println("scratchSpace " + c.getString("scratchSpace") + " can not be created.");
            System.exit(7);
        }
        // make dir if not found
        String HDFSClientBaseDirPath = c.getString("HDFSClientBaseDirPath");
        if (!doesHDFSDirExists(HDFSClientBaseDirPath)) {
            HDFSHelper.makeDir(HDFSClientBaseDirPath, logger);
        }
        String HDFSOptBaseDirPath = c.getString("HDFSOptBaseDirPath");
        if (!doesHDFSDirExists(HDFSOptBaseDirPath))
        {
            HDFSHelper.makeDir(HDFSOptBaseDirPath, logger);
        }
        String dupDispStateSavePathInHDFS = c.getString("dupDispStateSavePathInHDFS");
        if (!doesHDFSDirExists(dupDispStateSavePathInHDFS)) {
            HDFSHelper.makeDir(dupDispStateSavePathInHDFS, logger);
        }
        String userProgHDFSDir = c.getString("userProgHDFSDir");
        if (!doesHDFSDirExists(userProgHDFSDir)) {
            HDFSHelper.makeDir(userProgHDFSDir, logger);
            HDFSHelper.chmod(userProgHDFSDir, "777");
        }
    }

    private static boolean doesLocalDirExists(String name) {
        File f = new File(name);
        return (f.exists() && f.isDirectory());
    }

    public static boolean doesLocalFileExists(String name) {
        File f = new File(name);
        return (f.exists() && f.isFile());
    }

    private static boolean doesHDFSDirExists(String name) {
        return (HDFSHelper.doesXExists(name, true) == 0);
    }

    public static boolean doesHDFSFileExists(String name) {
        return (HDFSHelper.doesXExists(name, false) == 0);

    }

    public static boolean makeLocalDirIfDoesntExist(String name) {
        File dir = new File(name);
        if(dir.exists()) return true;

        return dir.mkdir();
    }
}
