// code copied from: https://github.com/alexholmes/hiped2/blob/master/src/main/java/hip/ch10/dstat/ApplicationMaster.java
// and then modified

package edu.isi.ant.JobAPI;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.isi.ant.yamlseranddeser.Stage;
import io.grpc.lander.queue.api.RegExReply;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.yarn.api.protocolrecords.AllocateResponse;
import org.apache.hadoop.yarn.api.records.FinalApplicationStatus;
import org.apache.hadoop.yarn.client.api.AMRMClient;
import org.apache.hadoop.yarn.client.api.AMRMClient.ContainerRequest;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.exceptions.YarnException;

import javax.net.ssl.SSLException;


//ToDo: Current code assumes that there are only single inputs to each
//stage and hence uses reserveTop function.  On multiple queues we need
//to be very careful that which two files to reserve from two different
//queues.
//ToDo: With multi-input stages, there can be a host of concurrency issues
//and we will need careful thought and probably we will need transaction like
//facility where we alwys are in good state.
// Probably enforcing windowing of size 1 block on all inputs do the trick
// Above will be simpler when across-window-scheduling is off but that can be slow.
// With across-window-scheduling on, again things can be quite tricky.


/**
 * The ApplicationMaster, which is responsible for sending container
 * requests to the ResourceManager, and talking to NodeManagers when
 * the RM provides us with available containers.
 */
public class YARNApplicationMaster {
    private static long t1,t2;

    public static void main(String[] args)  {
        t1 = System.nanoTime();
        int retCode = -1;
        Logger logger = Logger.getAnonymousLogger();
        logger.addHandler(new StreamHandler(System.out, new SimpleFormatter()));
        t2 = System.nanoTime();
        logger.info("time to get loggers = " + (t2-t1)/1000000000);


        t1 = System.nanoTime();
        /* History: We stopped copying and fetching code from HDFS
           Now we assume latest version is on the server.
        // Fetch out QueuingSystem code out of HDFS
        if(!MyUtils.fetchQueuingSystemCode(logger))
        {
            logger.info("Failed to fetch out Queuing system code out of HDFS.");
            return;
        }
        */

        Map<String, Integer> dupTable = (Map<String, Integer>) MyUtils.deSeralizeObjectFromHDFS(
                ServerConfig.getString("dupPathForAppMaster"),
                ServerConfig.getString("dupPathInHDFS"), logger);
        //dupTable = (Map<String, Integer>) MyUtils.deSeralizeObjectFromLocal(ServerConfig.getString("dupPath"), logger);

        Map<String, Stage> dispTable = (Map<String, Stage>) MyUtils.deSeralizeObjectFromHDFS(
                ServerConfig.getString("dispPathForAppMaster"),
                ServerConfig.getString("dispPathInHDFS"), logger);
        //dispTable = (Map<String, Stage>) MyUtils.deSeralizeObjectFromLocal(ServerConfig.getString("dispPath"), logger);
        t2 = System.nanoTime();
        logger.info("time to fetch queuesystem, dup and disp from hdfs = " + (t2-t1)/1000000000);


        /*
        System.out.println("Stage making");
        Stage x = new Stage(null, null, null);
        System.out.println("Done...");
        run(new ArrayList<String>());
        ClassLoader cl = ClassLoader.getSystemClassLoader();

        URL[] urls = ((URLClassLoader)cl).getURLs();

        for(URL url: urls){
            System.out.println(url.getFile());
        }

        System.out.println("Dump debug output");
        Map<String, String> envs1 = System.getenv();
        System.out.println("classpath = " +envs1.get("CLASSPATH"));
        System.exit(0);
        */


        t1 = System.nanoTime();



        try {

            Configuration conf = new YarnConfiguration();

            // create a client to talk to the ResourceManager
            AMRMClient<ContainerRequest> rmClient = AMRMClient.createAMRMClient();
            rmClient.init(conf);
            rmClient.start();




            // register with ResourceManager
            System.out.println("registerApplicationMaster: pending");
            rmClient.registerApplicationMaster("", 0, "");
            System.out.println("registerApplicationMaster: complete");


            // sending pings to RM periodically so that it stays alive
            Thread t = new Thread() {
                public void run() {
                    //Random random = new Random();
                    long execTimeInMinutes = 0;

                    while (true) {
                        try {
                            AllocateResponse response = rmClient.allocate(0);
                            TimeUnit.SECONDS.sleep(60);
                            //ToDo:Measuring wall-clock time or duration is not a good idea using
                            //sleep because under load, sleep might fire late.
                            //I tried using System.currentTimeMillis()/(1000*60) but it is performing
                            //worse (e.g. 5 seconds elapses but it shows 0 or 1)
                            execTimeInMinutes += 1;

                            // We don't want duplicate jobs running on same file(s)
                            if(ServerConfig.getInt("maxYARNJobRunTime") < execTimeInMinutes)
                            {
                                System.out.println("Max time to run a YARN job exceeded. Exiting...");
                                System.exit(5);

                            }
                        } catch (YarnException | InterruptedException | IOException e) {
                            e.printStackTrace();
                        }

                        // Following only for testing failure recovery. 10% error rate
                        //if(random.nextInt(100) < 10)
                        //{
                        //    System.exit(2);
                        //}

                    }
                }
            };
            t.start();
            t2 = System.nanoTime();
            logger.info("time to register AM and starting ping thread = " + (t2-t1)/1000000000);


            t1 = System.nanoTime();

            // ToDo: I will come back and add my code here later after initial skeleton is working
            System.out.println("Dump debug output");
            Map<String, String> envs = System.getenv();
            for (Map.Entry<String, String> env : envs.entrySet()) {
                System.out.println("System env: key=" + env.getKey() + ", val=" + env.getValue());

            }

            t2 = System.nanoTime();
            logger.info("time to print environment = " + (t2-t1)/1000000000);



            t1 = System.nanoTime();
            retCode = runIteration(dispTable, dupTable, envs, logger);
            //retCode = run(envs);
            t2 = System.nanoTime();
            logger.info("time to runIteration = " + (t2-t1)/1000000000);







            System.out.println("unregister");
            // Un-register with ResourceManager
            if(retCode == 0)
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.SUCCEEDED, "", "");
            else
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.FAILED, "", "");

            System.out.println("exiting");
        } catch(Throwable t) {
            t.printStackTrace();
        }

        System.exit(retCode);
    }

    private static int run(List<String> args,
                           Stage stage,
                           List<TezQRep> in,
                           List<TezQRep> out) {
        int retCode = 1;
        try {



            args.add(0, "./AppMaster.shellscript");




            ProcessBuilder pb = new ProcessBuilder(args)
                    .inheritIO();


            pb.redirectErrorStream(true);
            setEnv(pb.environment(), stage, in, out);
            Process p = pb.start();
            retCode = p.waitFor();
            BufferedReader reader =  new BufferedReader(new InputStreamReader(p.getInputStream()));
            String line = "";
            while ( (line = reader.readLine()) != null) {
            	System.out.println(line);
            }

            System.out.println("Retval = " + retCode);
        } catch (Exception exp) {
            exp.printStackTrace();
        }

        return retCode;


    }



    private static String getReservedFile(Stage stage, String q, Logger logger) {
        QueueHelper.QResult result;
        //ToDo: The following can become an infinite loop in a scenario, where the file(s) this stage was supposed
        // to do is already done by someone else.
        do {
            result = QueueHelper.reserveTopFileRepeatOnError(q, logger, 1);
            if (result.qSize == 0)
            {
                result.reservedHDFSFile = null;
                break;
            }

        }
        while (result.retCode != 0);

        return result.reservedHDFSFile;
    }

    // Follwing functin is depricated.  Use its variant instead.
    private static String getNewFileName_v1(String old, String newExt) {
        //old == hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap_1/20151005-021001-00083337.lax.pcap.xz
        System.out.println("old = " + old + " and newExt = " + newExt);
        String firstPart = old.substring(0, old.lastIndexOf("/"));
        String lastPart = old.substring(old.lastIndexOf("/") + 1);
        String lastPartWithoutExt = lastPart.substring(0, lastPart.indexOf("."));
        int bi = lastPart.indexOf(".")+1;
        int ei = bi + 3; // +3 because String's substring actually goes till ei - 1
        String siteName = lastPart.substring(bi, ei);

        return String.format("%s/%s.%s.%s", firstPart, lastPartWithoutExt, siteName, newExt);
    }

    private static String getNewFileName(String oldFullURL, String newQName) {
        //old == hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap_1/20151005-021001-00083337.lax.pcap.xz
        String fileName = oldFullURL.substring(oldFullURL.lastIndexOf("/")+1); // +1 because we want to keep last /
        System.out.println(fileName);

        // This parsing functionality is duplicated here (already done in regex.py of Queue)
        // See next ToDo  This function is a stop-gap solution until we are able to reuse queue's regex.py
        Pattern namePattern = Pattern.compile("(\\d{8})-(\\d{6})-(\\d+)((-([^.]+))?)\\.([^.]+)\\.(.+)", Pattern.CASE_INSENSITIVE);
        Matcher matcher = namePattern.matcher(fileName);
        String oldQName = "";
        if(!matcher.matches())
        {
            System.err.println("Unable to extract queue name from file name string.");
            System.exit(100);
        }
        else
        {
            oldQName = matcher.group(8);
        }

        System.out.println("oldFullURL = " + oldFullURL);
        System.out.println("oldQName = " + oldQName);
        System.out.println("newQName = " + newQName);
        int startIndex = oldFullURL.lastIndexOf(oldQName) - 1;
        if ((startIndex <= 0)) throw new AssertionError();

        String firstPart = oldFullURL.substring(0, startIndex+1); // Java substring is from startIndex to (endIndex -1)

        return String.format("%s%s", firstPart, newQName); // . already be there
    }

    /*
    ToDO:
    The purpose of following function was to reuse regex.py from the queue code base via
    an RPC.  This effort initially failed because required gPRC, netty, io.chanel etc classes
    were not present to this code (Remember it runs on some cluster box).  I tried different
    things but they failed:
    (1) I tried making a fat jar (everything in one big jar) by overriding jar in gradle build.
    The problem was that different jars were overwriting each other in META-INF and when I tried
    to run my JoApiServcie main class, code couldn't find it.
    (2) Then I tried a gradle plug-in called shadow-jar but that I couldn't make it work.
    (3) Yet an unexplored option is to copy required jars (just like our main jar is done in Dispatcher
    one in the HDFS) and then keep copying it via local resource.  One thing to observe is that will
    it slow down the whole process?
    (4) So finding a reasonable way to do the work is still needs to be done.
     */

    private static String getNewFileName_forfuture(String oldFullURL, String newQName) {
        //old == hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap_1/20151005-021001-00083337.lax.pcap.xz
        String fileName = oldFullURL.substring(oldFullURL.lastIndexOf("/"));
        String oldQName = "";
        try {
            NamingServices ns = new NamingServices();
            RegExReply response = ns.getQueueName(fileName);
            if (!response.getResult())
            {
                System.err.println("Unable to contact queue service.");
                System.exit(100);
            }
            oldQName = response.getQueueName();
        } catch (SSLException e) {
            e.printStackTrace();
            System.err.println("Unable to contact queue service.");
            System.exit(101);
        }
        System.out.println("oldFullURL = " + oldFullURL);
        System.out.println("oldQName = " + oldQName);
        System.out.println("newQName = " + newQName);
        int startIndex = oldFullURL.lastIndexOf(oldQName) - 1;
        if ((startIndex <= 0)) throw new AssertionError();

        String firstPart = oldFullURL.substring(0, startIndex);

        return String.format("%s%s", firstPart, newQName); // . already be there
    }

    private static String getNewFileName_v20200519(String oldFullURL, String oldQName, String newQName) {
        // HEADS UP: This function produces wrong result for derived queues because oldQName comes as
        // pcap.sz_1 etc while we expected pcap.sz
        //old == hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap_1/20151005-021001-00083337.lax.pcap.xz
        System.out.println("oldFullURL = " + oldFullURL);
        System.out.println("oldQName = " + oldQName);
        System.out.println("newQName = " + newQName);
        int startIndex = oldFullURL.lastIndexOf(oldQName) - 1;
        if ((startIndex <= 0)) throw new AssertionError();

        String firstPart = oldFullURL.substring(0, startIndex);

        return String.format("%s.%s", firstPart, newQName);
    }

    private static List<String> getCmdLineForYarn(Stage stage,
                                           List<TezQRep> in,
                                           List<TezQRep> out) {
        List<String> builder = new ArrayList<>();


        for (TezQRep tqr : in) {
            //dont put any spaces in following parameters else getopts
            //in bash fails (probably process builder puts in these spaces
            //in some way that causes getopts to stop
            builder.add(String.format("%s\"%s\"", "-i", tqr.hdfsURI));
        }

        for (String s : stage.getProgram()) {
            builder.add(String.format("%s\"%s\"", "-p", s));
        }

        for (TezQRep tqr : out) {
            builder.add(String.format("%s\"%s\"", "-o", tqr.hdfsURI));
        }

        return builder;
    }

    private static List<String> getCmdLineForYarnv2(Stage stage,
                                                  List<TezQRep> in,
                                                  List<TezQRep> out) {
        List<String> builder = new ArrayList<>();


        for (TezQRep tqr : in) {
            //dont put any spaces in following parameters else getopts
            //in bash fails (probably process builder puts in these spaces
            //in some way that causes getopts to stop
            builder.add(String.format("%s%s", "-i", tqr.hdfsURI));
        }

        for (String s : stage.getProgram()) {
            builder.add(String.format("%s%s", "-p", s));
        }

        for (TezQRep tqr : out) {
            builder.add(String.format("%s%s", "-o", tqr.hdfsURI));
        }

        return builder;
    }

    private static List<String> getCmdLineForYarnv3(Stage stage,
                                                    List<TezQRep> in,
                                                    List<TezQRep> out) {
        List<String> builder = new ArrayList<>();


        for (TezQRep tqr : in) {
            //dont put any spaces in following parameters else getopts
            //in bash fails (probably process builder puts in these spaces
            //in some way that causes getopts to stop
            builder.add(String.format("%s%s", "-i", tqr.hdfsURI));
        }

        for (String s : stage.getProgram()) {
            s = s.replaceAll("\"", "'");
            builder.add(String.format("%s\"%s\"", "-p", s));
        }

        for (TezQRep tqr : out) {
            builder.add(String.format("%s%s", "-o", tqr.hdfsURI));
        }

        return builder;
    }


    private static int runIteration(Map<String, Stage> dispTable,
                              Map<String, Integer> dupTable,
                              Map<String, String> envs,
                                     Logger logger)
    {
        t1 = System.nanoTime();


        Stage stage = dispTable.get(envs.get("args_1")); // args_1 should be stage ID

        List<String> inQueues = stage.getInput();
        List<TezQRep> in = new ArrayList<TezQRep>();

        List<String> outQueues = stage.getOutput();
        List<TezQRep> out = new ArrayList<TezQRep>();

        // wait until all input queues have at least one item in it
        for(String q:inQueues)
        {


            // ToFix: This logic wont work when reading from multiple queues
            // ToDo: There is real danger of deadlock when reserving for multiple
            //       inputs.  Either all need to use same order (like it is done in
            //       GFS or we need to make multi-file reservation atomic.

            // reserve top file in the queue
            String aReservedFile = getReservedFile(stage, q, logger);
            // ToDo: For multiple inputs, it might be the case that few inputs are reserved while on some it
            // fails.  In that case we need to un-reserve already reserved files.
            if (aReservedFile == null)
            {
                return 5;
            }
            
            TezQRep tqr = new TezQRep(q, aReservedFile);
            in.add(tqr);
        }

        // tmp hack
        // ToDo: Following works okay for stages that have a singel input.  As of 2020-05-16 we don't have a use case
        // for multiple inputs into the stage.  For that, name of output will be extracted based on the name of the
        // first input.  That might be okay or might not be!
        String file = in.get(0).hdfsURI;
        String qOfFile = in.get(0).qName;


        boolean isAnyOutputQueueNone = isAnyQueueNone(outQueues);

        if(!isAnyOutputQueueNone) {
            // get output queues(s) URIs
            for (String q : outQueues) {
                int count = dupTable.get(q);
                // Actually it doesn't matter what is the intiail part of the newFileName.  The thing that matter is
                // the last part of the name (basename).  getEnqueuURI and getDistribute methods below only extract
                // the last part of the name while they construct the rest by themselves.
                String newFileName = getNewFileName(file, q);


                if (count <= 1) {
                    logger.info("Asking for enque URL " + q + " and newFileName = " + newFileName);
                    TezQRep tqr = new TezQRep(q, QueueHelper.getEnqueuURIRepeatOnError(q, newFileName, logger).reservedHDFSFile);

                    out.add(tqr);
                } else {
                    logger.info("Asking for distribute URL " + q + " and newFileName = " + newFileName);
                    TezQRep tqr = new TezQRep(q, QueueHelper.getDistributeURIRepeatOnError(q, newFileName, logger).reservedHDFSFile);
                    out.add(tqr);
                }
            }
        }else
        {
            out.add(new TezQRep("None", "None"));
        }

        List<String> cmdLine = getCmdLineForYarnv2(stage, in, out);
        logger.info("cmdLine = " + Arrays.toString(cmdLine.toArray())+ "inLenght="+in.size()+"OutLength="+out.get(0).hdfsURI.length());
        t2 = System.nanoTime();
        logger.info("time for file resevations etc. = " + (t2-t1)/1000000000);

        t1 = System.nanoTime();
        int retCode = run(cmdLine, stage, in, out);
        t2 = System.nanoTime();
        logger.info("Return code form run = " + retCode);
        logger.info("time run job = " + (t2-t1)/1000000000);
        t1 = System.nanoTime();
        if(retCode == 0)
        {

            // First enqueu new files then, release file(s).  This is because
            // killer thread can kill at the wrong time, when files are released
            // but newly made files are not written in DB.
            //ToDO: Following correct behavior is based on ordering of events.
            // Generally this is not good in concurrent programming.  Probably
            // use locks here.


            if(!isAnyOutputQueueNone) {
                for (TezQRep tqr : out) {

                    int count = dupTable.get(tqr.qName);

                    if (count <= 1) {
                        logger.info("Enquing file " + tqr.hdfsURI + " to queue = " + tqr.qName);
                        QueueHelper.enqueueFileRepeatOnError(tqr.qName, tqr.hdfsURI, logger);
                    } else {
                        logger.info("Distributing file " + tqr.hdfsURI + " to queue = " + tqr.qName);
                        QueueHelper.distributeFileRepeatOnError(tqr.qName, tqr.hdfsURI, logger);
                    }

                }
            }

            for (TezQRep tqr:in)
            {

                logger.info("Releasing file " + tqr.hdfsURI + " from queue = " + tqr.qName );

                QueueHelper.releaseFileRepeatOnError(tqr.qName, tqr.hdfsURI, logger);

            }


        } else
        {
            logger.info("Execution failed...");
        }
        t2 = System.nanoTime();
        logger.info("time post job run = " + (t2-t1)/1000000000);


        return retCode;

    }




    static class StageControllerCell {
        public String stageName;
        public Stage stage;
        public int backlog;
        public int maxAllowedWorkers;

    }


    static class TezQRep {

        public TezQRep(String qName, String hdfsURI) {
            this.qName = qName;
            this.hdfsURI = hdfsURI;
        }

        public String qName;
        public String hdfsURI;
    }

    private static boolean isAnyQueueNone(List<String> outQueues)
    {
        boolean result = false;
        for(String q:outQueues)
        {
            if (q.equalsIgnoreCase("None"))
            {
                result = true;
                break;
            }
        }
        return result;
    }

    private static void setEnv(Map<String, String> env,
                               Stage stage,
                               List<TezQRep> in,
                               List<TezQRep> out)
    {

        env.put("__plumb_no_of_inputs", in.size()+"");
        int i = 1;
        for (TezQRep tqr : in)
        {
            env.put("__plumb_i"+i, tqr.hdfsURI);
            i++;
        }



        env.put("__plumb_no_of_outputs", out.size()+"");
        i = 1;
        for (TezQRep tqr : out)
        {
            env.put("__plumb_o"+i, tqr.hdfsURI);
            i++;
        }

        String userProg = "";

        for (String s : stage.getProgram()) {
            userProg += s;
        }

        String parts[] = userProg.split(" ", 2);
        if (null != parts && parts.length > 0)
        {
            env.put("__plumb_prog_name", parts[0]);
        }

        if (null != parts && parts.length > 1)
        {
            env.put("__plumb_prog_args", parts[1]);
        }

        // Put path to the Queue regex.py so that yarn wrapper can sym link to it
        env.put("pathToRegEx", ServerConfig.getString("pathToRegEx"));

    }
}
