package edu.isi.ant.JobAPI;

import edu.isi.ant.yamlseranddeser.Stage;
import edu.isi.ant.yamlseranddeser.YAMLSerAndDeSer;
import io.grpc.lander.queue.api.RegExReply;
import io.grpc.lander.queue.api.WindowOperationResponse;
import io.grpc.lander.queue.api.WindowState;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.yarn.api.protocolrecords.AllocateResponse;
import org.apache.hadoop.yarn.api.records.FinalApplicationStatus;
import org.apache.hadoop.yarn.client.api.AMRMClient;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.exceptions.YarnException;
import org.apache.hadoop.yarn.util.Apps;

import javax.net.ssl.SSLException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static edu.isi.ant.JobAPI.MyUtils.getNewFileName;


public class YARNApplicationMasterForWindows {
    private static long t1,t2;
    private static WindowClient windowClient;
    private static String myName;

    public static void main(String[] args)  {
        t1 = System.nanoTime();
        int retCode = -1;
        Logger logger = Logger.getAnonymousLogger();
        logger.addHandler(new StreamHandler(System.out, new SimpleFormatter()));
        t2 = System.nanoTime();
        logger.info("time to get loggers = " + (t2-t1)/1000000000);


        t1 = System.nanoTime();
        Map<String, Integer> dupTable;
        dupTable = (Map<String, Integer>) MyUtils.deSeralizeObjectFromHDFS(ServerConfig.getString("dupPathForAppMaster"),
                ServerConfig.getString("dupPathInHDFS"), logger);


        Map<String, Stage> dispTable;
        dispTable = (Map<String, Stage>) MyUtils.deSeralizeObjectFromHDFS(ServerConfig.getString("dispPathForAppMaster"),
                ServerConfig.getString("dispPathInHDFS"), logger);

        t2 = System.nanoTime();
        logger.info("time to fetch dup and disp from hdfs = " + (t2-t1)/1000000000);




        t1 = System.nanoTime();



        try {

            Configuration conf = new YarnConfiguration();

            // create a client to talk to the ResourceManager
            AMRMClient<AMRMClient.ContainerRequest> rmClient = AMRMClient.createAMRMClient();
            rmClient.init(conf);
            rmClient.start();




            // register with ResourceManager
            System.out.println("registerApplicationMaster: pending");
            rmClient.registerApplicationMaster("", 0, "");
            System.out.println("registerApplicationMaster: complete");
            //myName = rmClient.getName(); // getName gives us: org.apache.hadoop.yarn.client.api.impl.AMRMClientImpl
            myName = MyUtils.getYARNApplicationID(System.getenv("CONTAINER_ID"));
            windowClient = new WindowClient(ServerConfig.getString("QueueServerHostName"), ServerConfig.getInt("QueueServerListeningPort"));


            // sending pings to RM periodically so that it stays alive
            Thread t = new Thread() {
                public void run() {
                    //Random random = new Random();
                    long execTimeInMinutes = 0;

                    while (true) {
                        try {
                            AllocateResponse response = rmClient.allocate(0);
                            TimeUnit.SECONDS.sleep(60);
                            //ToDo:Measuring wall-clock time or duration is not a good idea using
                            //sleep because under load, sleep might fire late.
                            //I tried using System.currentTimeMillis()/(1000*60) but it is performing
                            //worse (e.g. 5 seconds elapses but it shows 0 or 1)
                            execTimeInMinutes += 1;

                            // We don't want duplicate jobs running on same file(s)
                            if(ServerConfig.getInt("maxYARNJobRunTime") < execTimeInMinutes)
                            {
                                System.out.println("Max time to run a YARN job exceeded. Exiting...");
                                System.exit(5);

                            }
                        } catch (YarnException | InterruptedException | IOException e) {
                            e.printStackTrace();
                        }
                    }
                }
            };
            t.start();
            t2 = System.nanoTime();
            logger.info("time to register AM and starting ping thread = " + (t2-t1)/1000000000);


            t1 = System.nanoTime();


            System.out.println("Dump debug output");
            Map<String, String> envs = System.getenv();
            for (Map.Entry<String, String> env : envs.entrySet()) {
                System.out.println("System env: key=" + env.getKey() + ", val=" + env.getValue());

            }

            t2 = System.nanoTime();
            logger.info("time to print environment = " + (t2-t1)/1000000000);



            t1 = System.nanoTime();
            retCode = runIteration(dispTable, dupTable, envs, logger);
            t2 = System.nanoTime();
            logger.info("time to runIteration = " + (t2-t1)/1000000000);

            // Test we can call window client functions from here
            windowClient.testCommunication();







            System.out.println("unregister");
            // Un-register with ResourceManager
            if(retCode == 0)
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.SUCCEEDED, "", "");
            else
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.FAILED, "", "");

            System.out.println("exiting");
        } catch(Throwable t) {
            t.printStackTrace();
        }

        System.exit(retCode);
    }





    private static int runIteration(Map<String, Stage> dispTable,
                                    Map<String, Integer> dupTable,
                                    Map<String, String> envs,
                                    Logger logger)
    {
        t1 = System.nanoTime();
        Stage stage = dispTable.get(envs.get("cell_stageName")); // cell_stageName should be stage ID
        List<String> inQueues = stage.getInput();
        List<YARNApplicationMaster.TezQRep> in = new ArrayList<YARNApplicationMaster.TezQRep>();

        List<String> outQueues = stage.getOutput();
        List<YARNApplicationMaster.TezQRep> out = new ArrayList<YARNApplicationMaster.TezQRep>();

        //ToDo: For all input queues, we should have done the reservation.  For now we have just one input window
        // Lets do the resrvation of window
        WindowOperationResponse response;
        response = windowClient.changeWindowState(inQueues.get(0), Long.parseLong(envs.get("replywindow_windowNo")),
                                    envs.get("replywindow_site"),
                                    WindowState.FORMED, WindowState.UNDERPROCESSING, myName);

        manageFailure(response, "Change window state from FORMED to UNDERPROCESSING", 2);


        List cmdLine = new ArrayList<String>();
        int retCode = run(cmdLine, stage, in, out);
        t2 = System.nanoTime();
        logger.info("Return code form run = " + retCode);
        logger.info("time run job = " + (t2-t1)/1000000000);

        if (retCode == 0)
        {
            // store the results first
            try {
                Map<String,Object> result = YAMLSerAndDeSer.readYAMLFileForWindowsAbstraction("./outyaml");
                if (result.containsKey("ret_code") == false || (Integer)(result.get("ret_code")) != 0)
                {
                    logger.info("YARNApplicationMasterForWindows: outyaml either does not have ret_code or ret_code not zero.");
                    return 201;
                }
                //"output_rssacint.xz": list_of_output_hdfs_files
                for (String q : outQueues) {
                    String key = "output_"+q;
                    if (result.containsKey(key) == false || !(result.get(key) instanceof List))
                    {
                        logger.info("YARNApplicationMasterForWindows: Either key: " + key + " not present in outyaml or not an instance of a List");
                        return 202;
                    }
                    List<String> hdfsFileList = (List<String>)result.get(key);
                    if (hdfsFileList.size() == 0)
                    {
                        logger.info("YARNApplicationMasterForWindows: For key: " + key + " List of hdfs files is empty.");
                        return 203;
                    }
                    String newFileName = getNewFileName(hdfsFileList.get(0), q);

                    int count = dupTable.get(q);
                    if (count <= 1) {
                        logger.info("Asking for enque URL " + q + " and newFileName = " + newFileName);
                        YARNApplicationMaster.TezQRep tqr = new YARNApplicationMaster.TezQRep(q, QueueHelper.getEnqueuURIRepeatOnError(q, newFileName, logger).reservedHDFSFile);
                        retCode = QueueHelper.enqueueFileRepeatOnError(q, tqr.hdfsURI, logger);
                        if (retCode != 0)
                        {
                            logger.info("YARNApplicationMasterForWindows: Failed to put " + tqr.hdfsURI + " into queue: " + q);
                            return 204;
                        }


                    } else {
                        logger.info("Asking for distribute URL " + q + " and newFileName = " + newFileName);
                        YARNApplicationMaster.TezQRep tqr = new YARNApplicationMaster.TezQRep(q, QueueHelper.getDistributeURIRepeatOnError(q, newFileName, logger).reservedHDFSFile);
                        retCode = QueueHelper.distributeFileRepeatOnError(q, tqr.hdfsURI, logger);
                        if (retCode != 0)
                        {
                            logger.info("YARNApplicationMasterForWindows: Failed to put " + tqr.hdfsURI + " into queue: " + q);
                            return 205;
                        }

                    }
                    // Now un-reserve windows files
                    long windowNO = Long.parseLong(envs.get("replywindow_windowNo"));
                    response = windowClient.unreserveWindow(q, windowNO);
                    if (response.getResult() == false)
                    {
                        logger.info("YARNApplicationMasterForWindows: Failed to un reserve window no: " + windowNO);
                        return 205;
                    }

                }


                // Now del hdfs files if user asked so. "hdfs_cleanup": to_del_list,
                if (result.containsKey("hdfs_cleanup") && (result.get("hdfs_cleanup") instanceof List))
                {
                    List<String> hdfsFilesToDel = (List<String>)result.get("hdfs_cleanup");
                    for (String f:hdfsFilesToDel)
                    {
                        retCode = HDFSHelper.delOneFile(f, logger);
                        logger.info("YARNApplicationMasterForWindows: Failed to del hdfs file: " + f);

                    }


                }


            } catch (IOException e) {
                e.printStackTrace();
                logger.info("YARNApplicationMasterForWindows failed due to YAML deserialization error.");
                return 200;
            }

            //Map<String,Object> YAMLSerAndDreadYAMLFileForWindowsAbstraction(String yamlFilePath)
            response = windowClient.changeWindowState(inQueues.get(0), Long.parseLong(envs.get("replywindow_windowNo")),
                    envs.get("replywindow_site"),
                    WindowState.UNDERPROCESSING, WindowState.DONE, myName);
            manageFailure(response, "Change window state from UNDERPROCESSING to DONE", 3);


        }
        else
        {
            // Plumb fault detector will detect that window state has been UNDERPROCESSING for long and will
            // re-try or switch to faulty state
        }



        return retCode;

    }

    private static void manageFailure(WindowOperationResponse response, String action, int exitCode)
    {
        System.out.println(response);
        if (response.getResult() == false)
        {
            System.err.println("Failed windowing operation while doing: " + action);
            System.err.println(response);
            System.exit(exitCode);
        }

    }

    private static int run(List<String> args,
                           Stage stage,
                           List<YARNApplicationMaster.TezQRep> in,
                           List<YARNApplicationMaster.TezQRep> out)
    {
        int retCode = 1;
        try {



            args.add(0, "./AppMaster.shellscript");




            ProcessBuilder pb = new ProcessBuilder(args)
                    .inheritIO();


            pb.redirectErrorStream(true);
            setEnv(pb.environment(), stage, in, out);

            Process p = pb.start();
            retCode = p.waitFor();
            BufferedReader reader =  new BufferedReader(new InputStreamReader(p.getInputStream()));
            String line = "";
            while ( (line = reader.readLine()) != null) {
                System.out.println(line);
            }

            System.out.println("Retval = " + retCode);
        } catch (Exception exp) {
            exp.printStackTrace();
        }

        return retCode;
    }



    private static void setEnv(Map<String, String> env,
                               Stage stage,
                               List<YARNApplicationMaster.TezQRep> in,
                               List<YARNApplicationMaster.TezQRep> out)
    {

        env.put("__plumb_no_of_inputs", in.size()+"");
        int i = 1;
        for (YARNApplicationMaster.TezQRep tqr : in)
        {
            env.put("__plumb_i"+i, tqr.hdfsURI);
            i++;
        }



        env.put("__plumb_no_of_outputs", out.size()+"");
        i = 1;
        for (YARNApplicationMaster.TezQRep tqr : out)
        {
            env.put("__plumb_o"+i, tqr.hdfsURI);
            i++;
        }

        String userProg = "";

        for (String s : stage.getProgram()) {
            userProg += s;
        }

        String parts[] = userProg.split(" ", 2);
        if (null != parts && parts.length > 0)
        {
            env.put("__plumb_prog_name", parts[0]);
        }

        if (null != parts && parts.length > 1)
        {
            env.put("__plumb_prog_args", parts[1]);
        }

        // Following two sent from here so that yarn wrapper could sym link to them without hard-coding.
        env.put("pathToRegEx", ServerConfig.getString("pathToRegEx"));
        env.put("pathToYamlUtil",ServerConfig.getString("pathToYamlUtil"));
        env.put("inyaml","./inyaml");
        env.put("outyaml","./outyaml");
        env.put("erryaml","./erryaml");



    }


}

