// Code copied from: https://github.com/alexholmes/hiped2/blob/master/src/main/java/hip/ch10/dstat/Client.java
// and then modified

package edu.isi.ant.JobAPI;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sun.xml.bind.v2.runtime.reflect.opt.Const;
import edu.isi.ant.yamlseranddeser.Stage;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.util.ClassUtil;
import org.apache.hadoop.yarn.api.ApplicationConstants;
import org.apache.hadoop.yarn.api.ApplicationConstants.Environment;
import org.apache.hadoop.yarn.api.records.*;
import org.apache.hadoop.yarn.client.api.YarnClient;
import org.apache.hadoop.yarn.client.api.YarnClientApplication;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
import org.apache.hadoop.yarn.exceptions.YarnException;
import org.apache.hadoop.yarn.util.ConverterUtils;
import org.apache.hadoop.yarn.util.Records;
import org.apache.hadoop.yarn.util.Apps;

import java.io.IOException;
import java.util.*;
import java.io.File;
import java.util.concurrent.TimeUnit;

public class YARNJobSubmitter extends JobSubmitter{

    private YarnClient yarnClient;
    private ApplicationId appId;

    public YARNJobSubmitter()
    {

    }



    public static boolean submitJob(List<String> args)
    {
        boolean result = true;
        try {
            submitJobInternal(args);
            result = false;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return result;
    }

    public static YarnConfiguration getConf()
    {
        YarnConfiguration conf = new YarnConfiguration();
        ////////////////
        ///conf.set("yarn.resourcemanager.hostname", "hcpu001.ant.isi.edu");
        ///conf.set("fs.default.name", "hdfs://hcpu001.ant.isi.edu:9000/");
        conf.set(ServerConfig.getString("yarn_resourcemanager_hostname_key"),
                 ServerConfig.getString("yarn_resourcemanager_hostname_val"));
        conf.set(ServerConfig.getString("fs_default_name_key"),
                 ServerConfig.getString("fs_default_name_val"));

        ///////////////

        return conf;

    }

    private static LocalResource getLocalResourceFile(String filename) throws Exception {
        YarnConfiguration conf = getConf();
        FileSystem fs = FileSystem.get(conf);
        Path src = new Path(filename);
        Path dest = new Path(ServerConfig.getString("HDFSScratchSpace"), src.getName());
        if (!fs.exists(dest))
            System.out.println("WARNING: file " + dest + "doesn't exist");
        FileStatus fileStat = FileSystem.get(conf).getFileStatus(dest);
        LocalResource lrFile = Records.newRecord(LocalResource.class);
        lrFile.setResource(ConverterUtils.getYarnUrlFromPath(dest));
        lrFile.setSize(fileStat.getLen());
        lrFile.setTimestamp(fileStat.getModificationTime());
        lrFile.setType(LocalResourceType.FILE);
        lrFile.setVisibility(LocalResourceVisibility.APPLICATION);

        return lrFile;
    }
    private static void submitJobInternal(List<String> args) throws Exception{
        // create yarn client
        YarnConfiguration conf = getConf();
        YarnClient yarnClient = YarnClient.createYarnClient();
        yarnClient.init(conf);
        yarnClient.start();

        // create an application
        YarnClientApplication app = yarnClient.createApplication();

        // Create context for user application submission
        ContainerLaunchContext container = Records.newRecord(ContainerLaunchContext.class);
        String amLaunchCmd;
        amLaunchCmd = String.format(
                        "$JAVA_HOME/bin/java -Xmx256M %s 1>%s/stdout 2>%s/stderr",
                        YARNApplicationMaster.class.getName(),
                        ApplicationConstants.LOG_DIR_EXPANSION_VAR,
                        ApplicationConstants.LOG_DIR_EXPANSION_VAR);

        container.setCommands(Lists.newArrayList(amLaunchCmd));


        // setup jar for AM
        String appMasterJarName = ClassUtil.findContainingJar(YARNJobSubmitter.class);
        LocalResource appMasterJar = makeLocalResource(appMasterJarName);

        //also provide shell script to Yarn.  It should have been done in Dispatcher already
        String shellScriptName = ServerConfig.getString("internalScriptsPath") + "/yarn_job_wrapper.sh";
        LocalResource shellScript = makeLocalResource(shellScriptName);

        //also provide configuration file.  It should have been done in Dispatcher already
        LocalResource confFile = makeLocalResource(ServerConfig.CONFIGURATION_FILE);

        container.setLocalResources(
            ImmutableMap.of(
                "AppMaster.jar", appMasterJar,
                "AppMaster.shellscript", shellScript,
                ServerConfig.CONFIGURATION_FILE, confFile
            )
        );

        // setup classpath for AM
        String classPathSeparator = ":";
        Map<String, String> appMasterEnv = Maps.newHashMap();
        for (String c : conf.getStrings(
                YarnConfiguration.YARN_APPLICATION_CLASSPATH,
                YarnConfiguration.DEFAULT_YARN_APPLICATION_CLASSPATH)) {
            Apps.addToEnvironment(appMasterEnv, Environment.CLASSPATH.name(),
                    c.trim(), classPathSeparator);
        }
        Apps.addToEnvironment(appMasterEnv,
                Environment.CLASSPATH.name(),
                Environment.PWD.$() + File.separator + "*", classPathSeparator);

        String prepPATH = ServerConfig.getString("environment.path.prepend", "");
        if (prepPATH != "") {
            System.out.println("prepending " + prepPATH + " to PATH env");
            Apps.setEnvFromInputString(appMasterEnv, "PATH=" + prepPATH + ":$PATH", File.pathSeparator);
        }

        //TODO: delete?
        Apps.addToEnvironment(appMasterEnv,
                Environment.HOME.name(),
                "abc_home", "");

        int noOfArgs = args.size();

        Apps.addToEnvironment(appMasterEnv,
                "args_count",
                noOfArgs+"", "");

        for(int i=1; i<=noOfArgs;i++)
        {
            Apps.addToEnvironment(appMasterEnv, "args_"+i, "\""+args.get(i-1)+"\"", "");

        }

        //XXX appMasterEnv was setup but not used going forward
        //container.setEnvironment(appMasterEnv);

        // specify resource requirements for AM
        Resource capability = Records.newRecord(Resource.class);
        capability.setMemory(2560);
        capability.setVirtualCores(1);

        // Finally, set-up ApplicationSubmissionContext for the application
        ApplicationSubmissionContext appContext = app.getApplicationSubmissionContext();
        appContext.setApplicationName("Plumb"); // application name
        appContext.setAMContainerSpec(container);
        appContext.setResource(capability);
        appContext.setQueue("default"); // queue

        // submit the application
        ApplicationId appId = appContext.getApplicationId();
        System.out.println("Submitting application " + appId);
        yarnClient.submitApplication(appContext);

        ApplicationReport report = yarnClient.getApplicationReport(appId);
        YarnApplicationState state = report.getYarnApplicationState();

        EnumSet terminalStates = EnumSet.of(YarnApplicationState.FINISHED,
                                            YarnApplicationState.KILLED,
                                            YarnApplicationState.FAILED);

        // wait for the application to complete
        Random random = new Random();
        int l = 1;
        int h = 10;
        while (!terminalStates.contains(state)) {

            TimeUnit.SECONDS.sleep(random.nextInt(h-l)+l);
            report = yarnClient.getApplicationReport(appId);
            state = report.getYarnApplicationState();
        }

        System.out.printf("Application %s finished with state %s%n", appId, state);

    }

//........................................................

    public boolean submitJobNonBlocking(List<String> args)
    {
        boolean result = false;
        try {
            submitJobNonBlockingInternal(args);
            result = true;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return result;
    }

    //TODO factor out common code with submitJobInternal
    private void submitJobNonBlockingInternal(List<String> args) throws Exception{
        // create yarn client
        YarnConfiguration conf = getConf();
        this.yarnClient = YarnClient.createYarnClient();
        yarnClient.init(conf);
        yarnClient.start();

        // create an application
        YarnClientApplication app = yarnClient.createApplication();

        // Create context for user application submission
        ContainerLaunchContext container = Records.newRecord(ContainerLaunchContext.class);
        String amLaunchCmd;
        // Java parallel garbage collector seems to spawn too many threads on servers that have
        // like 32 cores.  As per advice from https://confluence.csiro.au/pages/viewpage.action?pageId=278167841
        // I am using a serial garbage collector for now.
        amLaunchCmd = String.format(
                //"$JAVA_HOME/bin/java -Xmx4G %s 1>%s/stdout 2>%s/stderr",
                "$JAVA_HOME/bin/java -XX:+UseSerialGC %s 1>%s/stdout 2>%s/stderr",
                YARNApplicationMaster.class.getName(),
                ApplicationConstants.LOG_DIR_EXPANSION_VAR,
                ApplicationConstants.LOG_DIR_EXPANSION_VAR);

        container.setCommands(Lists.newArrayList(amLaunchCmd));


        // setup jar for AM
        String appMasterJarName = ClassUtil.findContainingJar(YARNJobSubmitter.class);
        LocalResource appMasterJar = makeLocalResource(appMasterJarName);

        // Get Stage jar
        String stageJarName = ClassUtil.findContainingJar(Stage.class);
        LocalResource stageJar = makeLocalResource(stageJarName);

        //also provide shell script to Yarn.  It should have been done in Dispatcher already
        String shellScriptName = ServerConfig.getString("internalScriptsPath") + "/yarn_job_wrapper.sh";
        LocalResource shellScript = makeLocalResource(shellScriptName);
        
        //also provide configuration file.  It should have been done in Dispatcher already
        LocalResource confFile = makeLocalResource(ServerConfig.CONFIGURATION_FILE);

        // get dup and disp packaged
        //LocalResource disp = this.makeLocalResource(Constants.dispPath, conf, fs);
        //LocalResource dup = this.makeLocalResource(Constants.dupPath, conf, fs);

        container.setLocalResources(
            ImmutableMap.of(
                "AppMaster.jar", appMasterJar,
                "Stage.jar", stageJar,
                "AppMaster.shellscript", shellScript,
                ServerConfig.CONFIGURATION_FILE, confFile
            )
        );



        // setup classpath for AM
        String classPathSeparator = ":";
        Map<String, String> appMasterEnv = Maps.newHashMap();
        for (String c : conf.getStrings(
                YarnConfiguration.YARN_APPLICATION_CLASSPATH,
                YarnConfiguration.DEFAULT_YARN_APPLICATION_CLASSPATH)) {
            Apps.addToEnvironment(appMasterEnv, Environment.CLASSPATH.name(),
                    c.trim(), classPathSeparator);
        }
        //TODO addToEnvironment is depricated switch everything to setEnvFromInputString
        Apps.addToEnvironment(appMasterEnv,
                Environment.CLASSPATH.name(),
                Environment.PWD.$() + File.separator + "*", classPathSeparator);

        String prepPATH = ServerConfig.getString("environment.path.prepend", "");
        if (prepPATH != "") {
            System.out.println("prepending " + prepPATH + " to PATH env");
            Apps.setEnvFromInputString(appMasterEnv, "PATH=" + prepPATH + ":$PATH", File.pathSeparator);
        }

        Apps.setEnvFromInputString(appMasterEnv, "HOME=/nfs/jade/hadoop_u", "");
        Apps.setEnvFromInputString(appMasterEnv, "USER=hadoop_u", "");



        int noOfArgs = args.size();

        Apps.addToEnvironment(appMasterEnv,
                "args_count",
                noOfArgs+"", "");

        for(int i=1; i<=noOfArgs;i++)
        {
            Apps.addToEnvironment(appMasterEnv, "args_"+i, "\""+args.get(i-1)+"\"", "");
        }

        container.setEnvironment(appMasterEnv);


        // specify resource requirements for AM
        Resource capability = Records.newRecord(Resource.class);
        capability.setMemory(2560);
        capability.setVirtualCores(1);

        // Finally, set-up ApplicationSubmissionContext for the application
        ApplicationSubmissionContext appContext = app.getApplicationSubmissionContext();
        String nameAppend = args.size() > 0?args.get(0):"???";
        appContext.setApplicationName("Plumb ("+nameAppend+")"); // application name
        appContext.setAMContainerSpec(container);
        appContext.setResource(capability);
        appContext.setQueue("default"); // queue




        // submit the application
        this.appId = appContext.getApplicationId();
        System.out.println("Submitting application " + appId);
        yarnClient.submitApplication(appContext);


        ApplicationReport report = yarnClient.getApplicationReport(appId);
        YarnApplicationState state = report.getYarnApplicationState();



        System.out.printf("Application %s submitted to cluster.%n", appId);

    }

    public boolean isJobFinished()
    {
        EnumSet terminalStates = EnumSet.of(YarnApplicationState.FINISHED,
                                            YarnApplicationState.KILLED,
                                            YarnApplicationState.FAILED);

        boolean result = false;

        YarnApplicationState state;
        try {
            state = yarnClient.getApplicationReport(appId).getYarnApplicationState();
            result = terminalStates.contains(state);
        } catch (ApplicationNotFoundException anf) {
            anf.printStackTrace();
            // Added this new catch clause on 2019-02-15 because if Hadoop cluster is restarted
            // we in application here still have old IDs and getting their state will get
            // obvious exception of application not found.  First bad thing we were retrying
            // something that will never succeeded and it was filling server log very quickly.

            result = true;

        }
        catch (YarnException | IOException e) {
            e.printStackTrace();
        }

        return result;

    }

    private static LocalResource makeLocalResource(String srcPath) throws Exception {
        YarnConfiguration conf = getConf();
        FileSystem fs = FileSystem.get(conf);
        Path src = new Path(srcPath);
        Path dest = new Path(ServerConfig.getString("HDFSScratchSpace"), src.getName());
        if (!fs.exists(dest))
            System.out.println("WARNING: File " + dest.toString() + " is missing");
        FileStatus fileStat = FileSystem.get(conf).getFileStatus(dest);
        LocalResource lrFile = Records.newRecord(LocalResource.class);
        lrFile.setResource(ConverterUtils.getYarnUrlFromPath(dest));
        lrFile.setSize(fileStat.getLen());
        lrFile.setTimestamp(fileStat.getModificationTime());
        lrFile.setType(LocalResourceType.FILE);
        lrFile.setVisibility(LocalResourceVisibility.APPLICATION);

        return lrFile;
    }

}
