package edu.isi.ant.JobAPI;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import edu.isi.ant.yamlseranddeser.Stage;
import edu.isi.ant.yamlseranddeser.YAMLSerAndDeSer;
import io.grpc.lander.queue.api.ReplyWindow;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.util.ClassUtil;
import org.apache.hadoop.yarn.api.ApplicationConstants;
import org.apache.hadoop.yarn.api.records.*;
import org.apache.hadoop.yarn.client.api.YarnClient;
import org.apache.hadoop.yarn.client.api.YarnClientApplication;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
import org.apache.hadoop.yarn.exceptions.YarnException;
import org.apache.hadoop.yarn.util.Apps;
import org.apache.hadoop.yarn.util.ConverterUtils;
import org.apache.hadoop.yarn.util.Records;

import java.io.File;
import java.io.IOException;
import java.util.*;


/*
 * Purpose of this class: Getting YARN's Application Master's environment ready and then to get it started
 */





public class YARNJobSubmitterForWindows extends JobSubmitter{

    private YarnClient yarnClient;
    private ApplicationId appId;

    public YARNJobSubmitterForWindows()
    {

    }


    public static YarnConfiguration getConf()
    {
        YarnConfiguration conf = new YarnConfiguration();
        conf.set(ServerConfig.getString("yarn_resourcemanager_hostname_key"), ServerConfig.getString("yarn_resourcemanager_hostname_val"));
        conf.set(ServerConfig.getString("fs_default_name_key"), ServerConfig.getString("fs_default_name_val"));

        return conf;
    }


    //ToDo: There should be as may replyWindows as there are number of inputs.  For now its just one.
    public boolean submitJobNonBlocking(LumpinessAwareFastDispatcher.StageControllerCell cell, ReplyWindow replyWindow)
    {
        boolean result = false;
        try {
            submitJobNonBlockingInternal(cell, replyWindow);
            result = true;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return result;
    }


    private void submitJobNonBlockingInternal(LumpinessAwareFastDispatcher.StageControllerCell cell,
                                              ReplyWindow replyWindow) throws Exception{

        // create yarn client
        YarnConfiguration conf = getConf();
        this.yarnClient = YarnClient.createYarnClient();
        yarnClient.init(conf);
        yarnClient.start();

        // create an application
        YarnClientApplication app = yarnClient.createApplication();


        // Create context for user application submission
        ContainerLaunchContext container = Records.newRecord(ContainerLaunchContext.class);
        String amLaunchCmd;
        // Java parallel garbage collector seems to spawn too many threads on servers that have
        // like 32 cores.  As per advice from https://confluence.csiro.au/pages/viewpage.action?pageId=278167841
        // I am using a serial garbage collector for now.
        amLaunchCmd = String.format(
                //"$JAVA_HOME/bin/java -Xmx4G %s 1>%s/stdout 2>%s/stderr",
                "$JAVA_HOME/bin/java -XX:+UseSerialGC %s %s 1>%s/stdout 2>%s/stderr",
                "-classpath \"$CLASSPATH\"",
                YARNJobSubmitterForWindows.class.getName(),
                ApplicationConstants.LOG_DIR_EXPANSION_VAR,
                ApplicationConstants.LOG_DIR_EXPANSION_VAR);

        container.setCommands(Lists.newArrayList(amLaunchCmd));


        // setup jar for AM
        LocalResource appMasterJar = MyUtils.makeLocalResource(ClassUtil.findContainingJar(YARNJobSubmitterForWindows.class), conf, false);

        // Get Stage jar
        LocalResource stageJar =  MyUtils.makeLocalResource(ClassUtil.findContainingJar(Stage.class), conf, false);

        //also provide shell script to Yarn.
        // Actual data copy now done in Dispatcher due to job failures (This code changing it while some other window job
        // is trying to read it.)
        LocalResource shellScript = MyUtils.makeLocalResource("./scripts/yarn_window_job_wrapper.sh", conf, false);

        //Convert URLs to YAML and make a file and copy it to HDFS
        Map map = this.getMapOfYAMLForUser(replyWindow);
        // ToDo: Del following file from local.  Later make sure AppMaster del it from HDFS
        // ToDo: Even after above, there are chances of such file accumulationg over time.  Need to garbage collect them.
        String outFile = String.format("%s/%s_%s_in.yaml", ServerConfig.getString("scratchSpaceForYAMLs"), cell.stageName, new Random().nextInt(Integer.MAX_VALUE));
        if (false == YAMLSerAndDeSer.writeYAMLFile(map, outFile)) throw new Exception("YAMLSerAndDeSer.writeYAMLFile(map, outFile) returned false.");
        LocalResource inYaml = MyUtils.makeLocalResource(outFile, conf, true);


        // Now set all local resources
        container.setLocalResources(ImmutableMap.of("AppMaster.jar", appMasterJar,
                "Stage.jar", stageJar,
                "AppMaster.shellscript", shellScript,
                "inyaml", inYaml)); // ToDo: This is also done in AppMaster.  Remove from here?


        // setup classpath for AM
        String classPathSeparator = ":";
        Map<String, String> appMasterEnv = Maps.newHashMap();
        // To enable AppMaster do gRPC, appropriate nfs based lib path should be on class path
        String cwd = System.getProperty("user.dir");
        // HEADS-UP: looks like there are different netty versions in first two directory.  So just using what
        // gradle has put in its build.
        //String libs[] = {cwd+"/../required_libs/*", cwd+"/../required_yarn_libs/*", cwd+"/build/install/JobServer/lib/*"};
        String libs[] = {cwd+"/build/install/JobServer/lib/*"};
        for (String c : libs) {
            Apps.addToEnvironment(appMasterEnv, ApplicationConstants.Environment.CLASSPATH.name(),
                    c.trim(), classPathSeparator);
        }


        Apps.addToEnvironment(appMasterEnv,
                ApplicationConstants.Environment.CLASSPATH.name(),
                ApplicationConstants.Environment.PWD.$() + File.separator + "*", classPathSeparator);

        // (1) Following elements of classpath are necessary. Else calls to rmClient.registerApplicationMaster() in
        // ApplicatonMaster will be stuck and ultimately AM killed by resource manager.
        // (2) Following must be the last items in classpath.  An altenate version of netty comes from here, for which
        // gRPC communciation doesn't work in WindowClient.  The netty version that gradle used, and we put prevoiusly
        // is the required on. ToDo: Fix this ordering issue.
        for (String c : conf.getStrings(
                YarnConfiguration.YARN_APPLICATION_CLASSPATH,
                YarnConfiguration.DEFAULT_YARN_APPLICATION_CLASSPATH)) {
            Apps.addToEnvironment(appMasterEnv, ApplicationConstants.Environment.CLASSPATH.name(),
                    c.trim(), classPathSeparator);
        }

        //    private void submitJobNonBlockingInternal(LumpinessAwareFastDispatcher.StageControllerCell cell,
        //                                              ReplyWindow replyWindow) throws Exception{


        Apps.addToEnvironment(appMasterEnv,"cell_stageName",cell.stageName, "");
        Apps.addToEnvironment(appMasterEnv,"cell_userProg",String.join(" ",cell.stage.getProgram()), "");
        //cell.stage.getCmdProgram(); // this gives strings like " -p /mapreduce_wrapper.sh "  We don't need -p here.

        //ToDo: Fow now following suffice for one inpup per stage window.  Later we will have at least two
        // choices: (1) Either decode the whole list of replyWindows as YAMl and then get it at AppMaster
        // (2) Plumb does some light weight query on window and e.g. pass on just window numbers or just stage ID
        // and App Master itself does the RPC call to get the actual data.  (In my view second solution is more
        // elegant.
        Apps.addToEnvironment(appMasterEnv,"replywindow_queueID",replyWindow.getQueueID(), "");
        Apps.addToEnvironment(appMasterEnv,"replywindow_windowNo",replyWindow.getWindowNo()+"", "");
        Apps.addToEnvironment(appMasterEnv,"replywindow_site",replyWindow.getQSeqNoInst()+"", "");


        Apps.addToEnvironment(appMasterEnv,"args_count",3+"", "");



        container.setEnvironment(appMasterEnv);


        // specify resource requirements for AM
        Resource capability = Records.newRecord(Resource.class);
        capability.setMemory(2560);
        capability.setVirtualCores(1);

        // Finally, set-up ApplicationSubmissionContext for the application
        ApplicationSubmissionContext appContext = app.getApplicationSubmissionContext();
        String nameAppend = cell.stageName;
        appContext.setApplicationName("Plumb-Window ("+nameAppend+")"); // application name
        appContext.setAMContainerSpec(container);
        appContext.setResource(capability);
        appContext.setQueue("default"); // queue




        // submit the application
        this.appId = appContext.getApplicationId();
        System.out.println("Submitting application " + appId);
        yarnClient.submitApplication(appContext);


        ApplicationReport report = yarnClient.getApplicationReport(appId);
        YarnApplicationState state = report.getYarnApplicationState();



        System.out.printf("Application %s submitted to cluster.%n", appId);

    }

    public boolean isJobFinished()
    {
        EnumSet terminalStates = EnumSet.of(YarnApplicationState.FINISHED,
                YarnApplicationState.KILLED,
                YarnApplicationState.FAILED);

        boolean result = false;

        YarnApplicationState state;
        try {
            state = yarnClient.getApplicationReport(appId).getYarnApplicationState();
            result = terminalStates.contains(state);
        } catch (ApplicationNotFoundException anf) {
            anf.printStackTrace();
            // Added this new catch clause on 2019-02-15 because if Hadoop cluster is restarted
            // we in application here still have old IDs and getting their state will get
            // obvious exception of application not found.  First bad thing we were retrying
            // something that will never succeeded and it was filling server log very quickly.

            result = true;

        }
        catch (YarnException | IOException e) {
            e.printStackTrace();
        }

        return result;

    }

    private Map<String, Object> getMapOfYAMLForUser(ReplyWindow replyWindow)
    {
        Map<String, Object> map = new HashMap<String, Object>();
        map.put("section_count", 1);
        map.put("input_"+replyWindow.getQueueID(), replyWindow.getListOfBlocksInWindowInThisWindowList());
        return map;
    }


}
