package edu.isi.ant.yamlseranddeser;

import java.io.Serializable;
import java.time.ZonedDateTime;
import static java.time.ZonedDateTime.now;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 * @author aqadeer
 */
public class Window implements Serializable
{
    public static enum ABSTRACTION
    {
        Large_Block_Straming,
        Record_Level_Streaming,
        Batch,
        Continuous,
        LB_WINDOW_UNORDERED_MapReduce,
        LB_CONTINUOUS_UNORDERED_Plumb,
        Unknown
    }
    
    private static Map<ABSTRACTION,Boolean> allAbstractions = new HashMap<>();
    //String to ABSTRACTION mapping
    private static Map<String,ABSTRACTION> abstractionMappingFromStringToABSTRACTION = new HashMap<>();
    static {
     
        for (ABSTRACTION anAbstraction : ABSTRACTION.values()) 
        {
            allAbstractions.put(anAbstraction, Boolean.TRUE);
            abstractionMappingFromStringToABSTRACTION.put(anAbstraction.toString(), anAbstraction);
            
        }
        // remove Unknown
        allAbstractions.remove(ABSTRACTION.Unknown);
        abstractionMappingFromStringToABSTRACTION.remove(ABSTRACTION.Unknown.toString());
        

        
    }
    public static ABSTRACTION mapStringToAbstraction(String absAsString)
    {
        ABSTRACTION x = abstractionMappingFromStringToABSTRACTION.get(absAsString);
        return ((x == null)?ABSTRACTION.Unknown:x);
    }
    
    public static enum SIZEMETRIC
    {
        Blocks,
        Time
    }
    
    public static enum TIMEMETRIC
    {
        TimeViaSizeAndStart,
        TimeViaStartAndEnd
    }
    
    public static enum MAXWAITEMETRIC
    {
        Blocks,
        Time,
        PercentBlocks
    }
    
    public static enum OUTOFORDERSCHEDULING
    {
        Yes,
        No
    }
    
    public static enum CUTOFFDELIVERYREQUIREMENT
    {
        Cutoff_Duplicate_Delivery,
        Cutoff_Exactly_Once_To_Current_Window,
        Cutoff_Exactly_Once_To_Next_Window
    }
    //String to CUTOFFDELIVERYREQUIREMENT mapping
    private static Map<String,CUTOFFDELIVERYREQUIREMENT> cutoffDeliveryMappingToString = new HashMap<>();
    static {
     
        for (CUTOFFDELIVERYREQUIREMENT cutoff : CUTOFFDELIVERYREQUIREMENT.values()) 
        {
            cutoffDeliveryMappingToString.put(cutoff.toString(), cutoff);
            
        }
    }
    
    //Heads up clients: following function can return null if unknown cutoff string provided
    public static CUTOFFDELIVERYREQUIREMENT mapStringToCutoff(String cutoffAsString)
    {
        return cutoffDeliveryMappingToString.get(cutoffAsString);
        
    }
    
    public static enum USEQUEUESEQUENCENUMBERINSTANCE
    {
        Yes,
        No
    }
    
    public ABSTRACTION abstacton = ABSTRACTION.Large_Block_Straming; // default is LBS
    public SIZEMETRIC sizeMetric = SIZEMETRIC.Blocks; // default: Window size in blocks
    public TIMEMETRIC timeMetric = TIMEMETRIC.TimeViaSizeAndStart;
    public long size; // either no of blocks or number of hours. depends on sizeMetric
    public int maxWait = 1; // this is internally always as an hour
    public MAXWAITEMETRIC maxWaitMetric = MAXWAITEMETRIC.Time;
    public OUTOFORDERSCHEDULING outOfOrderScheduling = OUTOFORDERSCHEDULING.Yes;
    public CUTOFFDELIVERYREQUIREMENT cutoffDeliveryRequirement = CUTOFFDELIVERYREQUIREMENT.Cutoff_Exactly_Once_To_Current_Window;
    // Following changed from now() to an old value
    public ZonedDateTime startOfWindow = ZonedDateTime.parse("1958 04 01 00 00 UTC", DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz"));//now();
    //public ZonedDateTime endOfWindow = Instant.ofEpochMilli(Long.MAX_VALUE).atZone(ZoneOffset.UTC);
    public ZonedDateTime endOfWindow = ZonedDateTime.parse("7000 04 01 00 00 UTC", DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz"));
    public USEQUEUESEQUENCENUMBERINSTANCE useQueueSequenceNumberInstance = USEQUEUESEQUENCENUMBERINSTANCE.No;
    // Though following is a list, but we save it as a comma separeted string (as user gave it) because later we need to put it in a hashmap.
    public String queueSequenceNumberInstancesList = new String();
    
    @Override
    public String toString()
    {
        return String.format("Abstraction=%s SizeMetric=%s Size=%s Max Wait=%s Max Wait Metric = %s Out of Order Window Scheduling=%s "
                           + " Window CutOff Method: %s Start of Window = %s End of Window = %s use_queue_sequence_number_instance = %s"
                           + " queue_sequence_number_instance_list = %s",
                             this.abstacton,this.sizeMetric,this.size,this.maxWait,
                             this.maxWaitMetric,this.outOfOrderScheduling,
                             this.cutoffDeliveryRequirement,this.startOfWindow,this.endOfWindow,
                             this.useQueueSequenceNumberInstance,
                             this.queueSequenceNumberInstancesList);
        
    }
    
    public Map<String,String> toMap()
    {
        Map<String,String> window = new HashMap<>();
        String sm = this.sizeMetric == this.sizeMetric.Time?" hour":" blocks";
        window.put("size", this.size+sm);
        window.put("start", this.startOfWindow.format(DateTimeFormatter.ofPattern("yyyy:MM:dd:HH:mm:zzz")));
        window.put("start_and_end", this.startOfWindow.format(DateTimeFormatter.ofPattern("yyyy:MM:dd:HH:mm:zzz")) +
                   " to " + 
                   this.endOfWindow.format(DateTimeFormatter.ofPattern("yyyy:MM:dd:HH:mm:zzz")));
        String wm = this.getWM();
        window.put("max_wait", this.maxWait + wm);
        window.put("across_window_ordering", this.outOfOrderScheduling.toString());
        window.put("abstraction", this.abstacton.toString());
        window.put("cutoff_delivery_requirement", this.cutoffDeliveryRequirement.toString());
        window.put("use_queue_sequence_number_instance", this.useQueueSequenceNumberInstance.toString());
        window.put("queue_sequence_number_instance_list", this.queueSequenceNumberInstancesList);
        
        return window;
    }
    
    private String getWM()
    {
        String result;
        switch(this.maxWaitMetric)
        {
            case Blocks:
                result=" Blocks";
                break;
                
            case Time:
                result=" hour";
                break;
                
            case PercentBlocks:
                result=" percent";
                break;
                
            default:
                result="";
                break;
            
        }
        return result;
    }
        
        
      
    
    
    public Window(ABSTRACTION abstraction,
                  SIZEMETRIC sizeMetric,
                  int size,
                  TIMEMETRIC timeMetric,
                  int maxWait,
                  MAXWAITEMETRIC maxWaitMetric,
                  OUTOFORDERSCHEDULING outOfOrderScheduling,
                  CUTOFFDELIVERYREQUIREMENT cutoffDeliveryRequirement,
                  ZonedDateTime startOfWindow,
                  ZonedDateTime endtOfWindow,
                  USEQUEUESEQUENCENUMBERINSTANCE useQueueSequenceNumberInstance,
                  String queue_sequence_number_instances_list)
    {
        this.abstacton = abstraction;
        this.sizeMetric = sizeMetric;
        this.size = size;
        this.timeMetric = timeMetric;
        this.maxWait = maxWait; // the metric of this can be different from sizeMetric
        this.maxWaitMetric = maxWaitMetric;
        this.outOfOrderScheduling = outOfOrderScheduling;          
        this.cutoffDeliveryRequirement = cutoffDeliveryRequirement;
        this.startOfWindow = startOfWindow;
        this.endOfWindow = endtOfWindow;
        this.useQueueSequenceNumberInstance = useQueueSequenceNumberInstance;
        this.queueSequenceNumberInstancesList = queue_sequence_number_instances_list;
        
    }  
    
    // default constructor
    public Window()
    {        
    }
    
}

