from __future__ import print_function

import grpc
import sys
import time
import threading
import os

from top.sup import hdfs, opsys

from top.sup import queue_api_pb2
from top.sup import queue_api_pb2_grpc

from top.conf import usrConfInfo

KeyPath = os.path.expanduser(usrConfInfo.KeyPath)
roots = "%s/roots.pem" % KeyPath
client_key = "%s/client.key.pem" % KeyPath
cert_chain = "%s/client_chain" % KeyPath

server_name = usrConfInfo.server_name
server_port = usrConfInfo.server_port


# server_name = "localhost6"
# server_name = "0.0.0.0"

def e(ret):
    exit(ret[0])


_DEBUG_THIS = True


def _r(response):
    ret = (response.retCode, response.result, response.errorMessage)
    if _DEBUG_THIS:
        output2 = ("retCode=%r result= %d errMessage = %s" % ret)
        sys.stderr.write(output2 + "\n")
    return ret


#####  RPC diagnostic functions.  Don't remove. #######
def run_plain():
    channel = grpc.insecure_channel(server_name + ':' + server_port)
    stub = queue_api_pb2_grpc.LanderQueueAPIImplementationStub(channel)
    response = stub.SayHello(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)
    response = stub.SayHelloAgain(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)


def run_onewayAuthentication():
    creds = grpc.ssl_channel_credentials(open(roots).read())
    channel = grpc.secure_channel(server_name + ':' + server_port, creds)
    stub = queue_api_pb2_grpc.LanderQueueAPIImplementationStub(channel)
    response = stub.SayHello(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)
    response = stub.SayHelloAgain(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)


def run_twowayAuthentication():
    creds = grpc.ssl_channel_credentials(open(roots).read(), open(client_key).read(), open(cert_chain).read())
    channel = grpc.secure_channel(server_name + ':' + server_port, creds)
    stub = queue_api_pb2_grpc.LanderQueueAPIImplementationStub(channel)
    response = stub.SayHello(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)
    response = stub.SayHelloAgain(queue_api_pb2.HelloRequest(name='you'))
    print("LanderQueueAPIImplementation client received: " + response.message)


def run_twowayStreaming():
    creds = grpc.ssl_channel_credentials(open(roots).read(), open(client_key).read(), open(cert_chain).read())
    channel = grpc.secure_channel(server_name + ':' + server_port, creds)
    stub = queue_api_pb2_grpc.LanderQueueAPIImplementationStub(channel)

    mygenerator = createGenerator()  # create a generator
    for response in stub.SayTwoWayStreamingHello(mygenerator):
        print("LanderQueueAPIImplementation client received: " + response.message)


########################################################################


def createGenerator():
    rList = [queue_api_pb2.HelloRequest(name='request 1'), queue_api_pb2.HelloRequest(name='request 2')]
    for request in rList:
        yield request


def getStub():
    creds = grpc.ssl_channel_credentials(open(roots, 'rb').read(), open(client_key, 'rb').read(),
                                         open(cert_chain, 'rb').read())
    channel = grpc.secure_channel(server_name + ':' + server_port, creds)
    stub = queue_api_pb2_grpc.LanderQueueAPIImplementationStub(channel)
    return stub


def listDerivedQueues(from_queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc, caller_id=caller_id_fc)
    response = stub.listDerivedQueues(request)
    for q in response.listOfDerivedQueues:
        print(q)
    #
    return _r(response)


def createQueue(queue_id_fc, data_type_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        data_type=data_type_fc,
                                        caller_id=caller_id_fc)
    response = stub.createQueue(request)
    # print (response.retCode)
    #
    return _r(response)


def deleteQueue(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc
                                        )
    response = stub.deleteQueue(request)
    # print (response.retCode)
    #
    return _r(response)


def connectDerivedQueue(from_queue_id_fc, to_queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                        to_queue_id=to_queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.connectDerivedQueue(request)
    # print (response.retCode)
    #
    return _r(response)


def activateQueue(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.activateQueue(request)
    # print(response.retCode)
    #
    return _r(response)


def deactivateQueue(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.deactivateQueue(request)
    # print(response.retCode)
    #
    return _r(response)


def deactivateAllQueues(caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(caller_id=caller_id_fc)
    response = stub.deactivateAllQueues(request)
    # print(response.retCode)
    #
    return _r(response)


# Following rpc call does the work in two-passes
# Pass-1: Server does some verification and returns an appropriate HDFS URI
# Client: Copies files at the provided URI
# Pass-2: Server verifies file is copied at URI and does other DB related house-keeping
# Mechanism: We use 2-way streaming facility of gRPC
def enqueueLocal(to_queue_id_fc, localfile_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)  # we want to generate 2 requests
    rg.insertRequest(
        queue_api_pb2.QAPIRequest(to_queue_id=to_queue_id_fc, localfile=localfile_fc, caller_id=caller_id_fc))
    phase = 1
    for response in stub.enqueueLocal(rg.generateRequest()):
        phase += 1
        if response.result == False:
            break
        if phase == 2:
            r = hdfs.copyFileFromLocal(localfile_fc, response.hdfs_file_uri, True)
            if r != 0:
                response.retCode = r
                response.result = False
                response.errorMessage = "Local client box: hdfs.copyFileFromLocal failed in %s" % __name__
                break
            rg.insertRequest(queue_api_pb2.QAPIRequest(to_queue_id=to_queue_id_fc,
                                                       localfile=localfile_fc,
                                                       hdfs_file_uri=response.hdfs_file_uri,
                                                       caller_id=caller_id_fc))
            # xxx error status?
    rg.stop()
    return _r(response)


# Following rpc call does the work in two-passes
# Pass-1: Server does some verification and returns an appropriate HDFS URI
# Client: Copies files at the provided URI
# Pass-2: Server verifies file is copied at URI and does other DB related house-keeping
# Mechanism: We use 2-way streaming facility of gRPC
def distributeLocal(from_queue_id_fc, localfile_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)  # we want to generate 2 requests
    rg.insertRequest(queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                               localfile=localfile_fc,
                                               caller_id=caller_id_fc))
    phase = 1
    for response in stub.distributeLocal(rg.generateRequest()):
        phase += 1
        if response.result == False:
            break

        if phase == 2:
            r = hdfs.copyFileFromLocal(localfile_fc, response.hdfs_file_uri, True)
            if r != 0:
                response.retCode = r
                response.result = False
                response.errorMessage = "Local client box: hdfs.copyFileFromLocal failed in %s" % __name__
                break
            rg.insertRequest(queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                                       localfile=localfile_fc,
                                                       hdfs_file_uri=response.hdfs_file_uri,
                                                       caller_id=caller_id_fc))
            # xxx error status?
    rg.stop()
    return _r(response)


def lander_queue_size(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.lander_queue_size(request)
    print(response.qSize)
    #
    return _r(response)


def lander_queue_files(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.lander_queue_files(request)
    for f in response.listOfFilesInQueue:
        print(f)
    #
    return _r(response)


def lander_queue_file_older(queue_id_fc,
                            hdfs_file_uri_fc,
                            ageseconds_fc,
                            caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        hdfs_file_uri=hdfs_file_uri_fc,
                                        ageseconds=int(ageseconds_fc),
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_file_older(request)
    # print (response.retCode)
    #
    return _r(response)


def lander_queue_reserve_top(queue_id_fc, localtmp_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        localtmp=localtmp_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_reserve_top(request)
    finalTmpFile = localtmp_fc + "/" + os.path.basename(response.hdfs_file_uri)

    if response.result == True:
        opsys.__deleteFile(finalTmpFile)  # old file if it already exists
        response.retCode = hdfs.copyToLocal(finalTmpFile, response.hdfs_file_uri)
        if response.retCode != 0:
            response.errorMessage = ("%s reserved in DB but HDFS to local copy failed."
                                     % response.hdfs_file_uri)
    if response.retCode == 0:
        print(response.hdfs_file_uri)
    #
    return _r(response)


def lander_queue_reserve_top_without_local_copy(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.lander_queue_reserve_top(request)
    print(response.hdfs_file_uri)
    #
    return _r(response)


def lander_queue_reserve_file(queue_id_fc, localtmp_fc, file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        localtmp=localtmp_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_reserve_file(request)
    finalTmpFile = localtmp_fc + "/" + os.path.basename(file_uri_fc)

    if response.result == True:
        opsys.__deleteFile(finalTmpFile)  # old file if it already exists
        response.retCode = hdfs.copyToLocal(finalTmpFile, file_uri_fc)
        if response.retCode != 0:
            response.errorMessage = "%s reserved in DB but HDFS to local copy failed." % file_uri_fc
        else:
            print(finalTmpFile)
    return _r(response)


def lander_queue_unreserve(queue_id_fc, file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_unreserve(request)
    return _r(response)


def lander_queue_release(queue_id_fc, file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_release(request)

    return _r(response)


def lander_queue_release_unconditional(queue_id_fc, file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_release_unconditional(request)

    return _r(response)


def lander_queue_wipe(queue_id_fc,
                      caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_wipe(request)
    # print (response.retCode)
    return _r(response)


def markAsFaulty(queue_id_fc,
                 file_uri_fc,
                 caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.markAsFaulty(request)
    # print (response.retCode)

    return _r(response)


def markAsHealthy(queue_id_fc,
                  file_uri_fc,
                  caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_uri=file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.markAsHealthy(request)
    # print (response.retCode)
    return _r(response)


def listFilesInStateX(queue_id_fc, file_state_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        file_state=file_state_fc,
                                        caller_id=caller_id_fc)
    response = stub.listFilesInStateX(request)
    # print (response.retCode)
    for f in response.listOfFilesInQueue:
        print(f)
    return _r(response)


def lander_queue_list(queue_id_fc, data_type_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        data_type=data_type_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_list(request)
    for q in response.listOfQueues:
        print(q)
    return _r(response)


def lander_queue_du(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_du(request)
    #
    print("%s" % (str(response.queueSizeInBytes)))
    #
    return _r(response)


def lander_queue_user_du(caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(caller_id=caller_id_fc)
    response = stub.lander_queue_user_du(request)
    #
    print("%s" % (str(response.userQueuesSizeInBytes)))
    #
    return _r(response)


def lander_queue_queue2user(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_queue2user(request)
    #
    print(response.queueOwner)
    #
    return _r(response)


def lander_queue_user2queues(caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(caller_id=caller_id_fc)
    response = stub.lander_queue_user2queues(request)
    # print(response.retCode)
    #
    for q in response.listOfQueues:
        print(q)
    #
    return _r(response)


def getEnqueueURI(to_queue_id_fc, localfile_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)
    rg.insertRequest(
        queue_api_pb2.QAPIRequest(to_queue_id=to_queue_id_fc, localfile=localfile_fc, caller_id=caller_id_fc))

    for response in stub.enqueueLocal(rg.generateRequest()):
        rg.stop()
        break
    print(response.hdfs_file_uri)  # Plumb needs this value on stdout
    return _r(response)


def enqueue(to_queue_id_fc, hdfs_uri_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)  # we want to generate 2 requests
    rg.insertRequest(
        queue_api_pb2.QAPIRequest(to_queue_id=to_queue_id_fc, localfile=hdfs_uri_fc, caller_id=caller_id_fc))

    phase = 1
    for response in stub.enqueueLocal(rg.generateRequest()):
        phase += 1
        if response.result == False:
            break

        if phase == 2:
            rg.insertRequest(queue_api_pb2.QAPIRequest(to_queue_id=to_queue_id_fc,
                                                       localfile=hdfs_uri_fc,
                                                       hdfs_file_uri=response.hdfs_file_uri,
                                                       caller_id=caller_id_fc))
    rg.stop()
    # print (response.retCode)
    #
    return _r(response)


def getDistributeURI(from_queue_id_fc, localfile_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)  # we want to generate 2 requests
    rg.insertRequest(queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                               localfile=localfile_fc,
                                               caller_id=caller_id_fc))

    for response in stub.distributeLocal(rg.generateRequest()):
        rg.stop()
        break
    print(response.hdfs_file_uri)  # Plumb needs this value on stdout
    #
    return _r(response)


def distribute(from_queue_id_fc, hdfs_file_uri_fc, caller_id_fc=None):
    stub = getStub()
    rList = []
    rg = RequestGenerator(rList, 2)  # we want to generate 2 requests
    rg.insertRequest(queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                               localfile=hdfs_file_uri_fc,
                                               caller_id=caller_id_fc))
    phase = 1
    for response in stub.distributeLocal(rg.generateRequest()):
        phase += 1
        if response.result == False:
            rg.stop()
            break

        if phase == 2:
            rg.insertRequest(queue_api_pb2.QAPIRequest(from_queue_id=from_queue_id_fc,
                                                       localfile=hdfs_file_uri_fc,
                                                       hdfs_file_uri=response.hdfs_file_uri,
                                                       caller_id=caller_id_fc))
    # print (response.retCode)
    #
    return _r(response)


def copyFromLocalToHDFS(localfile_uri_fc, hdfs_uri_fc, caller_id_fc=None):
    response = queue_api_pb2.QAPIReply(retCode=0, result=False, errorMessage="OK")
    response.retCode = hdfs.copyFileFromLocal(localfile_uri_fc, hdfs_uri_fc, True)
    if response.retCode != 0:
        response.result = False
        response.errorMessage = "Local client box: hdfs.copyFileFromLocal failed in %s" % __name__

    # print (response.retCode)
    #
    return _r(response)


def doErrorRecovery(queue_id_fc, how_old_fc, max_fault_count_fc, is_window_fc='no', caller_id_fc=None):
    is_window_boolean = False
    if is_window_fc.lower() == 'yes':
        is_window_boolean = True
    stub = getStub()
    # fc means from call
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        how_old=int(how_old_fc),
                                        max_fault_count=int(max_fault_count_fc),
                                        caller_id=caller_id_fc,
                                        is_window=is_window_boolean
                                        )
    response = stub.doErrorRecovery(request)
    print(response.fileCount)

    output2 = "result=%r errCode= %d errMessage = %s" % (response.result, response.retCode, response.errorMessage)
    sys.stderr.write(output2 + "\n")


def lander_queue_peek_next(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_peek_next(request)

    if response.retCode == 0:
        print("%s %s %s %s" % (str(response.count), str(response.gap), str(response.minutesSinceLastReserveNext),
                               str(response.hdfs_file_uri)))

    return _r(response)


def lander_queue_reserve_next(queue_id_fc, hdfs_file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc,
                                        hdfs_file_uri=hdfs_file_uri_fc,
                                        caller_id=caller_id_fc)
    response = stub.lander_queue_reserve_next(request)
    return _r(response)


def lander_queue_list_keys(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.lander_queue_list_keys(request)
    for f in response.listOfKeysInKeyVal:
        print(f)
    #
    return _r(response)


def get_value(queue_id_fc, key_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, key=key_fc, caller_id=caller_id_fc)
    response = stub.get_value(request)
    print(response.value)
    #
    return _r(response)


def put_value(queue_id_fc, key_fc, value_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, key=key_fc, value=value_fc, caller_id=caller_id_fc)
    response = stub.put_value(request)
    return _r(response)


def del_value(queue_id_fc, key_fc, value_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.QAPIRequest(queue_id=queue_id_fc, key=key_fc, value=value_fc, caller_id=caller_id_fc)
    response = stub.del_value(request)
    return _r(response)


# for windows
def ls_windowed_queues(caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(caller_id=caller_id_fc)
    response = stub.ls_windowed_queues(request)
    for f in response.listOfQueues:
        print(f)
    #
    return _r(response)


def ls_all_window_statuses(caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(caller_id=caller_id_fc)
    response = stub.ls_all_window_statuses(request)
    for f in response.listOfStatuses:
        print(f)
    #
    return _r(response)


def ls_win(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.ls_win(request)
    for f in response.listOfStatuses:
        print(f)
    #
    return _r(response)


def ls_win_in_status(queue_id_fc, status_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, status=status_fc, caller_id=caller_id_fc)
    response = stub.ls_win_in_status(request)
    for f in response.listOfWindows:
        print(f)
    #
    return _r(response)


def set_win_status(queue_id_fc, window_no_fc, status_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, window_no=window_no_fc, status=status_fc
                                          , caller_id=caller_id_fc)
    response = stub.set_win_status(request)
    return _r(response)


def ls_f_in_win(queue_id_fc, window_no_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, window_no=window_no_fc, caller_id=caller_id_fc)
    response = stub.ls_f_in_win(request)
    for f in response.listOfFiles:
        print(f)
    #
    return _r(response)


def mv_f_in_win(queue_id_fc, window_no_fc, file_uri_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, window_no=window_no_fc, file_uri=file_uri_fc
                                          , caller_id=caller_id_fc)
    response = stub.mv_f_in_win(request)
    return _r(response)


def ls_missing_files_in_a_window(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.ls_missing_files_in_a_window(request)
    for f in response.listOfFiles:
        print(f)
    #
    return _r(response)


def ls_site_info_in_a_window(queue_id_fc, caller_id_fc=None):
    stub = getStub()
    request = queue_api_pb2.WindowRequest(queue_id=queue_id_fc, caller_id=caller_id_fc)
    response = stub.ls_site_info_in_a_window(request)
    for f in response.siteInfoList:
        print(f)
    #
    return _r(response)


class RequestGenerator:
    # constructor
    def __init__(self, rList, howMany):
        self.lock = threading.Lock()
        self.rList = rList
        self.howMany = howMany
        self.cv = threading.Condition(self.lock)
        self.halt = False

    def stop(self):
        self.cv.acquire()
        self.halt = True
        self.cv.notify()
        self.cv.release()

    def insertRequest(self, request):
        self.cv.acquire()
        self.rList.append(request)
        self.cv.notify()
        self.cv.release()

    def generateRequest(self):
        generated = 0
        request = None

        while generated < self.howMany:
            self.cv.acquire()
            while self.halt == False and len(self.rList) < generated + 1:
                self.cv.wait(5)

            if self.halt == True:
                return

            request = self.rList[generated]
            generated += 1
            self.cv.release()

            yield request


if __name__ == '__main__':
    # run()  # for plain data
    # run_onewayAuthentication()  # for one way
    # run_twowayAuthentication()   # for two way
    run_twowayStreaming()
    # listDerivedQueues()
    # createQueue()
    # deleteQueue()
    # connectDerivedQueue()
    # activateQueue()
    # deactivateQueue()
    # enqueue()
    # distributeFromLocal()
    # lander_queue_size()
    # lander_queue_files()
    # lander_queue_file_older()
    # lander_queue_reserve_top()
    # lander_queue_reserve_file()
    # lander_queue_unreserve()
    # lander_queue_release()
    # lander_queue_wipe()
    # doErrorRecovery("rsscint", 30, 2, "aqadeer")
