#!/usr/bin/env python3

import random
import sys

from top.sup import opsys, db
from top.conf import sysConfInfo, syssetup




Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB

debug = False


def connectDerivedQueue_Inner(from_queue_id, to_queue_id, queue_owner):
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquistion failed inside %s" %(__name__)


    if db.__doesQueueExists(dbcon, cur, from_queue_id) != 0:
        return 3, "source queue with id %s doesn't exist. This failure is in %s" %(from_queue_id, __name__)


    if db.__doesQueueExists(dbcon, cur, to_queue_id) != 0:
        return 4, "destination queue with id %s doesn't exist. This failure is in %s" %(to_queue_id, __name__)


    # Confim queue_owner is truely owner of destination queue (i.e. to_queue_id)
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, to_queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 5, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, to_queue_id, __name__)


    # Verify queue_owner is member of Linux group associated with data type of source queue (i.e. from_queue_id)
    retCode, dataType = db. __getDataTypeFromQueueTable(dbcon, cur, from_queue_id)
    if retCode != 0: return 6, "failed to get data type associated with source queue %s. This failure is in %s" %(from_queue_id, __name__)


    validy, groupName = db.__getValFromKeyValTable(dbcon, cur, dataType)
    if validy != 0: return 7, "data type %s is not associated with any Linux group. This failure is in %s" %(dataType, __name__)


    if opsys.__doesUserBelongToGroup(queue_owner, groupName) != 0:
        return 8, "user %s does not belong to group %s.  This failure is in %s" %(queue_owner, groupName, __name__)


    if 0 != db.__UpdateQueueParentsTable(dbcon, cur, from_queue_id, to_queue_id):
        return 0, "source queue %s is already connected to destination queue %s.  This failure is in %s" %(from_queue_id, to_queue_id, __name__)
    	

    cur.close()
    db.__disconnectFromDB(dbcon)

    
    return 0, "OK"
#..........................................................................

def connectDerivedQueue(from_queue_id, to_queue_id, queue_owner):
    retVal, retMessage = connectDerivedQueue_Inner(from_queue_id, to_queue_id, queue_owner)
    print (retVal)
    print (retMessage)
    return retVal, retMessage
#..........................................................................
if __name__ == '__main__':
    from_queue_id = "queue_2"
    to_queue_id  = "queue_3"
    queue_owner = "abc"

    connectDerivedQueue(from_queue_id, to_queue_id, queue_owner)
    connectDerivedQueue("queue_1", "queue_which_dont_exist", "myself")
