#!/usr/bin/env python3

import MySQLdb

from top.conf import sysConfInfo
from top.sup import db

Host = sysConfInfo.Host
User = sysConfInfo.User
Pass = sysConfInfo.Pass
DB = sysConfInfo.DB

debug = True


def create_window_state_inner(queue_id):
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon is None:
        return False, 1, "dbcon failed inside %s" % __name__
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return False, 2, "cursor acquisition failed inside %s. Error is: str(%s)" % __name__, str(e)

    dbcon.autocommit(False)

    if db.__doesQueueExists(dbcon, cur, queue_id) != 0:
        return False, 3, "queue with id %s does not exists. This failure is in %s" % (queue_id, __name__)

    ret_code, q_owner = db.__getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 4, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)

    ret_code = db.__createSubscriberWindowStateTable(dbcon, cur, queue_id, q_owner, True)
    if ret_code != 0:
        return False, 5, "db.__createSubscriberWindowStateTable failed. This failure is in %s" % __name__

    ret_code = db.__createSubscriberMissingBlockTable(dbcon, cur, queue_id, q_owner, True)
    if ret_code != 0:
        return False, 6, "db.__createSubscriberMissingBlockTable failed. This failure is in %s" % __name__

    ret_code = db.__createSubscriberSiteInfoTable(dbcon, cur, queue_id, q_owner, True)
    if ret_code != 0:
        return False, 7, "db.__createSubscriberSiteInfoTable failed. This failure is in %s" % __name__

    # if everything went well only now is the time to commit
    dbcon.commit()
    cur.close()
    db.__disconnectFromDB(dbcon)

    return True, 0, "OK"


# ..........................................................................


def create_window_state(queue_id):
    ret_val, ret_code, ret_message = create_window_state_inner(queue_id)
    print(ret_val)
    print(ret_message)
    return ret_val, ret_code, ret_message


# ..........................................................................


if __name__ == '__main__':
    queue_identification = "queue_2"
    create_window_state(queue_identification)
