#!/usr/bin/env python3

import random
import sys

from top.sup import opsys, db, hdfs
from top.conf import sysConfInfo, syssetup



# createQueue() function is meant to be run by system admin

Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = False

# 1. Tables involed are: Queue, queuename_Queue, queuename_KeyVal, Data
# 2. Take exclusive locks on tables: queuename_Queue, Queue, queuename_KeyVal
# 3. If any file in the queue is in reserved mode, queue cannot be deleted
# 4. For each data file in queue, try to delete it from Data table.  If exception, move on. Else del actual file from hdfs
# 5. Change from_que to null, where this queue was the source
# 6. Delete qeue from the Queue table
# 7. Delete queuename_Queue and queuename_KeyVal table;

def deleteQueue_Inner(queue_id, queue_owner):
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquistion failed inside %s" %(__name__)
        
    dbcon.autocommit(False)

    # Confirm queue exists
    if db.__doesQueueExists(dbcon, cur, queue_id) != 0:
        return 3, "queue with id %s doesn't exist. This failure is in %s" %(queue_id, __name__)


    # Confim queue_owner is truely owner of queue_id
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 4, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, queue_id, __name__)


    # ToDo: System is asking to lock a lot of tables.  Similar mysql.user table is hard to get locked.
    # Need to lock table in one go
    queueName = db.__getDBLevelUserQueueTableName(queue_owner, queue_id)
    subscriberKeyValTable = db.__getDBLevelUserKeyValueTableName(queue_owner, queue_id)
    ##tables = {queueName:"WRITE", subscriberKeyValTable:"WRITE", "Queue":"WRITE", "Data":"WRITE", "user":"WRITE"}
    ##isSuccess = db.__lockTables(dbcon, cur, tables)
    ##if isSuccess != 0:
    ##    errMessage = "Failed to lock tables.\n"
    ##    sys.stderr.write(errMessage)
    ##    return 4


    # Iterate in user queue to find list of file
    isSuccess, files = db.__getFileListFromSubscriberQueue(dbcon, cur, queue_owner, queue_id);
    if isSuccess != 0:
        return 6, "failed to get list of files from user queue with id %s. This failure is in %s" %(queue_id, __name__)


    for f in files:
        # del file
        if db.__deleteFileFromQueueUnConditional(dbcon, cur, queue_owner, queue_id, f) != 0: 
            return 7, "db.__deleteFileFromQueueUnConditional failed. This failure is in %s" %(__name__)
        # try deleting from Data table in best effort way
        if 0 == db.__deleteRowFromDataTable(dbcon, cur, f, False):  hdfs.delHDFSFile(f)





    # dis-connect this queue from anyone else
    if db.__DisconnectUserQueue(dbcon, cur, queue_id) != 0: return 8, "db.__DisconnectUserQueue failed. This failure is in %s" %(__name__)
    # delete queue from Queue table
    if db.__deleteRowFromQueueTable(dbcon, cur, queue_id, False) != 0: return 9, "db.__deleteRowFromQueueTable failed. This failure is in %s" %(__name__)


    # drop user queue and keyval tables
    if db.__deleteDBTable(dbcon, cur, queueName, False) !=0: return 10, "db.__deleteDBTable failed for table: %s. This failure is in %s" %(queueName, __name__)
    if db.__deleteDBTable(dbcon, cur, subscriberKeyValTable, False) !=0: 
        return 11, "db.__deleteDBTable failed for table %s. This failure is in %s" %(subscriberKeyValTable, __name__)





    dbcon.commit()

    #unlock tables (which don't exist now?)
    ##db.__ublockTables(dbcon, cur)

    cur.close()
    db.__disconnectFromDB(dbcon)

    
    return 0, "OK"

#..........................................................................
def deleteQueue(queue_id, queue_owner):
    retVal, retMessage = deleteQueue_Inner(queue_id, queue_owner)
    print (retVal)
    print (retMessage)
    return retVal, retMessage

#..........................................................................
if __name__ == '__main__':
    queue_id = "queue_2"
    queue_owner = "abc"
    deleteQueue(queue_id, queue_owner)


