#!/usr/bin/env python3

import random
import sys

from top.sup import opsys, db
from top.conf import sysConfInfo


Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = False

def doErrorRecovery_Inner(queue_id, how_old, max_fault_count, queue_owner, is_window):
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, 0, "dbcon failed inside %s" %(__name__)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, 0, "cursor acquistion failed inside %s" %(__name__)

    if db.__doesQueueExists(dbcon, cur, queue_id) != 0:
        return 3, 0, "queue with id %s doesn't exist. This failure is in %s" %(queue_id, __name__)


    # Confim queue_owner is truely owner of queue
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 4, 0, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, queue_id, __name__)

    if is_window:
        retCode, message = db.change_windows_state_conditional(dbcon, cur, queue_id, queue_owner, how_old, max_fault_count, True)
    else:
        # ToDo: Aren't following two db calls race condition?
        # Toggle from RESERVED to UNRESERVED if its been too late in that state
        retCode1, result = db.__toggleStateConditional(dbcon, cur, queue_id, queue_owner, how_old, max_fault_count, True)

        # Mark as FAULTY if too may tries have been already done
        retCode2, result2 = db.__markFaultyConditional(dbcon, cur, queue_id, queue_owner, how_old, max_fault_count, True)

        retCode = retCode1 + retCode2
    
    
    if retCode != 0:
        return 5, 0, "Eithr of db.__toggleStateConditional, db.__markFaultyConditional or \
                      db.change_windows_state_conditional failed. This failure is in %s" %(__name__)

    result = int(result) + int(result2)

    cur.close()
    db.__disconnectFromDB(dbcon)

    return retCode, result, "OK"

#..........................................................................
def doErrorRecovery(queue_id, how_old, max_fault_count, queue_owner, is_window):
    retCode, retVal, retMessage = doErrorRecovery_Inner(queue_id, how_old, max_fault_count, queue_owner, is_window)
    print (retCode)
    print (retVal)
    print (retMessage)
    return retCode, retVal, retMessage

#..........................................................................
if __name__ == '__main__':
    queue_id = "rsscint"
    how_old = 30
    max_fault_count = 2
    queue_owner = "aqadeer"

    doErrorRecovery(queue_id, how_old, max_fault_count, queue_owner)
