import datetime
import sys
from datetime import datetime, timedelta
from enum import Enum

import MySQLdb
# ToDo: Needed the following import to access WindowManager class (that should have been accessible since it is in this
#  same file).  The whole import mechanism is busted and needs a thorough review.
from top import queue_admin
from top.sup import queue_api_pb2
from queue_api_pb2 import ReplyWindow, WindowOperationRequest
from top.conf import sysConfInfo
from top.queue_admin.createWindowState import create_window_state
from top.sup import db
from top.sup import regex

Host = sysConfInfo.Host
User = sysConfInfo.User
Pass = sysConfInfo.Pass
DB = sysConfInfo.DB

debug = True

'''
# Heads-up
# If you try to call any method that starts with a __ from this class methods, python will add class name to it before
# calling and it will fail as follows:
# Exception calling application: module 'top.sup.db' has no attribute '_WindowManager__connectToDB'
# So don't do it.  I had this trouble with my db functions.
'''


class WindowManager:
    listOfWindows = None
    windowMap = {}  # key=queue_id and value=gRPC Window object

    def __init__(self, list_of_window_objects):
        result, ret_code, err_message = self.set_state(list_of_window_objects)
        if not result:
            raise ValueError(err_message)

    def set_state(self, list_of_window_objects):
        # sanity checks
        if list_of_window_objects is None:
            return False, 1, "list_of_window_objects is null.  This error occurred in: %s" % __name__
        for window in list_of_window_objects:
            if window is None:
                return False, 2, "A window object is null.  This error occurred in: %s" % __name__
            else:  # generate DB level WindowState tables
                r, c, m = create_window_state(window.queueID)  # ToDo: potential issue in MissingBlockTable
                if not r:
                    return r, c, m
                self.windowMap[window.queueID] = window

        self.listOfWindows = list_of_window_objects
        return True, 0, "OK"

    def get_windows_to_schedule(self, queue_id):
        list_of_reply_windows = []
        if (self.windowMap is None) or (queue_id not in self.windowMap):
            return False, queue_api_pb2.WindowReply.ErrorType.UNINITIALIZED_STATE, \
                   "Can not find %s in my state. This error is in %s." % (queue_id, __name__), list_of_reply_windows

        window = self.windowMap[queue_id]
        dbcon = db.connectToDB(Host, User, Pass, DB)
        print(window)
        print(window.size)
        print(window.startTime)
        # ToDo: Do we need to disable auto db commit?
        if dbcon is None:
            return False, 1, "dbcon failed inside %s" % __name__, list_of_reply_windows
        try:
            cur = dbcon.cursor()
        except MySQLdb.Error as e:
            return False, 2, "cursor acquisition failed inside %s. Error is: str(%s)" % (__name__, str(e)), \
                   list_of_reply_windows

        r, c, msg = self.process_out_of_window_blocks(dbcon, cur, queue_id, window.startTime, window.endTime)
        self.print_rcm(r, c, msg)
        if not r:
            return r, c, "%s. This error is in: %s" % (msg, __name__), list_of_reply_windows

        # Find any un-discovered sites.  See note 1.
        r, c, msg = db.fill_site_info_table(dbcon, cur, queue_id, window.useQSeqNoInst, window.qSeqNoInstList)
        self.print_rcm(r, c, msg)
        if not r:
            return r, c, "%s. This error is in: %s" % (msg, __name__), list_of_reply_windows

        # find all new unique windows, that are not already present in WindowState. See note 2.
        r, c, msg = self.process_new_blocks(dbcon, cur, queue_id, window)
        if not r:
            return r, c, "%s. This error is in: %s" % (msg, __name__), list_of_reply_windows
        ignore_site = False
        if window.useQSeqNoInst.lower() == "no": ignore_site = True
        result, count_or_code, list_of_windows_or_err_message = db.get_new_window_numbers(dbcon, cur,
                                                                                          queue_id, ignore_site)
        if not result:
            return result, count_or_code, "%s. This error is in: %s" % (list_of_windows_or_err_message, __name__), \
                   list_of_reply_windows

        # now insert all newly found windows in WindowState
        # Be careful when we need ignore site. site will be NULL or None then
        for aWindow, site in list_of_windows_or_err_message:
            # 1 in the following means 1 window across
            r, c, m = self.populate_new_window(dbcon, cur, queue_id, aWindow, site, window, ignore_site)
            if not r:
                return r, c, "%s. This error is in: %s" % (m, __name__), list_of_reply_windows

        # Mark those blocks "INWINDOW" that have just been recorded in some window
        r, c, m = db.mark_blocks_as_inwindow(dbcon, cur, queue_id)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__), list_of_reply_windows

        # Find windows that should be scheduled now.  See note 3
        # ToDo: Verify following works fine both with and without site
        r, c, complete_windows, missing_data_windows, inconclusive_windows, window_size_metric = \
            self.get_windows_whose_time_has_come(dbcon, cur, queue_id, window)
        if not r:
            return r, c, "Failed to get schedule-able windows. This error is in: %s" % __name__, list_of_reply_windows
        '''
        Each entry in above list is either of the following two tuples:
        time-based:  (winNo, site, s, e, elapsed_hours)
        block based: (window_no, site, min_block, max_block, count)
        '''

        # ToDo: For missing and in-conclusive, set current time, only if it was originally null.
        # ToDo: default that to null in window table def.  Currently its not.
        r, c, msg, ready_to_trigger_from_inconclusive = self.manage_inconclusive_windows(dbcon, cur, queue_id,
                                                                                         inconclusive_windows, window,
                                                                                         window_size_metric,
                                                                                         ignore_site)

        r, c, msg, ready_to_trigger_from_missing = self.manage_missing_windows(dbcon, cur, queue_id,
                                                                               missing_data_windows, window,
                                                                               window_size_metric, ignore_site)
        # Get a final list of ready to trigger windows.  If more than one, see user scheduling order request and
        # move on with just the oldest one.
        accumulated_windows_ready_to_trigger = []
        for window_no, site, *_ in complete_windows:
            accumulated_windows_ready_to_trigger.append((window_no, site))

        for window_no, site in ready_to_trigger_from_inconclusive:
            accumulated_windows_ready_to_trigger.append((window_no, site))

        for window_no, site in ready_to_trigger_from_missing:
            accumulated_windows_ready_to_trigger.append((window_no, site))

        accumulated_windows_ready_to_trigger.sort()
        print(accumulated_windows_ready_to_trigger)
        if window.outOfOrderWinSched.lower() == "no" and len(accumulated_windows_ready_to_trigger) > 0:
            min_win, min_site = accumulated_windows_ready_to_trigger[0]
            r, c, m, result = db.get_min_window(dbcon, cur, queue_id)
            if not r:
                return False, 5, "db.get_min_window failed inside %s" % __name__, list_of_reply_windows
            w, s = result[0]
            if (w, s) != (min_win, min_site):
                accumulated_windows_ready_to_trigger = []
            else:
                accumulated_windows_ready_to_trigger = [(w, s)]

            print("accumulated_windows_ready_to_trigger = " + str(accumulated_windows_ready_to_trigger))
            print("w = " + str(w) + "s = " + str(s) + "min_win = " + str(min_win) + "min_site =" + str(min_site))

        # Ready complete windows for shipment. For time based windows, be careful what user asked for and accordingly
        # Either change status (to InDupWindow) or correctly set window no (if it should go to the next one).
        for window_no, site in accumulated_windows_ready_to_trigger:
            list_of_reply_windows.append(
                self.ship_this_window_to_plumb(dbcon, cur, queue_id, window_no, site, ignore_site, window))
            # Also change status to FORMED
            r, c, m = db.change_window_state(dbcon, cur, queue_id, window_no, site,
                                   "UNDERFORMATION", "FORMED", "By Window Manager")
            if not r:
                return r, c, "db.change_window_state failed with message %s inside %s" % (str(m), __name__), \
                       list_of_reply_windows

        # ToDo: make sure window replies have qSeqNoInst set.
        return True, 0, "okay", list_of_reply_windows

    @staticmethod
    def print_rcm(r, c, m):
        toPrint = "Returns: %s, return code: %d, message: %s" % (str(r), c, str(m))
        print(toPrint)

    def process_out_of_window_blocks(self, dbcon, cur, queue_id, window_start_time, window_end_time):
        # We have used two different strategies to handle those blocks that are out of the window.
        # (1) Deleting them or (2) Marking them appropriately, but not deleting them
        # (1) is simple (and it can potentially simplify many subsequent queries)
        # (2) allows extra information, that might be needed when blocks are missing or delayed
        # On 2021-05-05 I am switching to (2) but leaving (1) commented below if there is a need later.

        # mark any block that lie before start time of windowing
        r, c, m = db.mark_blocks(dbcon, cur, queue_id, window_start_time, True, self.BlockState.OUTOFLEFTBOUNDARY.name)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)
        # If window has an end time, mark any blocks after end time appropriately
        r, c, m = db.mark_blocks(dbcon, cur, queue_id, window_end_time, False, self.BlockState.OUTOFRIGHTBOUNDARY.name)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        '''
        # If window had a start time, we need to remove all blocks whose arrival time is before start date
        r, c, msg = db.delete_window_blocks(dbcon, cur, queue_id, window.startTime, "<")
        if not r:
            return r, c, "%s. This error is in: %s" % (msg, __name__), list_of_reply_windows
        # If window had an end time, we need to remove all blocks whose arrival time is after end date
        r, c, msg = db.delete_window_blocks(dbcon, cur, queue_id, window.endTime, ">")
        if not r:
            return r, c, "%s. This error is in: %s" % (msg, __name__), list_of_reply_windows
        '''
        return r, c, m

    def manage_missing_windows(self, dbcon, cur, queue_id, missing_data_windows, window,
                               window_size_metric, ignore_site):
        if window.waitMetric == "Time":
            # same handling for time-based and block-based
            r, c, msg, result = self.manage_T_missing(dbcon, cur, queue_id, missing_data_windows, window,
                                                      window_size_metric, ignore_site)

        elif window.waitMetric == "Blocks" and window.sizeMetric == "Blocks":
            r, c, msg, result = self.manage_BB_missing(dbcon, cur, queue_id, missing_data_windows, window,
                                                       window_size_metric, ignore_site)
        elif window.waitMetric == "Blocks" and window.sizeMetric == "Time":
            r, c, msg, result = self.manage_BT_missing(dbcon, cur, queue_id, missing_data_windows, window,
                                                       window_size_metric, ignore_site)
        elif window.waitMetric == "PercentBlocks" and window.sizeMetric == "Blocks":
            r, c, msg, result = self.manage_PB_missing(dbcon, cur, queue_id, missing_data_windows, window,
                                                       window_size_metric, ignore_site)
        elif window.waitMetric == "PercentBlocks" and window.sizeMetric == "Time":
            r, c, msg, result = self.manage_PT_missing(dbcon, cur, queue_id, missing_data_windows, window,
                                                       window_size_metric, ignore_site)
        else:
            # error
            r, c, msg, result = False, 100, "Unexpected window options", []

        return r, c, msg, result

    def manage_T_missing(self, dbcon, cur, queue_id, missing_data_windows, window,
                         window_size_metric, ignore_site):
        # return: True, code, message, ready_to_trigger
        ready_to_trigger = []

        '''
        Each entry in above list is either of the following two tuples:
        time-based:  (winNo, site, s, e, elapsed_hours)
        block based: (window_no, site, min_block, max_block, count)
        '''

        # For all missing windows, init the waiting
        for window_no, site, *_ in missing_data_windows:
            db.init_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)

        # For all missing windows, check if timeout has happened
        for window_no, site, *_ in missing_data_windows:
            r, c, msg, duration = db.get_wait_time_in_hours(dbcon, cur, queue_id, window_no, site, ignore_site)
            print("window.maxWait = " + str(window.maxWait))
            print("Window no: %d for site %s has been waiting for %d hours:" % (window_no, site, duration))
            if r and duration > window.maxWait:
                ready_to_trigger.append((window_no, site))

        return True, 0, "Okay", ready_to_trigger

    def manage_BB_missing(self, dbcon, cur, queue_id, missing_data_windows, window,
                          window_size_metric, ignore_site):
        ready_to_trigger = []
        required_blocks = window.size
        max_missing_allowed = window.maxWait
        # ToDo: Bug. Following count is probably not the actual block count -- fixed on 20201012
        for window_no, site, _, _, count in missing_data_windows:
            if (required_blocks - count) <= max_missing_allowed:
                ready_to_trigger.append((window_no, site))
        return True, 0, "Okay", ready_to_trigger

    def manage_BT_missing(self, dbcon, cur, queue_id, missing_data_windows, window,
                          window_size_metric, ignore_site):
        ready_to_trigger = []
        max_missing_allowed = window.maxWait
        # time - based: (winNo, site, s, e, elapsed_hours)
        for window_no, site, *_ in missing_data_windows:
            r, c, m, window_no, site, min_b, max_b, block_count = \
                db.get_a_specific_window_info(dbcon, cur, queue_id, window_no, site, ignore_site)
            if (block_count - (max_b - min_b + 1)) <= max_missing_allowed:
                ready_to_trigger.append((window_no, site))
        return True, 0, "Okay", ready_to_trigger

    def manage_PB_missing(self, dbcon, cur, queue_id, missing_data_windows, window,
                          window_size_metric, ignore_site):
        ready_to_trigger = []
        required_blocks = window.size
        max_missing_allowed_percentage = int(window.maxWait)
        max_missing_allowed = (max_missing_allowed_percentage / 100.0) * required_blocks
        for window_no, site, _, _, count in missing_data_windows:
            if (required_blocks - count) <= max_missing_allowed:
                ready_to_trigger.append((window_no, site))
        return True, 0, "Okay", ready_to_trigger

    def manage_PT_missing(self, dbcon, cur, queue_id, missing_data_windows, window,
                          window_size_metric, ignore_site):
        ready_to_trigger = []
        max_missing_allowed_percentage = int(window.maxWait)
        # time - based: (winNo, site, s, e, elapsed_hours)
        for window_no, site, *_ in missing_data_windows:
            r, c, m, window_no, site, min_b, max_b, block_count = \
                db.get_a_specific_window_info(dbcon, cur, queue_id, window_no, site, ignore_site)
            max_missing_allowed = (max_missing_allowed_percentage / 100.0) * block_count
            if (block_count - (max_b - min_b + 1)) <= max_missing_allowed:
                ready_to_trigger.append((window_no, site))
        return True, 0, "Okay", ready_to_trigger

    def manage_inconclusive_windows(self, dbcon, cur, queue_id, inconclusive_windows, window,
                                    window_size_metric, ignore_site):
        # return: True, code, message, ready_to_trigger
        ready_to_trigger = []
        max_wait_for_inconclusive_windows = 40  # in hours
        '''
        Each entry in above list is either of the following two tuples:
        time-based:  (winNo, site, s, e, elapsed_hours)
        block based: (window_no, site, min_block, max_block, count)
        '''

        # For all missing and inconclusive windows, init the waiting
        for window_no, site, *_ in inconclusive_windows:
            db.init_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)

        # For inconclusive windows: If there was a new block after last recorded wait time, then reset wait time to 0
        for window_no, site, *_ in inconclusive_windows:
            r, c, msg = db.did_window_receive_new_block(dbcon, cur, queue_id, window_no, site, ignore_site)
            if r:
                db.update_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)

        # For all inconclusive windows, check if timeout has happened
        for window_no, site, *_ in inconclusive_windows:
            r, c, msg, duration = db.get_wait_time_in_hours(dbcon, cur, queue_id, window_no, site, ignore_site)
            print("window.maxWait = " + str(window.maxWait))
            print("Window no: %d for site %s has been waiting for %d hours:" % (window_no, site, duration))

            if r and duration > max_wait_for_inconclusive_windows:
                ready_to_trigger.append((window_no, site))

        print("Inconclusive ready to trigger: %s" % (str(ready_to_trigger)))
        return True, 0, "Okay", ready_to_trigger

    def manage_missing_and_inconclusive_windows(self, dbcon, cur, queue_id, missing_data_windows,
                                                inconclusive_windows, window, window_size_metric, ignore_site):
        # return: True, code, message, ready_to_trigger
        ready_to_trigger = []
        max_wait_for_inconclusive_windows = 48  # in hours
        '''
        Each entry in above list is either of the following two tuples:
        time-based:  (winNo, site, s, e, elapsed_hours)
        block based: (window_no, site, min_block, max_block, count)
        '''

        # For all missing and inconclusive windows, init the waiting
        for window_no, site, *_ in missing_data_windows:
            db.init_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)
        for window_no, site, *_ in inconclusive_windows:
            db.init_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)

        # For inconclusive windows: If there was a new block after last recorded wait time, then reset wait time to 0
        for window_no, site, *_ in inconclusive_windows:
            r, c, msg = db.did_window_receive_new_block(dbcon, cur, queue_id, window_no, site, ignore_site)
            if r:
                db.update_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site)

        # For all missing and inconclusive windows, check if timeout has happened
        for window_no, site, *_ in missing_data_windows:
            r, c, msg, duration = db.get_wait_time_in_hours(dbcon, cur, queue_id, window_no, site, ignore_site)
            if r and duration > window.maxWait:
                ready_to_trigger.append((window_no, site))

        for window_no, site, *_ in inconclusive_windows:
            r, c, msg, duration = db.get_wait_time_in_hours(dbcon, cur, queue_id, window_no, site, ignore_site)
            if r and duration > max_wait_for_inconclusive_windows:
                ready_to_trigger.append((window_no, site))

        return True, 0, "Okay", ready_to_trigger

    @staticmethod
    def get_windows_whose_time_has_come(dbcon, cur, queue_id, window):
        if window.sizeMetric == "Time":
            r, c, complete_windows, missing_data_windows, inconclusive_windows = \
                db.for_time_windows_get_windows_whose_time_has_come(dbcon, cur, queue_id)
        else:
            r, c, complete_windows, missing_data_windows, inconclusive_windows = \
                db.for_block_windows_get_windows_whose_time_has_come(dbcon, cur, queue_id, window)

        return r, c, complete_windows, missing_data_windows, inconclusive_windows, window.sizeMetric

    def process_new_blocks(self, dbcon, cur, queue_id, window):
        # purpose: (1) block state change from UNRESERVED => Some other state (2) Finding new window numbers
        # Wrinkle: window size can be either of hours or block-count
        # Solution: call a generic function here and then use a specialized function at db layer
        r, c, m = db.assign_window_no_to_blocks(dbcon, cur, queue_id, window)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        # detect and mark any late blocks
        r, c, m = db.mark_blocks_as_late(dbcon, cur, queue_id)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        # Also remove those blocks from missing block table those have been marked late
        r, c, m = db.delete_late_blocks_from_missing_table(dbcon, cur, queue_id)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        # detect and mark any duplicate blocks
        r, c, m = db.mark_blocks_as_duplicate(dbcon, cur, queue_id)
        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        return r, c, m

    def populate_new_window(self, dbcon, cur, queue_id, new_window_no, site, window, ignore_site):
        # ToDo: we will fill start and end block numbers later for both time based and block based windows at the
        #       completion time.
        if window.sizeMetric == "Time":
            int_w, str_s, str_e, str_fmt = self.get_window_no_start_and_end(new_window_no,
                                                                            window.startTime, window.size, 1)
            r, c, m = db.insert_new_window_time_based(dbcon, cur, queue_id, int_w, str_s, str_e, str_fmt,
                                                      site, ignore_site)
        else:
            r, c, m = db.insert_new_window_block_based(dbcon, cur, queue_id, new_window_no, site, ignore_site)

        if not r:
            return r, c, "%s. This error is in: %s" % (m, __name__)

        return r, c, "OK"

    @staticmethod
    def ship_this_window_to_plumb(dbcon, cur, queue_id, window_no, site, ignore_site, window):
        # ToDo: Filling other information later
        rw = ReplyWindow()
        rw.queueID = queue_id
        rw.windowNo = window_no
        # rw.startTime = s.strftime("%Y %m %d %H %M %S UTC")  # Convert time to Plumb style string
        # rw.endTime = e.strftime("%Y %m %d %H %M %S UTC")
        r, c, blocks = db.get_blocks_in_window(dbcon, cur, queue_id, window_no, site, ignore_site)
        if not r:
            return rw
        # cut-off does not apply to windows where user provided exact start and end time.
        if window.timeMetric != "TimeViaStartAndEnd":
            if window.cutoff.lower() == "Cutoff_Exactly_Once_To_Next_Window" and len(blocks) > 0:
                # Just exclude last block in this window.  Change this block number's window no to the next one
                last_data_uri = blocks[-1]
                blocks.remove(last_data_uri)
                # We will now keep original window numbers and wont change it.
                # db.update_window_no(dbcon, cur, queue_id, last_data_uri)
                # If I am not the first window, include the block just before the first block of this window
                if window_no != 1:
                    r, c, msg_or_url = WindowManager.get_next_block_uri(dbcon, cur, queue_id, blocks[0])
                    if r and (msg_or_url != "" or msg_or_url is not None):
                        blocks.insert(0, msg_or_url)
                        db.update_window_state(dbcon, cur, queue_id, blocks[0])
            elif window.cutoff.lower() == "Cutoff_Duplicate_Delivery" and len(blocks) > 0:
                # Just change first and last block's state to INMULTIWINDOW
                db.update_window_state(dbcon, cur, queue_id, blocks[-1])
                if window_no != 1:
                    r, c, msg_or_url = WindowManager.get_next_block_uri(dbcon, cur, queue_id, blocks[0])
                    if r and (msg_or_url != "" or msg_or_url is not None):
                        blocks.insert(0, msg_or_url)
                        db.update_window_state(dbcon, cur, queue_id, blocks[0])
            else:
                pass  # No need to do anything for Cutoff_Exactly_Once_To_Current_Window

        rw.listOfBlocksInWindowInThisWindow.extend(blocks)
        rw.noOfBlocksInThisWindow = len(blocks)
        return rw

    @staticmethod
    def get_next_block_uri(dbcon, cur, queue_id, first_block):
        site_from_url = regex.get_queue_sequence_number_instance(first_block)
        block_no_from_url = int(regex.get_sequence_number(first_block)) - 1
        r, c, msg_or_url = db.get_block_URI(dbcon, cur, queue_id, site_from_url, block_no_from_url)
        return r, c, msg_or_url

    @staticmethod
    def get_window_no_start_and_end(str_window_no, str_window_start, str_window_size, str_increment):
        int_window_no = int(str_window_no)
        int_window_size = int(str_window_size)
        time_format_string = "%Y %m %d %H %M %Z"
        out_time_format_string = "%Y %m %d %H %M"
        out_time_format_string_for_mysql_db = "%Y %m %d %H %i"  # unfortunately last MySQL format specifier is different
        datetime_obj = datetime.strptime(str_window_start, time_format_string)
        delta_hours_start = int_window_size * (int_window_no - str_increment)
        delta_hours_end = int_window_size * int_window_no

        start = datetime_obj + timedelta(hours=delta_hours_start)
        end = datetime_obj + timedelta(hours=delta_hours_end)

        return (int_window_no, str(start.strftime(out_time_format_string)),
                str(end.strftime(out_time_format_string)),
                out_time_format_string_for_mysql_db)

    @staticmethod
    def get_db_connection_and_cursor():
        dbcon = db.connectToDB(Host, User, Pass, DB)
        cur = ""
        # ToDo: Do we need to disable auto db commit?
        if dbcon is None:
            return False, 1, "dbcon failed inside %s" % __name__, dbcon, cur
        try:
            cur = dbcon.cursor()
        except MySQLdb.Error as e:
            return False, 2, "cursor acquisition failed inside %s. Error is: str(%s)" % (__name__, str(e)), \
                   dbcon, cur

        return True, 0, "Db connection and cursor acquired.", dbcon, cur

    def get_number_of_windows(self):
        if self.listOfWindows is None:
            err_message = "WindowManager's listOfWindows is Null!"
        else:
            err_message = "WindowManager has a list windows with count: " + str(len(self.listOfWindows))

        return err_message

    def change_window_state(self, queue_id, window_no, site, old_state, new_state, app_id):
        r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
        if not r:
            return r, c, m
        os = WindowOperationRequest.DESCRIPTOR.fields_by_name['currentWindowState'].enum_type.values_by_number[
            old_state].name
        ns = WindowOperationRequest.DESCRIPTOR.fields_by_name['nextWindowState'].enum_type.values_by_number[
            new_state].name
        return db.change_window_state(dbcon, cur, queue_id, window_no, site, os, ns, app_id)

    def unreserve_window(self, queue_id, window_no):
        r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
        if not r:
            return r, c, m
        return db.unreserve_window(dbcon, cur, queue_id, window_no)

    class BlockState(Enum):
        RESERVED = 1
        UNRESERVED = 2
        FAULTY = 3
        LATE = 4
        INWINDOW = 5
        INMULTIWINDOW = 6
        DUPLICATE = 7
        OUTOFLEFTBOUNDARY = 8
        OUTOFRIGHTBOUNDARY = 9
        UNWANTED = 10

    def ls_windowed_queues(self):
        return True, 0, "OK", self.windowMap.keys()

    def ls_all_window_statuses(self):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.get_all_window_statuses(dbcon, cur)
            if c != 0:
                return False, c, "db.get_all_window_statuses failed.  This failure is in %s" % __name__, result
            return True, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 230, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def ls_win(self, queue_id):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.get_window_status(dbcon, cur, queue_id)
            if c != 0:
                return False, c, "db.get_window_status failed.  This failure is in %s" % __name__, result
            return True, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 230, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def ls_win_in_status(self, queue_id, status):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.get_window_in_status(dbcon, cur, queue_id, status)
            if c != 0:
                return False, c, "db.get_window_in_status failed.  This failure is in %s" % __name__, result
            return True, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 230, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def set_win_status(self, queue_id, window_no, status):
        r, c, m = True, 0, "OK"
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m
            r, c, m = db.set_win_status_unconditional(dbcon, cur, queue_id, status, window_no, True)
            if not r:
                return r, c, "db.set_win_status_unconditional failed. %s. This failure is in %s" % (m, __name__)
            return True, c, "OK"
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 232, str(e)
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def ls_f_in_win(self, queue_id, window_no):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.get_files_in_a_window(dbcon, cur, queue_id, window_no)
            if c != 0:
                return False, c, "db.get_files_in_a_window failed.  This failure is in %s" % __name__, result
            return True, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 231, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def mv_f_in_win(self, queue_id, window_no, file_uri):
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m
            r, c, m = db.move_f_in_win(dbcon, cur, queue_id, window_no, file_uri)
            if not r:
                return r, c, "db.move_f_in_win failed. %s. This failure is in %s" % (m, __name__)
            return True, c, "OK"
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 233, str(e)
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def ls_missing_files_in_a_window(self, queue_id):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.ls_missing_files_in_a_window(dbcon, cur, queue_id)
            if c != 0:
                return r, c, "db.ls_missing_files_in_a_window failed.  This failure is in %s" % __name__, result
            return r, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 231, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()

    def ls_site_info_in_a_window(self, queue_id):
        r, c, m, result = True, 0, "OK", []
        dbcon, cur = None, None
        try:
            r, c, m, dbcon, cur = self.get_db_connection_and_cursor()
            if not r:
                return r, c, m, result
            r, c, m, result = db.ls_site_info_in_a_window(dbcon, cur, queue_id)
            if c != 0:
                return r, c, "db.ls_site_info_in_a_window failed.  This failure is in %s" % __name__, result
            return r, c, "OK", result
        except Exception as e:
            sys.stderr.write("Exception: " + str(e) + "\n")
            return False, 231, str(e), result
        finally:
            if (dbcon is not None) and (cur is not None):
                cur.close()
                dbcon.close()


if __name__ == "__main__":
    # This is for local testing only.  First off we need some test data
    print("Hello")
    q_name = "rssacint_1"
    win_list = [queue_api_pb2.Window(queueID=q_name, size=24, sizeMetric="Time", timeMetric="TimeViaStartAndEnd",
                                     startTime="2019 08 19 00 00 UTC", endTime="2019 08 19 23 59 UTC",
                                     timeFormat="%Y %m %d %H %M %UTC", maxWait=3, waitMetric="Time",
                                     cutoff="Cutoff_Exactly_Once_To_Current_Window", outOfOrderWinSched="No",
                                     useQSeqNoInst="Yes", qSeqNoInstList="lax,mia,ari,ams,iad,sin")]
    wm = WindowManager(win_list)

    ##print(wm.get_windows_to_schedule(q_name))
    print(wm.ls_all_window_statuses())

    # Following for testing.
    # [queueID: "rsscint"
    # size: 24
    # sizeMetric: "Time"
    # timeMetric: "TimeViaSizeAndStart"
    # startTime: "2020 04 01 00 00 Z"
    # endTime: "7000 04 01 00 00 UTC"
    # timeFormat: "%Y %m %d %H %M %Z"
    # maxWait: 3
    # waitMetric: "Time"
    # cutoff: "Cutoff_Exactly_Once_To_Current_Window"
    # outOfOrderWinSched: "Yes"
    # useQSeqNoInst: "No"
    # , queueID: "message.fsdb"
    # size: 72
    # sizeMetric: "Blocks"
    # timeMetric: "TimeViaStartAndEnd"
    # startTime: "2020 04 02 00 00 UTC"
    # endTime: "2020 04 05 00 00 Z"
    # timeFormat: "%Y %m %d %H %M %Z"
    # maxWait: 1
    # waitMetric: "Time"
    # cutoff: "Cutoff_Exactly_Once_To_Current_Window"
    # outOfOrderWinSched: "Yes"
    # useQSeqNoInst: "No"
    # ]

'''
This space is for relatively detailed notes, especially related to windowing algorithm.

(note 1) An end-user can provide a list of sites in their YAML (in queue_sequence_number_instance_list) but that is
not a mandatory field.  If not given, system needs to detect any sites.  If given, system need to ignore all those
sites not given by the user.  For ignoring purposes, a new file state name 'UNWANTED' is now available.  All blocks
whose site name is something not given by the user, should be tagged as UNWANTED.
Additionally, when user gives window size in number of blocks, the start of window is
still in terms of time.  Hence we need to record the block sequence number for the first block of the window because
we need it to find out unique window numbers.
Currently we run this algorithm on each call to find any schedule-able windows.  Later if we find that doing so has
high cost, then we can do one of the two:
(a) For the case when user gives list of sites, and when we detect that those sites are in the site info table, we
stop doing so.  Henceforth, it will happen only once if window manager server is restarted. But be careful about out-
of-order blocks for a site where a smaller block no comes later (When you have already recorded a bigger no in the
siteInfoTable).  Probably we should not change start block no in site info table, once it is recorded because some
windows might be in progress using old no and a new onw might generate duplicate window numbers.
(b) For system-detection of any new sites, we can run this algo say after every x calls (instead of every call) when
we have some small (at least 1) number of sites present in the site info table.
<20210518 update> Finding min for each site can become tricky if blocks come out-of-order. To avoid un-necessary
complexity, now we use the fact that each site's sequence number starts at 0.  We use this fact to calculate unique
window numbers now. To avoid excessive code changes, we will record this 0 in site_info_table so that rest of the
code still works without more changes.


(note 2) We need to see in which window a new block (with status UNRESERVED) goes. For that purpose we need a mapping
from block no to window no.  There are two details:
(a) a window might be (start time, duration) or (start time, size in blocks)
(b) depending on user request of cutoff_delivery_requirement, a block can be in 1 or 2 windows
We actually don't assign a block any window number.  Rather we record new windows in WindowState table.  If a block
for a window that is already in progress comes, we just change that block's status (InWindow or InDupWindow etc).
A block might be slapped as OUTOFLEFTBOUNDARY (block before window start), OUTOFRIGHTBOUNDARY (block after window end),
LATE, or DUPLICATE statuses.
In summary this step has two purposes: (1) For new blocks (State = UNRESERVED),decide what will be its next state.
(2) To identify if a new window is starting (and hence we need to start keeping its state)
2020-09-28 update: We ended up having a WindowNo filed for each blcok in the queue table because:
(a) Later queries were becoming unwieldy whenever we needed a window no for a block
(b) Formula for block based window used siteInfoTable while time based doesn't. Doing it once, we hide that details
at one place and it doesn't come into play again.
(c) More importantly, we assume window numbers don't change.  So why re-calculate?

(note 2a) What are duplicates? Those blocks whose window has already processed. We need to be careful what kind of
cut-off is requested by the user (if this block is a boundary block).
For Cutoff_Duplicate_Delivery, if either of the window is not under-formation state, then we mark it as duplicate.
How to detect duplicate:
If window related to the block has processed AND this block number is not in Missing Block Record table, then mark
it as duplicate.
By the way, if window is still processing, that specific block will be in the DB and arrival of duplicate at that time
will not be allowed to insert in db.  These conditions are for after when a window has been processed, processed blocks
released, and then a duplicate comes.

(note 2b) What are late? If a block is in missing table, it is late.  After marking it late, remove it from missing
table.


(note 3) About detecting window completeness.  First question is when to run a completion algorithm.
For windows whose size is in terms of time => when current time > window end time
For windows whose size is in terms of blocks => we need to detect completeness (or missing stuff)
=> If any block from the next window has arrived, consider time for the previous as up
=> See percentage of missing, if user threshold meets, trigger the window
=> For 100% block requirement, check if missing block(s) are in faulty sate somewhere upstream
=> When unable to detect completeness, trigger window after two days (48 hours)
'''

'''
Test case:
    win_list = [queue_api_pb2.Window(queueID = q_name,size = 24,sizeMetric = "Time",timeMetric = "TimeViaSizeAndStart",
                         startTime = "2019 08 20 00 00 UTC",endTime = "7000 04 01 00 00 UTC",
                         timeFormat = "%Y %m %d %H %M %UTC",maxWait = 3,waitMetric = "Time",
                         cutoff = "Cutoff_Exactly_Once_To_Current_Window",outOfOrderWinSched = "No",
                        useQSeqNoInst = "Yes", qSeqNoInstList = "lax,mia,ari,ams,iad,sin")]
'''

'''
    # ToDo: Heads-up
    # When doing clean-up in specific WindowState tables, probably you should leave the final processed window so that
    # we can see which windows are done

    # ToDo: mark_blocks_as_late relies on missing blocks in MissingData table.  Dont forget to put missing blocks
    # there.  That will happen when we think window is ready to trigger and we can put any missing blocks at that
    # time.
    
    # ToDo: populate_new_window put few info in the window table. Some other fields will need to fill in when we
    # are ready to trigger the window.
    
    # ToDo: Which window waits and 3 edge cases: If end block belongs to next window, then current window should not
    #       wait if last block is missing.  If last block goes to both current and next windows, and if that block is
    #       missing will both windows wait?  Thats wasteful but it needs to think what should happen then. Another
    #       option is to always let the current window wait for missing, and not the next one.  In that case, if block
    #       is missing, next one can ignore it.  But will that not violate user API semantics of max wait?  Probably
    #       not if last window waited for it, but yes if this window is completing earlier than previous.
    
    # ToDo: Improvements for reduced wait delay: If some block is missing in a window, walk the YAML queue DAG to
    #       find out if missing block is marked as faulty somewhere upstream (and hence no need to wait for it).
    
    # ToDo: Clean-up queue tables being careful about blocks that are in multi-window.  Del blocks for windows
    #       that are Done. We are doing it here to avoid race conditions especially for blocks that are in two
    #       widows.


'''
