#!/usr/bin/env python3

import time

import MySQLdb

from top.sup import db
from top.conf import sysConfInfo


Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = False
#..........................................................................
# return (errCode, errMessage, count, gap, minutesSinceLastReserveNext, uri
#

def lander_queue_peek_next_Inner(queue_id, queue_owner):

    errCode = 0
    errMessage = "OK"
    count = 0
    gap = 1
    minutesSinceLastReserveNext = 0
    uri = "UN-INITIALIZED"

   
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__), count, gap, minutesSinceLastReserveNext, uri
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquisition failed inside %s" %(__name__), count, gap, minutesSinceLastReserveNext, uri


    if db.__doesQueueExists(dbcon, cur, queue_id) != 0:
        return 3, "queue with id %s does not exist.  This failure is in %s" %(queue_id, __name__), count, gap, minutesSinceLastReserveNext, uri


    # Confim queue_owner is truely owner of queue_id
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 4, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, queue_id, __name__), count, gap, minutesSinceLastReserveNext, uri

    # Direct delivery is only possible if to_queue is active
    (isSuccess, val) = db.__getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, "isActive")
    if isSuccess != 0 or val != "1":
        return (5, "Queue %s is not active and hence can't receive data.  This failure is in %s"  %(queue_id, __name__), count, gap, minutesSinceLastReserveNext, uri)

    isSuccess, uri, site, seqNo  = db.peekNext(dbcon, cur, queue_id, queue_owner)
    if isSuccess != 0:
        return isSuccess, "db.peekNext failed. This failure is in %s"  %(__name__), count, gap, minutesSinceLastReserveNext, uri

    count = 1

    # get block no and last time from key, val
    if db.__doesKeyExistsInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, site):
        retCode, last_block = db.__getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, site)
        if retCode != 0:
            return retCode, "db.__getValFromSubscriberKeyValTable failed. This failure is in %s"  %(__name__), count, gap, minutesSinceLastReserveNext, uri
        gap = int(seqNo) - int(last_block)

    key = "%s_reservation_timestamp" %(site)
    if db.__doesKeyExistsInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key):
        retCode, t1 = db.__getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key)
        if retCode != 0:
            return retCode, "db.__getValFromSubscriberKeyValTable failed. This failure is in %s"  %(__name__), count, gap, minutesSinceLastReserveNext, uri

        minutesSinceLastReserveNext = int((int(time.time()) - int(t1)) / 60)



    cur.close()
    db.__disconnectFromDB(dbcon)
    
    return errCode, errMessage, count, gap, minutesSinceLastReserveNext, uri

#..........................................................................
def lander_queue_peek_next(queue_id, queue_owner):
    errCode, errMessage, count, gap, minutesSinceLastReserveNext, uri = lander_queue_peek_next_Inner(queue_id, queue_owner)
    print (errCode)
    print (errMessage)
    print (count)
    print (gap)
    print (minutesSinceLastReserveNext)
    print (uri)

    return errCode, errMessage, count, gap, minutesSinceLastReserveNext, uri

#..........................................................................
if __name__ == '__main__':
    queue_id  = "pcap.sz_1"
    queue_owner = "hlander"
    

    result = lander_queue_peek_next(queue_id, queue_owner)
    
    print (result)


