#!/usr/bin/env python3
import os
import time

import MySQLdb

from top.queue_pull.lander_queue_peek_next import lander_queue_peek_next
from top.sup import db, regex
from top.conf import sysConfInfo


Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = False
#..........................................................................
# return (errCode, errMessage, count, gap, minutesSinceLastReserveNext, uri
#

def lander_queue_reserve_next_Inner(queue_id, uri, queue_owner):


    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquisition failed inside %s" %(__name__)


    if db.__doesQueueExists(dbcon, cur, queue_id) != 0:
        return 3, "queue with id %s does not exist.  This failure is in %s" %(queue_id, __name__)


    # Confim queue_owner is truely owner of queue_id
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 4, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, queue_id, __name__)



    isSuccess = db.__doesFileExistsInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, uri)
    if isSuccess != 0:
        return 5, "%s file is not in queue %s. This failure is in %s" %(uri, queue_id, __name__)

    isSuccess, err_message, _, sequence_no, _, site, _ = regex.match_file_name(os.path.basename(uri))
    if isSuccess != 0:
        return 6, "regex.match_file_name failed with error: %s. This failure is in %s" %(err_message, __name__)

    # (1) Reserve the file, but dont commit
    isSuccess = db.__doAtomicReserveInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, uri, False)
    if isSuccess != 0:
        return 7, "__doAtomicReserveInSubscriberQueueTable failed.  This failure is in %s" %(__name__)
    # (2) Add / update block number for this site
    last_block = 0
    if db.__doesKeyExistsInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, site):
        retCode, last_block = db.__getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, site)

    key = site
    val = max(int(sequence_no), int(last_block))
    isSuccess = db.__populateSubscriberKeyValTableWrapper(dbcon, cur, queue_owner, queue_id, key, val, False)
    if isSuccess != 0:
        return isSuccess, "db.__populateSubscriberKeyValTableWrapper failed.  This failure is in %s" %(__name__)
    # (3) Add / update time stamp for this site
    key = "%s_reservation_timestamp" %(site)
    val = str(int(time.time()))
    isSuccess = db.__populateSubscriberKeyValTableWrapper(dbcon, cur, queue_owner, queue_id, key, val, False)
    if isSuccess != 0:
        return isSuccess, "db.__populateSubscriberKeyValTableWrapper failed.  This failure is in %s" %(__name__)

    # if all is okay, now commit
    try:
        dbcon.commit()
    except MySQLdb.Error as e:
        return 9, "Commit failed with error: %s. This error is in: %s" %(str(e), __name__)

    cur.close()
    db.__disconnectFromDB(dbcon)

    return isSuccess, "OK"


#..........................................................................
def lander_queue_reserve_next(queue_id, uri, queue_owner):
    errCode, errMessage = lander_queue_reserve_next_Inner(queue_id, uri, queue_owner)
    print (errCode)
    print (errMessage)

    return errCode, errMessage

#..........................................................................
if __name__ == '__main__':
    queue_id  = "pcap.sz_1"
    queue_owner = "hlander"
    uri = "/user/plumb/hle/raw/pcap.sz/conqdata/20201127-044200-01884886.lax.pcap.sz"

    a,b,c,d,e,uri = lander_queue_peek_next(queue_id, queue_owner)
    if a == 0:
        result = lander_queue_reserve_next(queue_id, uri, queue_owner)
        print (result)


