#!/usr/bin/env python3

import os
import MySQLdb

from .top.sup import opsys, db, hdfs
from .top.conf import sysConfInfo
from .top.queue_admin import listDerivedQueues
from .top.sup import regex





Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = True



#..........................................................................
def __doDeliveryToQueue(dbcon, cur, hdfsURI, modTime, queue_id, queue_owner, eventTime, eventTimeFormat, siteName, blockNo, info, tailQID, shouldCommit):

    # Direct delivery is only possible if queue is active
    isSuccess, val = db.__getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, "isActive")
    if isSuccess != 0 or val != "1":
        return 20, " connected queue %s is not active. Hence no data for it!  This failure is in %s" %(queue_id, __name__)

    fileState = "UNRESERVED"
    if db.__populateSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfsURI, fileState,
                                   modTime, eventTime, eventTimeFormat, siteName, blockNo, info, tailQID, False) != 0:
        return 21, " db.__populateSubscriberQueueTable failed..  This failure is in %s" %(__name__)

    # Now lets try to commit the transaction
    try:
        if shouldCommit == True:
            dbcon.commit()

        return 0, "OK"
    except MySQLdb.Error as e:
        if debug: print (e)
        return 22, " DB commit failure.  This failure is in %s" %(__name__)




#..........................................................................
def distributeLocal_Phase1(from_queue_id, localfile, queue_owner):

    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__), None
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquisition failed inside %s" %(__name__), None


    if db.__doesQueueExists(dbcon, cur, from_queue_id) != 0:
        return 3, "queue with id %s does not exist.  This failure is in %s" %(from_queue_id, __name__), None

    # Confim queue_owner is truely owner of  from_queue_id
    retCode, queue_owner_fromDB = db.__getQueueOwner(dbcon, cur, from_queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return 4, "%s is not owner of queue %s. This failure is in %s" %(queue_owner, from_queue_id, __name__), None


    # Direct delivery is only possible if to_queue is active
    isSuccess, val = db.__getValFromSubscriberKeyValTable(dbcon, cur, from_queue_id, queue_owner, "isActive")
    if isSuccess != 0 or val != "1":
        return 5, " Queue %s is not active and hence can't receive data.  This failure is in %s" %(from_queue_id, __name__), None

    # find what is HDFS base dir
    isSuccess, hdfsBaseDir = db.__getValFromKeyValTable(dbcon, cur, "HDFS_Base_Dir")
    hdfsURI = ""
    data_type = ""
    if isSuccess != 0:
        return (6, " Failed to get HDFS_Base_Dir from Db.  This failure is in %s " %(__name__), None)



    # find data_type associated with the from_queue_id
    isSuccess, data_type = db.__getDataTypeFromQueueTable(dbcon, cur, from_queue_id)
    if isSuccess != 0:
        return (7, " Failed to get data_type associated with queue %s.  This failure is in %s " %(from_queue_id, __name__), None)


    # find Linux group associated with data_type
    isSuccess, groupName = db.__getValFromKeyValTable(dbcon, cur, data_type)
    if isSuccess != 0:
        return (8, " Failed to get Linux group associated with data_type %s.  This failure is in %s " %(data_type, __name__), None)


    # lets make hdfs_uri
    connectedQueueData = "conqdata"
    dstHdfsURI =  "%s/%s/%s/%s/%s" %(hdfsBaseDir, data_type, from_queue_id, connectedQueueData, os.path.basename(localfile))
    dstHdfsDir =  "%s/%s/%s/%s" %(hdfsBaseDir, data_type, from_queue_id, connectedQueueData)


    # make dir hierarchy if does not exist
    if hdfs.makeHDFSDirIfDoesNotExist(dstHdfsDir) != 0:
        return (9, " Failed to make HDFS directory hierarchy %s.  This failure is in %s " %(dstHdfsDir, __name__), None)

    # protect directory
    result = hdfs.setFileOrDirRights(dstHdfsDir, sysConfInfo.DirRights)
    if result != 0:
        return (10, " Failed to set dir rights.  This failure is in %s " %(__name__), None)

    result = hdfs.setFileOrDirGroup(dstHdfsDir, groupName)
    if result != 0:
        return (11, " Failed to set dir group.  This failure is in %s " %(__name__), None)



    cur.close()
    db.__disconnectFromDB(dbcon)


    return (0, "OK", dstHdfsURI)

#.......................................................................................
def distributeLocal_Phase2(from_queue_id, localfile, hdfs_file_uri, queue_owner,
                           eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID):
    # some checks from phase1 needs repeatition because we don't want to be in a situation
    # where client only does round two of call and something changes between two rounds
    retCode, errMessage, dstHdfsURI = distributeLocal_Phase1(from_queue_id, localfile, queue_owner)

    if retCode != 0: return retCode, errMessage, dstHdfsURI

    if dstHdfsURI != hdfs_file_uri:
        return 12, "client provided uri %s is different from what system asked for %s" %(hdfs_file_uri, dstHdfsURI), None

    # protect file
    result = hdfs.setFileOrDirRights(hdfs_file_uri, sysConfInfo.FileRights)
    if result != 0:
        return (13, " Failed to set file rights.  This failure is in %s " %(__name__), None)


    result, modTime = hdfs.getHDFSFileModTime(hdfs_file_uri)
    if result != 0:
        return 14, " Failed to get mod time of HDFS file %s.  This failure is in %s" %(hdfs_file_uri, __name__), None


    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__), None
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquisition failed inside %s" %(__name__), None



    # Actual file sits in source queue (i.e. from_queue_id)
    if db.__populateDataTable(dbcon, cur, hdfs_file_uri, from_queue_id, False) != 0:
        return 20, " db.__populateDataTable failed..  This failure is in %s" %(__name__), hdfs_file_uri


    # find connected children queues
    result, errMessage, queueList = listDerivedQueues.listDerivedQueues(from_queue_id, queue_owner)
    if result != 0:
        return result, errMessage, hdfs_file_uri
    finalResult = 0
    errMessage = ""
    atLeastOneSuccess = False

    for q in queueList:
        # get owner of q
        # ToDo: Currently if code fails to find owner of ANY connected queue, whole distribution fails.
        #       We might like to keep going with partial success?
        retCode, q_owner_fromDB = db.__getQueueOwner(dbcon, cur, q)
        if retCode != 0 :
            db.__cleanup(dbcon, cur)
            return 21, "Failed to find owner of the queue %s.  This failure is in %s" %(q, __name__), None

        result, message = __doDeliveryToQueue(dbcon, cur, hdfs_file_uri, modTime, q, q_owner_fromDB, \
                                               eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID, False)
        finalResult += result
        if result != 0: errMessage += message
        else: atLeastOneSuccess = True

    #  If we are able to deliver to at least one derived queue, we consider it a success
    if atLeastOneSuccess is False:
        db.__cleanup(dbcon, cur)
        return 22, " %s  This failure is in %s" % (errMessage, __name__), hdfs_file_uri

    # Now lets try to commit the transaction
    try:
        dbcon.commit()
    except MySQLdb.Error as e:
        if debug:
            print (e)
            db.__cleanup(dbcon, cur)
            return 23, " DB commit failure.  This failure is in %s" %(__name__), hdfs_file_uri

    db.__cleanup(dbcon, cur)
    return 0,  "OK", hdfs_file_uri




#..........................................................................
def distributeLocal(request, phaseNo):
    retVal, retMessage, dstHdfsURI  = 30, "Empty", "Empty"

    # enforece our naming conventoins
    retCode, retMessage, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        regex.match_file_name(os.path.basename(request.localfile), request.from_queue_id)
    if 0 != retCode:
        return (32, retMessage, "Empty")


    if phaseNo == 1:
      retVal, retMessage, dstHdfsURI = distributeLocal_Phase1(request.from_queue_id, request.localfile, request.caller_id)

    elif phaseNo == 2:
      retVal, retMessage, dstHdfsURI = distributeLocal_Phase2(request.from_queue_id, request.localfile, request.hdfs_file_uri, request.caller_id, \
                                                              date, regex.get_event_time_format(), sequence_no, optional_info, sequence_no_instance, queue_name)

    else:
        retVal, retMessage, dstHdfsURI = 31, "distributeLocal has exactly 2 rounds.", "Empty"


    if dstHdfsURI == None: dstHdfsURI = "deadbeef"

    print (retVal)
    print (retMessage)
    print (dstHdfsURI)
    return retVal, retMessage, dstHdfsURI


#..........................................................................
if __name__ == '__main__':
    localfile = "/nfs/ruby/aqadeer/working/traces/20141117-065927-00358611.erf.sz"
    from_queue_id = "queue_1"
    
    
    retVal =  distributeLocal(from_queue_id, localfile)
    print ("retVal = " + str(retVal))




