#!/bin/bash

#LANDER_DEBUG_SHELL_HLE=y
LANDER_DEBUG_SHELL_HLE=""

# Caution: Following is for debugging only.  When multiple calls happend
# concurrently, they will try to go to a single file and that is bad.
##errF="/tmp/errF"
errF="/dev/null"


CALLHDR="import top.client_rpc_proxy; from top.client_rpc_proxy import lander_queue_api_client_proxy as x"


function _debug() {
    local err=$?
    [ -n "$LANDER_DEBUG_SHELL_HLE" ] && echo "DEBUG: $(cat $errF)" >&2
    return $err
}


# $1 = retCode
# $2 = toPrint
function _print_stdout() {
    [[ $1 == 0 ]] && [ -n "$2" ] && echo  "$2" 
}


function lander_queue_list_derived_queues_hle() {
    callString="$CALLHDR; x.e(x.listDerivedQueues('$1', '$2'))"              
    retString=$((python3 -c"$callString")2>$errF)
    _debug
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    return $retCode

}


function lander_queue_create_hle() {
    callString="$CALLHDR; x.e(x.createQueue('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug 
}


function lander_queue_delete_hle() {
    callString="$CALLHDR; x.e(x.deleteQueue('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_connect_derived_queue_hle() {
    callString="$CALLHDR; x.e(x.connectDerivedQueue('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_queue_activate_hle() {
    callString="$CALLHDR; x.e(x.activateQueue('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_queue_deactivate_hle() {
    callString="$CALLHDR; x.e(x.deactivateQueue('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_queue_deactivate_all_queues_hle() {
    callString="$CALLHDR; x.e(x.deactivateAllQueues('$1'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug 
}

function lander_queue_enqueue_local_hle() {
    callString="$CALLHDR; x.e(x.enqueueLocal('$1', '$2', '$3'));"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}


lander_queue_distribute_local_hle() {
    callString="$CALLHDR; x.e(x.distributeLocal('$1', '$2','$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}


#xxx: currently unused, bypassed by lander_queue_size() calling lander_queue_files
function lander_queue_size_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_size('$1', '$2'))"
    _lander_qsize=0
    _lander_qsize=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$_lander_qsize"  # need this printing for Plumb that expects it
    _debug
    return $retCode
}


function lander_queue_files_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_files('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}


function lander_queue_reserve_top_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_reserve_top('$1', '$2', '$3'))"

    reserved_remote=$((python3 -c"$callString")2>$errF)
    retCode=$?
    reserved_local=$2/$(basename "$reserved_remote")
    _debug
    [ "$retCode" != "0" ] && {
        reserved_remote=""
        reserved_local=""
        return 1
    }
    _print_stdout "$retCode" "reserved_local=$reserved_local"
    return $retCode
}


function lander_queue_reserve_top_without_local_copy_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_reserve_top_without_local_copy('$1', '$2'))"
    reserved_local=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$reserved_local" # Plumb reads this output from stdout
    return $retCode
}


function lander_queue_reserve_file_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_reserve_file('$1', '$2', '$3' , '$4'))"
    reserved_local=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$reserved_local"
    return $retCode
}

function lander_queue_unreserve_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_unreserve('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug 
}

function lander_queue_release_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_release('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_queue_release_unconditional_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_release_unconditional('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}

function lander_queue_wipe_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_wipe('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}


##########################################################################################################################

function lander_mark_as_faulty_hle() {
    callString="$CALLHDR; x.e(x.markAsFaulty('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug 
}



function lander_mark_as_healthy_hle() {
    callString="$CALLHDR; x.e(x.markAsHealthy('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}



function lander_queue_files_in_state_x_hle() {
    callString="$CALLHDR; x.e(x.listFilesInStateX('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode     
}


function lander_queue_list_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_list('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode    
}



function lander_queue_du_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_du('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}




function lander_queue_user_du_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_user_du('$1'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}



function lander_queue_queue2user_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_queue2user('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}




function lander_queue_user2queues_hle() {
    callString="$CALLHDR; x.e(x.lander_queue_user2queues('$1'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}



function lander_getEnqueueURI_hle() {
    callString="$CALLHDR; x.e(x.getEnqueueURI('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}




function lander_queue_enqueue_remote_hle() {
    callString="$CALLHDR; x.e(x.enqueue('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}


function lander_copyFromLocalToHDFS_hle() {
    callString="$CALLHDR; x.e(x.copyFromLocalToHDFS('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug

}



lander_getDistributeURI_hle() {
    callString="$CALLHDR; x.e(x.getDistributeURI('$1', '$2','$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}


lander_distribute_hle() {
    callString="$CALLHDR; x.e(x.distribute('$1', '$2','$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    _debug
}


lander_do_Error_Recovery_hle() {
    callString="import top.client_rpc_proxy; from top.client_rpc_proxy import lander_queue_api_client_proxy as x; x.doErrorRecovery('$1', '$2', '$3', '$4', '$5')"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode
}

function lander_queue_peek_next() {
    callString="$CALLHDR; x.e(x.lander_queue_peek_next('$1', '$2'))"

    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    _print_stdout "$retCode" "$retString"
    return $retCode

}


function lander_queue_reserve_next() {
    callString="$CALLHDR; x.e(x.lander_queue_reserve_next('$1', '$2', '$3'))"

    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    return $retCode

}

function lander_queue_list_keys() {
    callString="$CALLHDR; x.e(x.lander_queue_list_keys('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}

function get_value() {
    callString="$CALLHDR; x.e(x.get_value('$1', '$2', '$3'))"
    value=""
    value=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$value"  # need this printing for Plumb that expects it
    _debug
    return $retCode
}

function put_value() {
    callString="$CALLHDR; x.e(x.put_value('$1', '$2', '$3', '$4'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    return $retCode
}

function del_value() {
    callString="$CALLHDR; x.e(x.del_value('$1', '$2', '$3', '$4'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    return $retCode
}


function ls_windowed_queues() {
    callString="$CALLHDR; x.e(x.ls_windowed_queues('$1'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}


function ls_all_window_statuses() {
    callString="$CALLHDR; x.e(x.ls_all_window_statuses('$1'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}


function ls_win() {
    callString="$CALLHDR; x.e(x.ls_win('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}


function ls_win_in_status() {
    callString="$CALLHDR; x.e(x.ls_win_in_status('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}

function set_win_status() {
    callString="$CALLHDR; x.e(x.set_win_status('$1', '$2', '$3', '$4'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    return $retCode
}


function ls_f_in_win() {
    callString="$CALLHDR; x.e(x.ls_f_in_win('$1', '$2', '$3'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}

function mv_f_in_win() {
    callString="$CALLHDR; x.e(x.mv_f_in_win('$1', '$2', '$3', '$4'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _debug
    return $retCode
}


function ls_missing_files_in_a_window() {
    callString="$CALLHDR; x.e(x.ls_missing_files_in_a_window('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}


function ls_site_info_in_a_window() {
    callString="$CALLHDR; x.e(x.ls_site_info_in_a_window('$1', '$2'))"
    retString=$((python3 -c"$callString")2>$errF)
    retCode=$?
    _print_stdout "$retCode"  "$retString"
    _debug
    return $retCode
}

# Some testing code



    # Ordinay client #
    #lander_create_queue_hle "test-q2" "raw"   # call as user that is member of group raw
    #lander_create_queue_hle "test-q3" "raw_funny"  "yuri"  # should fail if done by ordinary user.  Pass if done by admin
    #lander_create_queue_hle "yuri_q1"  "raw" "yuri"   # this shold fail if not called by Admin
    #lander_create_queue_hle "test_q1"  "raw"
    #lander_create_queue_hle "queue_1"  "raw" "nobody"
    #lander_create_queue_hle "q1"  "noop" "somebody"
    #lander_create_queue_hle "s3"  "raw"
    #lander_create_queue_hle "d2"  "raw"  
    #lander_create_queue_hle "q6"  "raw" "user3"
    #lander_create_queue_hle "q8"  "noop" "user4"
    #lander_create_queue_hle "q9"  "full" "user5"

    #lander_delete_queue_hle "test-q2" "aqadeer" 

    #lander_connect_derived_queue_hle "s2" "d2" "aqadeer" 
    #lander_connect_derived_queue_hle "queue_1" "queue_which_dont_exist"
    #lander_connect_derived_queue_hle "q5" "q6" 
    #lander_connect_derived_queue_hle "q4" "q6" 
    #lander_connect_derived_queue_hle "q4" "q7" 
    #lander_connect_derived_queue_hle "q4" "q8" 
    #lander_connect_derived_queue_hle "q7" "q9" 
    
    #lander_queue_activate_hle "queue_1"
    #lander_queue_activate_hle "non_existing_queue"
    #lander_queue_activate_hle "d2"  
    #lander_queue_activate_hle "q5"
    #lander_queue_activate_hle "q6"
    #lander_queue_activate_hle "q7"
    #lander_queue_activate_hle "q8"
    #lander_queue_activate_hle "q9"

    #lander_queue_deactivate_hle "d2" "aqadeer"
    #lander_queue_deactivate_hle "non_existing_queue"


    #directly put a file in a queue
    #lander_queue_enqueue_local_hle  "s2"  "/nfs/ruby/aqadeer/working/traces/faulty-20141117-063342-00358566.erf.sz" "aqadeer"
    #lander_enqueueLocal_hle "d2" "/nfs/ruby/aqadeer/working/traces/20151005-021612-00083343.pcap.xz"
    #lander_enqueueLocal_hle  "s2"  "/nfs/ruby/aqadeer/working/traces/faulty-20141117-063342-00358566.erf.sz"
    #lander_enqueueLocal_hle "pcap_1" "/nfs/ruby/aqadeer/working/traces/20151005-021001-00083337.pcap.xz" "aqadeer"


    # distribute file to connected queues
    #lander_distribute_local_hle  "s2"  "/nfs/ruby/aqadeer/working/traces/20151005-021612-00083343.pcap.xz" "aqadeer"


    # deliver a file into q1
    #lander_distribute_local_hle  "/nfs/ruby/aqadeer/working/traces/20141117-041326-00358331.erf.sz" "should_be_ignored" "q4"
    #lander_distribute_local_hle  "/abc/nfs/ruby/aqadeer/working/traces/faulty-20141117-063342-00358566.erf.sz"  "q7" ""
    #lander_distribute_local_hle  "/nfs/ruby/aqadeer/working/traces/20151005-021612-00083343.pcap.xz" "q1" "" 
    #lander_queue_activate_hle "queue_1"
    #lander_distribute_local_hle  "raw" "/nfs/ruby/aqadeer/working/traces/20141117-065927-00358611.erf.sz" "dummy_from_q" "queue_1"
    #lander_distribute_local_hle  "raw" "/nfs/ruby/aqadeer/working/traces/faulty-20141117-063342-00358566.erf.sz" "dummy_from_q" "queue_1" 
    #lander_distribute_local_hle  "/nfs/ruby/aqadeer/working/traces/20141117-041326-00358331.erf.sz" "should_be_ignored" "q9"


    #lander_queue_size_hle "queue_1"
    #lander_queue_size_hle "queue_2"
    #lander_queue_size_hle "queue_3"
    #lander_queue_size_hle "dummy_q"
    #lander_queue_size_hle "d2"  
    #lander_queue_size_hle "q2"
    #lander_queue_size_hle "q3"
    #lander_queue_size_hle "q4"
    #lander_queue_size_hle "q5"
    #lander_queue_size_hle "q6"
    #lander_queue_size_hle "q7"
    #lander_queue_size_hle "q8"
    #lander_queue_size_hle "q9"
    #lander_queue_size_hle "pcap_1" "aqadeer"


    #lander_queue_files_hle  "rsscint"
    #lander_queue_files_hle  "queue_2"
    #lander_queue_files_hle  "dummy_q"
    #lander_queue_files_hle  "dummy_q"
    #lander_queue_files_hle  "q1"
    #lander_queue_files_hle  "q2"
    #lander_queue_files_hle  "d2"
    #lander_queue_files_hle  "q4"
    #lander_queue_files_hle  "q5"
    #lander_queue_files_hle  "q6"
    #lander_queue_files_hle  "q7"
    #lander_queue_files_hle  "q8"
    #lander_queue_files_hle  "q9"

    #lander_queue_file_older_hle  "non_existing_queue" "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//anonymized/20141117-065927-00358611.erf.sz" "30"
    #lander_queue_file_older_hle  "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz" "30"
    #lander_queue_file_older_hle  "q9" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz" "30"
    #lander_queue_file_older_hle  "q9" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz" "999999999999999999"
    #lander_queue_file_older_hle  "queue_1" "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//anonymized/20141117-065927-00358611.erf.sz" "30"
    #lander_queue_file_older_hle  "queue_2" "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//raw/20141117-065927-00358611.erf.sz" "1"

    #lander_queue_reserve_top_hle "d2" "/tmp/" 
    #lander_queue_reserve_top_hle "queue_1" "/tmp/non_existing_path/"
    #lander_queue_reserve_top_hle "q9" "/tmp/"
    #lander_queue_reserve_top_hle "q9" "/tmp/"
    #lander_queue_reserve_top_hle "dummy_q" "/tmp/"

    #lander_queue_reserve_file_hle "q2" "/tmp/"  "/xyz//tmp/20141118-000002-00360582.erf"
    #lander_queue_reserve_file_hle "non_existing_queue" "/tmp/"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz"
    #lander_queue_reserve_file_hle "queue_1" "/tmp/non_existing_path/" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz"
    #lander_queue_reserve_file_hle "d2" "/tmp/" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz"
    #lander_queue_reserve_file_hle "q9" "/tmp/" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz"

    #lander_queue_unreserve_hle "queue_test" "/tmp/t/bar"
    #lander_queue_unreserve_hle  "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz"
    #lander_queue_unreserve_hle  "no_existing_queu" "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//anonymized/20141117-065927-00358611.erf.sz"
    #lander_queue_unreserve_hle  "q9" "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//raw/20141117-065927-00358611.erf.szzzzzzzzzz"
    #lander_queue_unreserve_hle  "q9" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//anonymized/faulty-20141117-063342-00358566.erf.sz"
    #lander_queue_unreserve_hle  "q9" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz"


    #lander_queue_release_hle  "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz"
    #lander_queue_release_hle  "non_existing_queue"  "hdfs://hcpu050.ant.isi.edu:9000/user/aqadeer/db2//raw/20141117-065927-00358611.erf.sz"
    #lander_queue_release_hle  "q9"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//full/20141117-041326-00358331.erf.sz"
    #lander_queue_release_hle  "q8"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//noop/20141117-041326-00358331.erf.sz"
    #lander_queue_release_hle  "q8"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//noop/20141117-041326-00358331.erf.sz"

    #lander_queue_wipe_hle "non_existing_q"
    #lander_queue_wipe_hle "d2"
    #lander_queue_wipe_hle "q7"
    #lander_queue_wipe_hle "q9"

    #lander_create_ordinary_db_user_hle "somebody"
    #lander_delete_ordinary_db_user_hle "somebody"
    #lander_delete_ordinary_db_user_hle "abc"

    #lander_mark_as_faulty_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz"
    #lander_mark_as_faulty_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz" "yuri"


    #lander_mark_as_healthy_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz"
    #lander_mark_as_healthy_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/20151005-021612-00083343.pcap.xz" "yuri"


    #lander_queue_files_in_state_x_hle "d2" "UNRESERVED"
    #lander_queue_files_in_state_x_hle "d2" "RESERVED" "yuri"


    #lander_queue_list_hle "queue_3" "anonymized" "yuri"
    #lander_queue_list_hle "d2" "anonymized" 

    #lander_queue_du_hle "d2"

    #lander_queue_user_du_hle 
    #lander_queue_user_du_hle "yuri"


    #lander_queue_queue2user_hle a2

    #lander_queue_user2queues_hle 
    #lander_queue_user2queues_hle "aqadeer"



    # Group of 3
    #
    #lander_getEnqueueURI_hle "d2" "/someting/abc.txt"
    
    #lander_copyFromLocalToHDFS_hle "/nfs/ruby/aqadeer/working/traces/20151005-021612-00083343.pcap.xz"  \
    #                               "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/abc.txt"

    #lander_enqueue_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/abc.txt"



    # Group of 3
    #
    #lander_getDistributeURI_hle "d2" "/something/def.txt"

    #lander_copyFromLocalToHDFS_hle "/nfs/ruby/aqadeer/working/traces/20151005-021612-00083343.pcap.xz"  \
    #                               "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/conqdata/def.txt"



    #lander_distribute_hle "d2" "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/d2/conqdata/def.txt"

    #lander_do_Error_Recovery_hle "rsscint"  30  2  "aqadeer"
    #lander_queue_size_hle "pcap.xz_in"  "aaqadeer"
    #lander_queue_reserve_top_without_local_copy_hle "pcap.xz_in" "aqadeer"
    #lander_getEnqueueURI_hle "pcap.xz_in"  "afile"  "aaqadeer"
    #lander_getDistributeURI_hle "pcap.xz_inn" "afile" "aqadeer"
    #lander_queue_create_hle "b" "raw" "aaqadeer"
    #lander_connect_derived_queue_hle "a" "bbbbb" "aqadeer"
    #lander_queue_activate_hle "a" "aaqadeer"
    #lander_queue_release_hle "pcap.xz_in"  "20160622-000137-00354295.pcap.xz" "aqadeer"
    #lander_queue_enqueue_remote_hle "pcap.xz_in"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap.xz_in/20160622-000137-00354295.pcap.xz" "aqadeer"
    #lander_distribute_hle "pcap.xz_in"  "hdfs://hcpu001.ant.isi.edu:9000/user/aqadeer/db2//raw/pcap.xz_in/20160622-000137-00354295.pcap.xz" "aqadeer"



    # Test
    #lander_queue_list_derived_queues_hle "s1" "yuri"
    #lander_queue_list_derived_queues_hle "a" "aqadeer"
    #lander_queue_create_hle "c" "raw" "aqadeer"
    #lander_queue_create_hle "a" "raw" "aqadeer"
    #lander_queue_delete_hle "a" "aqadeer"
    #lander_connect_derived_queue_hle "a" "c" "aqadeer"
    #lander_queue_activate_hle "b"  "aqadeer"
    #lander_queue_deactivate_hle "a" "aqadeer"
    #lander_queue_enqueue_local_hle  "a"  "/nfs/hle/for_Testing_only/ANT_GIT/CODE/lander_hard_link_emulation/wip/shell/20141117-063342-00358567.lax.pcap.sz" "aqadeer"
    #lander_queue_distribute_local_hle  "a"  "/nfs/hle/for_Testing_only/ANT_GIT/CODE/lander_hard_link_emulation/wip/shell/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_size_hle "c" "aqadeer"
    #lander_queue_files_hle "a" "aqadeer"
    #lander_queue_reserve_top_hle "a" "/tmp/" "aqadeer"
    #lander_queue_reserve_top_without_local_copy_hle "a" "aqadeer"
    #lander_queue_reserve_file_hle "a" "/tmp/"  "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_unreserve_hle "a" "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_release_hle  "a" "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_release_unconditional_hle "a" "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_wipe_hle "a" "aqadeer"
    #lander_mark_as_faulty_hle "a" "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_mark_as_healthy_hle "a" "/user/plumb/hle/raw/a/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_queue_files_in_state_x_hle "a" "UNRESERVED" "aqadeer"
    #lander_queue_list_hle "x" "raw" "aqadeer"
    #lander_queue_du_hle "c" "aqadeer"
    #lander_queue_user_du_hle "yuri"
    #lander_queue_queue2user_hle a "dummy"
    #lander_queue_user2queues_hle "aqadeer"
    #lander_getEnqueueURI_hle "b"  "20141117-063342-00358566.lax.pcap.sz"  "aqadeer"
    #lander_copyFromLocalToHDFS_hle "/nfs/hle/for_Testing_only/ANT_GIT/CODE/lander_hard_link_emulation/wip/shell/20141117-063342-00358566.lax.pcap.sz"  "/user/plumb/hle/raw/b/20141117-063342-00358566.lax.pcap.sz"
    # Following call assumes that files has already been placed in hdfs
    #lander_queue_enqueue_remote_hle "b"  "/user/plumb/hle/raw/b/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_getDistributeURI_hle "a" "/user/plumb/hle/raw/b/20141117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_copyFromLocalToHDFS_hle "/nfs/hle/for_Testing_only/ANT_GIT/CODE/lander_hard_link_emulation/wip/shell/20141117-063342-00358566.lax.pcap.sz"  "/user/plumb/hle/raw/a/conqdata/20181117-063342-00358566.lax.pcap.sz"
    #lander_distribute_hle "a"  "/user/plumb/hle/raw/a/conqdata/20181117-063342-00358566.lax.pcap.sz" "aqadeer"
    #lander_do_Error_Recovery_hle "b"  30  2  "aqadeer"

    # Tests for stateful-streaming
    #x=$(lander_queue_peek_next "pcap.sz_1" "hlander")
    #if [ $? == 0 ];
    #then
    #    echo $x
    #    read -ra arr <<<"$x"
    #    lander_queue_reserve_next "pcap.sz_1" "${arr[3]}" "hlander"
    #fi
    #put_value "pcap.sz_1" "key_k1" "k1->v1" "hlander"
    #lander_queue_list_keys "pcap.sz_1" "hlander"
    #get_value "pcap.sz_1" "key_k1" "hlander"
    #del_value "pcap.sz_1" "key_k1" "k1->v1" "hlander"

    # for windows
    # ls_windowed_queues "hlander"


