#!/usr/bin/env python3
import inspect
import os
import random
import sys

import MySQLdb

_DEBUG_THIS = True


def _debug(s):
  if _DEBUG_THIS:
    sys.stderr.write("DEBUG: " + str(s) + "\n")

# Following function exists because I can't call the same names function that starts with two underscores
# from a class (because python adds _classname before calling function and hence fails)
# For example see here: https://stackoverflow.com/questions/52853615/how-to-call-global-function-from-class-method
def connectToDB(Host, User, Pass, DB):
    return __connectToDB(Host, User, Pass, DB)


def __connectToDB(Host, User, Pass, DB):
    try:
        dbcon = MySQLdb.connect(host=Host,
                                user=User,
                                passwd=Pass,
                                db=DB)
        return dbcon


    except MySQLdb.Error as e:
        _debug(e)
        return None


def __disconnectFromDB(dbcon):
    if (dbcon is not None):
        dbcon.close()
        return True

    return False


def __cleanup(dbcon, cur):
    if cur is not None:
        cur.close()
    __disconnectFromDB(dbcon)


def __populateKeyValTable(dbcon, cur, key, val):
    query = 'INSERT INTO KeyVal (k, v) Values ("%s", "%s") ON DUPLICATE KEY UPDATE v="%s";' % (key, val, val)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __populateSubscriberKeyValTable(dbcon, cur, tableName, key, val, doCommit = False):
    query = 'INSERT INTO `%s` (k, v) Values ("%s", "%s") ON DUPLICATE KEY UPDATE v="%s";' % (tableName, key, val, val)
    _debug(query)

    try:
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __populateSubscriberKeyValTableWrapper(dbcon, cur, queue_owner, queue_id, key, val, doCommit = False):
    table_name = __getDBLevelUserKeyValueTableName(queue_owner, queue_id)
    return __populateSubscriberKeyValTable(dbcon, cur, table_name, key, val, doCommit)



def __initializeSubscriberKeyValTable(dbcon, cur, tableName, doCommit = False):
    retCode = 0
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "isActive", "0", doCommit = False)
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "isWindow", "0", doCommit = False)
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "deliveredUpToBlockNo", "-1", doCommit = False)
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "canDeliverUpToBlockNo", "-1", doCommit = False)
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "remainingWaitBudget", "0", doCommit = False)
    retCode += __populateSubscriberKeyValTable(dbcon, cur, tableName, "log", "", doCommit = False)

    return retCode



def __getValFromKeyValTable(dbcon, cur, key):
    query = 'Select v FROM KeyVal WHERE k="%s";' % (key)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return 201



def __doesQueueExists(dbcon, cur, queue_id):
    retVal = 0
    query = "SELECT count(*) FROM Queue WHERE queue_id='%s';"  %(queue_id)
    _debug(query)

    try:
        cur.execute(query)
        count = cur.fetchone()
        if count == None or cur.rowcount == 0 or count[0] == 0:
            return 1
        else:
            return 0

    except MySQLdb.Error as e:
        _debug(e)
        retVal = 202


def __populateQueueTable(dbcon, cur, queue_id, queue_type):
    query = 'INSERT INTO Queue (queue_id, queue_type) Values ("%s", "%s");' % (queue_id, queue_type)
    _debug(query)


    try:
        cur.execute(query)
        dbcon.commit()

    except MySQLdb.Error as e:
        _debug(e)
        return 203


    return 0



def __getQueueSize(dbcon, cur, queue_id, queue_owner, state):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = ("SELECT count(*) as c FROM `%s` WHERE FileStateName = '%s';") %(queueTableName, state)
    _debug(query)

    try:
        cur.execute(query)
        cRow = cur.fetchone()
        if cRow == None:
            return (1, None)
        else:
            return (0, cRow[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (200, None)


def __populateSubscriberTable(dbcon, cur, subscriberName, queue_id, queue_type, owner):
    query = "Insert into Subscriber (SubscriberName, queue_id, queue_type, owner) \
                             Values ('%s', '%s', '%s', '%s');"                                  \
                              %(subscriberName, queue_id, queue_type, owner)
    try:
        _debug(query)
        cur.execute(query)
        dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __createSubscriberKeyValTable_xxx(dbcon, cur, userName, queue_id):
    tableName = "%s_%s_%s"  %(userName, queue_id, "KeyVal")

    query = ("CREATE TABLE IF NOT EXISTS `%s`(k VARCHAR(255) NOT NULL UNIQUE, \
                                              v VARCHAR(1024) NOT NULL,       \
                                                 PRIMARY KEY (k)) ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;" \
            ) %(tableName)


    try:
        _debug(query)
        cur.execute(query)
        dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __createSubscriberKeyValTable(dbcon, cur, tableName, doCommit = False):

    query = ("CREATE TABLE IF NOT EXISTS `%s`(k VARCHAR(255) NOT NULL UNIQUE, \
                                              v VARCHAR(1024) NOT NULL,       \
                                                 PRIMARY KEY (k)) ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;" \
            ) %(tableName)


    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __createSubscriberQueueTable(dbcon, cur, tableName, doCommit = False):

    # ToDo: Isn't FileModTime same as EventTime?  Refactor and remove one of them after careful code review.
    # Note: I would have liked DataURI to be 1024 but unfortunately current DB engine doesn't allow index fields
    # bigger than 767.  It looks like we can get rid of this constraint by moving to next version of DB.
    # Note 2: In the following some filed lengths are 767 that are constituent parts of DataURI.  But since Yuri
    # didn't want to restrict these constituent parts max size, now if overall DataURI size increases more than 767,
    # data will not load.
    # Note: New entry will fail in table if collectively (SequenceNo, SequenceNoInstance) are not unique.
    query = ("CREATE TABLE IF NOT EXISTS `%s` (DataURI VARCHAR(767) NOT NULL, \
                                            FileStateName VARCHAR(20) NOT NULL,\
                                            RowEntryEpoch TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP, \
                                            FileModTime   TIMESTAMP NOT NULL, \
                                            Retries TINYINT UNSIGNED NOT NULL DEFAULT 0, \
                                            EventTime   DateTime NOT NULL, \
                                            SequenceNo BIGINT UNSIGNED NOT NULL,\
                                            AdditionalInfo VARCHAR(767), \
                                            SequenceNoInstance VARCHAR(767) NOT NULL, \
                                            QueueID VARCHAR(255)  NOT NULL          ,\
                                            WindowNo INTEGER UNSIGNED,\
                                            PRIMARY KEY (DataURI),\
                                            FOREIGN KEY (DataURI) REFERENCES Data(DataURI),\
                                            FOREIGN KEY (FileStateName) REFERENCES FileState(StateName), \
                                            FOREIGN KEY (QueueID) REFERENCES Queue(queue_id), \
                                            UNIQUE(SequenceNo, SequenceNoInstance)) \
                                            ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;"
            ) % tableName


    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __setSubscriberPrivileges(dbcon, cur, db_user, tables, doCommit = False):
    #tablestring = ",".join(tables)

    userstring = "'" + db_user +"'" + "@'%'"

    for table in tables:
        query = "grant SELECT, \
                       INSERT, \
                       DELETE, \
                       UPDATE  \
                       on %s to %s ;" %(table, userstring)
        try:
            _debug(query)
            cur.execute(query)
            if doCommit: dbcon.commit()
        except MySQLdb.Error as e:
            _debug(e)
            return 200


    # __addQueue needs additional select grants for user
    moreTables = ["KeyVal", "Queue"]
    return __setAdditionalSubscriberPrivileges(dbcon, cur, db_user, moreTables, doCommit)



# Following function added to fix a BUG on release queue function

def __setAdditionalSubscriberPrivileges(dbcon, cur, db_user, tables, doCommit = False):
    userstring = "'" + db_user +"'" + "@'%'"

    for table in tables:
        query = "grant SELECT  \
                       on %s to %s ;" %(table, userstring)
        try:
            _debug(query)
            cur.execute(query)
            if doCommit: dbcon.commit()
        except MySQLdb.Error as e:
            _debug(e)
            return 300

    return 0


def __createDBUser(dbcon, cur, db_user, db_pass, doCommit = False):
    # Delete user if it already exists.
    __deleteDBUser(dbcon, cur, db_user, doCommit)

    query = "create user '%s'  identified by '%s';" %(db_user, db_pass)
    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __deleteDBTable(dbcon, cur, db_table, doCommit = False):
    query = "drop table  `%s`;" %(db_table)
    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __deleteDBUser(dbcon, cur, db_user, doCommit = False):
    query = "drop user  `%s`;" %(db_user)


    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200


def __addSubscriber(dbcon, cur, userName, queue_id, queue_type):
    db_user = "%s_%s" %(userName, queue_id)
    db_pass = random.getrandbits(30)
    keyValTableName = "%s_%s_%s"  %(userName, queue_id, "KeyVal")
    key = "isActive"
    val = "0"

    queueTableName = "%s_%s_%s"  %(userName, queue_id, "Queue")
    tables = [keyValTableName, queueTableName]

    _debug("%s has passowrd %s" %(db_user, db_pass))

    result = 0
    result += __populateSubscriberTable(dbcon, cur, userName, queue_id, queue_type, db_user)
    result += __createSubscriberKeyValTable(dbcon, cur, keyValTableName)
    result += __populateSubscriberKeyValTable(dbcon, cur, keyValTableName, key, val)
    result += __createSubscriberQueueTable(dbcon, cur, queueTableName)
    result += __createDBUser(dbcon, cur, db_user, db_pass)
    result += __setSubscriberPrivileges(dbcon, cur, db_user, tables)

    return result


def __getSubsriberName(dbcon, cur, queue_id):
    query = "SELECT SubscriberName FROM Subscriber WHERE queue_id='%s';" % (queue_id)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __getQueueOwner(dbcon, cur, queue_id):
    query = "SELECT owner FROM Queue WHERE queue_id='%s';" % (queue_id)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __getValFromSubscriberKeyValTable(dbcon, cur, queue_id, key):
    tableName = "`%s_%s`" %(queue_id, "KeyVal")
    query = "Select v FROM %s WHERE k='%s';" % (tableName, key)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)

def __getValFromSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key):
    tableName = __getDBLevelUserKeyValueTableName(queue_owner, queue_id)
    query = "Select v FROM `%s` WHERE k='%s';" % (tableName, key)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __doesKeyExistsInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key):
    tableName = __getDBLevelUserKeyValueTableName(queue_owner, queue_id)
    query = "Select count(*) FROM `%s` WHERE k='%s';" % (tableName, key)
    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None or len(value) == 0 or value[0] == 0:
            return False
        else:
            return True
    except MySQLdb.Error as e:
        _debug(e)
        return True


def __setValInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key, val):
    tableName = __getDBLevelUserKeyValueTableName(queue_owner, queue_id)
    query = "UPDATE `%s` SET v='%s' where k='%s';" % (tableName, val, key)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 202


def __setValInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, fileStateName, dataURI):
    tableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "UPDATE `%s` SET FileStateName='%s' , RowEntryEpoch=CURRENT_TIMESTAMP  where DataURI='%s';" % (tableName, fileStateName, dataURI)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 202


def __getListOfActiveSubscribers(dbcon, cur, data_type):
    query = "SELECT SubscriberName, queue_id FROM Subscriber WHERE queue_type='%s';" % (data_type)
    _debug(query)

    resultArray = []

    try:
        cur.execute(query)
        subscriberArray = cur.fetchall()

        for userName, queue_id in subscriberArray:
            result, isActive = __getValFromSubscriberKeyValTable(dbcon, cur, queue_id, userName, "isActive")
            if result == 0 and isActive == "1":
                resultArray.append(queue_id)

        return resultArray
    except MySQLdb.Error as e:
        _debug(e)
        return None



def __populateQueueTable(dbcon, cur, queue_id, data_type, db_user, doCommit = False):
    query = 'INSERT INTO Queue (queue_id, data_type, owner) Values ("%s", "%s", "%s");' % (queue_id, data_type, db_user)
    _debug(query)

    try:
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 200



def __addQueue(dbcon, cur, queue_id, db_user, doCommit = False):
    keyValTableName = __getDBLevelUserKeyValueTableName(db_user, queue_id)
    key = "isActive"
    val = "0"

    queueTableName = __getDBLevelUserQueueTableName(db_user, queue_id)
    tables = [keyValTableName, queueTableName]

    result = 0
    result += __createSubscriberKeyValTable(dbcon, cur, keyValTableName)
    if result != 0: return result
    result += __populateSubscriberKeyValTable(dbcon, cur, keyValTableName, key, val)
    if result != 0: return result
    result += __createSubscriberQueueTable(dbcon, cur, queueTableName)
    if result != 0: return result
    # __createDBUser is now separetely called by admin
    #result += __createDBUser(dbcon, cur, db_user, db_pass)
    #if result != 0: return result
    # User is given restricted access to 2 tables s/he needs.  So even if s/he tries
    # to connect to the db directly, s/he only sees her/his stuff only.

    # end users now can't touch DB.  So no need for following.
    #result += __setSubscriberPrivileges(dbcon, cur, db_user, tables)
    #if result != 0: return result

    if doCommit: dbcon.commit()

    return result



def __UpdateQueueParentsTable(dbcon, cur, from_queue_id, to_queue_id):

    query = "INSERT INTO QueueParents (queue_id, from_queue_id) VALUES ('%s', '%s');" % (to_queue_id, from_queue_id)

    _debug(query)


    try:
        cur.execute(query)
        dbcon.commit()
        return 0

    except MySQLdb.Error as e:
        _debug(e)
        return 203


def __DisconnectUserQueue(dbcon, cur, queue_id):
    #There are 2 entanglements.  First in queue parents table. Second in Data table.
    query = "DELETE FROM QueueParents WHERE queue_id = '%s' OR from_queue_id = '%s';"  %(queue_id, queue_id)
    _debug(query)
    try:
        cur.execute(query)
    except MySQLdb.Error as e:
        _debug(e)
        return 203

    query = "UPDATE Data SET from_queue_id=NULL WHERE from_queue_id='%s';" % (queue_id)

    _debug(query)
    try:
        cur.execute(query)
        #dbcon.commit()
        return 0

    except MySQLdb.Error as e:
        _debug(e)
        return 204


def __SearchQueues_v1(dbcon, cur, from_queue_id, data_type):
    query = "SELECT queue_id from Queue WHERE from_queue_id='%s' AND data_type='%s';"  %(from_queue_id, data_type)

    try:
        cur.execute(query)
        queues = cur.fetchall()

        if queues == None:
            return (1, None)
        else:
            return (0, queues)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __SearchQueues(dbcon, cur, from_queue_id):
    query = "SELECT queue_id from Queue WHERE from_queue_id='%s';"  %(from_queue_id)

    try:
        cur.execute(query)
        queues = cur.fetchall()

        if queues == None:
            return (1, None)
        else:
            return (0, queues)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)



def __getQueuesByOwner(dbcon, cur, queue_owner):
    query = "SELECT queue_id FROM Queue WHERE owner='%s';"  %(queue_owner)

    try:
        cur.execute(query)
        queues = cur.fetchall()

        if queues == None:
            return (1, None)
        else:
            queues_array = [q[0] for q in queues]
            return (0, queues_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __findDerivedQueues(dbcon, cur, from_queue_id):
    query = "SELECT DISTINCT queue_id FROM QueueParents WHERE from_queue_id='%s';"  %(from_queue_id)

    try:
        cur.execute(query)
        queues = cur.fetchall()

        if queues == None:
            return (1, None)
        else:
            queues_array = [f[0] for f in queues]
            return (0, queues_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __ReverseSearchQueues(dbcon, cur, from_queue_id):
    query = "SELECT from_queue_id from Queue WHERE queue_id='%s';"  %(from_queue_id)
    _debug(query)

    try:
        cur.execute(query)
        queue = cur.fetchone()

        if queue == None:
            return (1, None)
        else:
            return (0, queue[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)




def __populateDataTable(dbcon, cur, dataURI, from_queue_id, shouldCommit):
    query = 'INSERT IGNORE INTO Data (DataURI, from_queue_id) Values \
            ("%s", "%s");' % (dataURI, from_queue_id)
    _debug(query)
    try:
        cur.execute(query)
        if shouldCommit == True:
            dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 205


def __populateSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, dataURI, fileState, modTime,
                                   eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance,
                                   tailQID, shouldCommit):
    # sequence no + sequence instance should be unique.  Catch if not unique.  We need to catch this condition here
    # because we want to ignore the case when exact same row is tried to insert again.
    retCode, result = __detectDuplicateInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, sequence_no,
                                                              sequence_no_instance, dataURI)
    if retCode == 0 and result:  # uniqueness violated
        return 207

    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = 'INSERT IGNORE INTO `%s` (DataURI, FileStateName, FileModTime, EventTime, SequenceNo, AdditionalInfo, SequenceNoInstance, QueueID) Values \
            ("%s", "%s", FROM_UNIXTIME(%d), STR_TO_DATE("%s", "%s" ), CAST("%s" as UNSIGNED), "%s", "%s", "%s");' \
            % (queueTableName, dataURI, fileState, int(modTime)/1000, eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID)


    _debug(query)

    try:
        cur.execute(query)
        if shouldCommit == True:
            dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 206


def __getFileListFromSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, file_state):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "SELECT DataURI from `%s` WHERE FileStateName='%s' order by FileModTime ;"  %(queueTableName, file_state)
    files_array = []

    try:
        cur.execute(query)
        files = cur.fetchall()



        if files == None:
            return (1, files_array)
        elif cur.rowcount == 0:
            return (0, files_array)
        else:
            files_array = [f[0] for f in files]
            return (0, files_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, files_array )


def __getModTimeForFileInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfs_file_uri):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "SELECT UNIX_TIMESTAMP(FileModTime) from `%s` WHERE DataURI='%s';"  %(queueTableName, hdfs_file_uri)
    _debug(query)
    try:
        cur.execute(query)
        epoch = cur.fetchone()
        if epoch == None:
            return (1, None)
        else:
            return (0, epoch[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __doUnreserveOperationInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfsURI):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "UPDATE `%s` SET FileStateName='UNRESERVED', RowEntryEpoch=CURRENT_TIMESTAMP  WHERE DataURI='%s' AND FileStateName='RESERVED';" %(queueTableName, hdfsURI)
    _debug(query)
    try:
        cur.execute(query)
        if cur.rowcount == 1:
            dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __doReleaseOperationInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfsURI):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "DELETE from `%s` WHERE DataURI='%s' AND FileStateName='RESERVED';" %(queueTableName, hdfsURI)
    _debug(query)

    try:
        cur.execute(query)
        if cur.rowcount == 1:
            dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __deleteFileFromQueueUnConditional(dbcon, cur, queue_owner, queue_id, hdfsURI):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "DELETE from `%s` WHERE DataURI='%s';" %(queueTableName, hdfsURI)
    _debug(query)
    try:
        cur.execute(query)
        if cur.rowcount == 1:
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __doWipeOperationInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "DELETE from `%s`;" %(queueTableName)
    _debug(query)

    try:
        cur.execute(query)
        if cur.rowcount >= 1:
            dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __findUnreservedURIWithOldestModTimeInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "SELECT DataURI from `%s` WHERE FileStateName='UNRESERVED' ORDER BY FileModTime LIMIT 1;" %(queueTableName)
    _debug(query)
    try:
        cur.execute(query)
        dataURI = cur.fetchone()
        if dataURI == None:
            return (1, None)
        else:
            return (0, dataURI[0])
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


def __doAtomicReserveInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfsURI, doCommit = True):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "UPDATE `%s` SET FileStateName='RESERVED', RowEntryEpoch=CURRENT_TIMESTAMP  WHERE DataURI='%s' AND FileStateName='UNRESERVED';" %(queueTableName, hdfsURI)
    _debug(query)

    try:
        cur.execute(query)
        if cur.rowcount == 1:
            if doCommit: dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __getFileListFromDataTable(dbcon, cur):
    query = "SELECT DataURI from `Data`;"
    files_array = []

    try:
        cur.execute(query)
        files = cur.fetchall()
        if files == None:
            return (1, files_array)
        elif cur.rowcount == 0:
            return (0, files_array)
        else:
            files_array = [f[0] for f in files]
            return (0, files_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, files_array )


def __getFileListFromSubscriberQueue(dbcon, cur, queue_owner, queue_id):
    query = "SELECT DataURI from `%s`;"  %(__getDBLevelUserQueueTableName(queue_owner, queue_id))
    files_array = []

    try:
        cur.execute(query)
        files = cur.fetchall()
        if files == None:
            return (1, files_array)
        elif cur.rowcount == 0:
            return (0, files_array)
        else:
            files_array = [f[0] for f in files]
            return (0, files_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, files_array )


def __deleteRowFromDataTable(dbcon, cur, dataURI, doCommit=True):
    query = "DELETE FROM Data WHERE DataURI='%s';" %(dataURI)
    _debug(query)
    try:
        cur.execute(query)
        if cur.rowcount >= 1:
            if doCommit: dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __deleteRowFromQueueTable(dbcon, cur, queue_id, doCommit=True):
    query = "DELETE FROM Queue WHERE queue_id='%s';" %(queue_id)
    _debug(query)
    try:
        cur.execute(query)
        if cur.rowcount >= 1:
            if doCommit: dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __getDataTypeFromQueueTable(dbcon, cur, queueID):
    query = "SELECT data_type FROM Queue WHERE queue_id='%s';" %(queueID)
    _debug(query)
    try:
        cur.execute(query)
        value = cur.fetchone()
        if value == None:
            return (1, value)
        else:
            return (0, value[0])
    except MySQLdb.Error as e:
        _debug(e)
        return 201


def __doesFileExistsInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, hdfsURI):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    # Remember in Mysql back quote i.e. ` is used with variable names
    # while ' with the value.  In the following queueTableName is variable
    # while hdfsURI is used as a value
    query = "SELECT DataURI FROM `%s` WHERE DataURI='%s';" %(queueTableName, hdfsURI)
    _debug(query)
    try:
        cur.execute(query)
        if cur.rowcount == 1:
            dbcon.commit()
            return 0
        else:
            return 1
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __makeHDFSURI(dbcon, cur, queue_id, file_uri):
    hdfsURI = ""
    isSuccess, hdfsBaseDir = __getValFromKeyValTable(dbcon, cur, "HDFS_Base_Dir")
    if isSuccess != 0: return (4, hdfsURI)

    isSuccess, from_queue_id= __ReverseSearchQueues(dbcon, cur, queue_id)
    if isSuccess != 0: return (5, hdfsURI)

    q = from_queue_id
    _debug("q=%s end queue_id=%s" %(q, queue_id))
    if q is None:
        q = queue_id
        _debug("q = %s" %(q))


    # find data_type associated with the to_queue_id
    isSuccess, data_type = __getDataTypeFromQueueTable(dbcon, cur, q)
    if isSuccess != 0: return (6, hdfsURI)

    # example: hdfs://hcpu001.ant.isi.edu:9000/tmp/hle//noop/20151005-021612-00083343.pcap.xz
    hdfsURI = "%s/%s/%s"  %(hdfsBaseDir, data_type, os.path.basename(file_uri))
    _debug(hdfsURI)

    return (0,hdfsURI)


def __lockTables(dbcon, cur, tables):
    tabString = ""
    for k,v in tables.items():
        tabString += "%s %s," %(k, v)

    tabString = tabString[:-1] # get rid of last comma

    query = "%s %s %s"  %("LOCK TABLES", tabString, ";")
    _debug(query)
    try:
        cur.execute(query)
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 1


def __ublockTables(dbcon, cur):
    query = "UNLOCK TABLES;"
    _debug(query)
    try:
        cur.execute(query)
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 1


# Some helper functions for easy management of code change

def __getDBLevelUserQueueTableName(user_id, queue_id):
    queueTableName = "%s_%s_%s"  %(user_id, queue_id, "Queue")
    return queueTableName


def __getDBLevelUserKeyValueTableName(user_id, queue_id):
    keyValTableName = "%s_%s_%s"  %(user_id, queue_id, "KeyVal")
    return keyValTableName


def __doesFileStateExists(dbcon, cur, file_state):
    query = "SELECT StateName FROM FileState WHERE StateName='%s';"  %(file_state)
    _debug(query)
    try:
        cur.execute(query)
        fileState = cur.fetchall()

        if fileState == None:
            return 1
        else:
            return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 201


def __listQueues(dbcon, cur, queue_owner=None, data_type=None):
    if data_type != None:
        query = "SELECT queue_id FROM Queue WHERE owner='%s' AND data_type='%s';"  %(queue_owner, data_type)
    elif queue_owner != None:
        query = "SELECT queue_id FROM Queue WHERE owner='%s';"  %(queue_owner)
    else:
        query = "SELECT queue_id FROM Queue;"

    _debug(query)
    try:
        cur.execute(query)
        queues = cur.fetchall()
        if queues == None:
            return (1, None)
        else:
            queues_array = [f[0] for f in queues]
            return (0, queues_array)
    except MySQLdb.Error as e:
        _debug(e)
        return (201, None)


#what HDFS URI will be if filename was placed in the queue_id
#returns (status, hdfs_uri, error)
#ToDo: Following function returns wrong hdfsURI when queue itself is one of the many childs.
# Example: We should have got: /user/plumb/hle/raw/pcap.sz/conqdata/20181023-233950-01207968.lax.pcap.sz instead of
# /user/plumb/hle/raw/pcap.sz_1/20181023-233950-01207968.lax.pcap.sz
# For now I am just returning filename as hdfsURI.  That is fine for Plumb code but it might not be okay
# for lander code.  Yuri needs to be consulted on this item.
#
def __getHdfsURI(dbcon, cur, queue_id, filename):
    return (0, filename, "")

    if filename.startswith("hdfs://"):
        #already good
        return (0, filename, "")
    # find what is HDFS base dir
    isSuccess, hdfsBaseDir = __getValFromKeyValTable(dbcon, cur, "HDFS_Base_Dir")
    hdfsURI = ""
    data_type = ""
    if isSuccess != 0:
        return (1, "", "Failed to get HDFS_Base_Dir from Db.  This failure is in %s "
                      %(__name__))
    # find data_type associated with the queue_id
    isSuccess, data_type = __getDataTypeFromQueueTable(dbcon, cur, queue_id)
    if isSuccess != 0:
        return (2, "", "Failed to get data_type associated with queue %s.  This failure is in %s "
                         %(queue_id, __name__))
    fileHdfsURI =  "%s/%s/%s/%s" %(hdfsBaseDir, data_type, queue_id, os.path.basename(filename))
    return (0, fileHdfsURI, "")





def __toggleStateConditional(dbcon, cur, queue_id, queue_owner, how_old, max_fault_count, doCommit = False):

    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("UPDATE  `%s` \
              SET FileStateName='UNRESERVED', RowEntryEpoch=CURRENT_TIMESTAMP, Retries=Retries+1 \
              WHERE FileStateName='RESERVED' \
              AND  CURRENT_TIMESTAMP > RowEntryEpoch + INTERVAL %d MINUTE \
              AND Retries < %d;"
            ) %(queueTableName, how_old, max_fault_count)

    searchQuery = ("SELECT count(*) FROM `%s` \
                    WHERE FileStateName='RESERVED' \
                    AND  CURRENT_TIMESTAMP > RowEntryEpoch + INTERVAL %d MINUTE \
                    AND Retries < %d;"
                  ) %(queueTableName, how_old, max_fault_count)


    try:
        _debug(searchQuery)
        cur.execute(searchQuery)
        count = cur.fetchone()
        if count == None or cur.rowcount == 0:
            return 1, 0

        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0, count[0]
    except MySQLdb.Error as e:
        _debug(e)
        return 500, 0






def __markFaultyConditional(dbcon, cur, queue_id, queue_owner, how_old, max_fault_count, doCommit = False):

    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("UPDATE  `%s` \
              SET FileStateName='FAULTY',  RowEntryEpoch=CURRENT_TIMESTAMP \
              WHERE FileStateName='RESERVED' \
              AND  CURRENT_TIMESTAMP > RowEntryEpoch + INTERVAL %d MINUTE \
              AND Retries = %d;"
            ) %(queueTableName, how_old, max_fault_count)


    searchQuery = ("SELECT count(*) FROM  `%s` \
                    WHERE FileStateName='RESERVED' \
                    AND  CURRENT_TIMESTAMP > RowEntryEpoch + INTERVAL %d MINUTE \
                    AND Retries = %d;"
                  ) %(queueTableName, how_old, max_fault_count)


    try:

        _debug(searchQuery)
        cur.execute(searchQuery)
        count = cur.fetchone()
        if count == None or cur.rowcount == 0:
            return 1, 0

        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0, count[0]
    except MySQLdb.Error as e:
        _debug(e)
        return 501, 0



def __disableAllQueues(dbcon, cur, doCommit = False):
    key="isActive"
    val="0"

    try:
        retCode,qList = __listQueues(dbcon, cur);
        if retCode != 0: return 502
        for queue_id in qList:
            retCode,queue_owner = __getQueueOwner(dbcon, cur, queue_id)
            __setValInSubscriberKeyValTable(dbcon, cur, queue_id, queue_owner, key, val)

        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 503


def __createSubscriberWindowStateTable(dbcon, cur, queue_id, db_user, doCommit = False):
    windowStateTableName = __getDBLevelUserWindowStateTableName(db_user, queue_id)
    query = ("CREATE TABLE IF NOT EXISTS `%s`(WindowNo INTEGER UNSIGNED , \
                                              SequenceNoInstance VARCHAR(767),\
                                              StartBlockNo BIGINT UNSIGNED ,       \
                                              EndBlockNo BIGINT UNSIGNED ,         \
                                              StartTime DateTime,                \
                                              EndTime DateTime,                 \
                                              NoOfMissingBlocks INTEGER UNSIGNED, \
                                              NoOfLateArrivals INTEGER UNSIGNED, \
                                              CurrentWindowState VARCHAR(20), \
                                              Retries TINYINT UNSIGNED NOT NULL DEFAULT 0, \
                                              RowEntryTime TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP, \
                                              WaitStartTime TIMESTAMP, \
                                              ProcessingStartTime TIMESTAMP, \
                                              WaitBudget INTEGER UNSIGNED, \
                                              RemainingWaitBudget INTEGER UNSIGNED, \
                                              Log VARCHAR(255),\
                                              AppLevelInfo VARCHAR(255),\
                                              PRIMARY KEY (WindowNo,SequenceNoInstance),\
                                              FOREIGN KEY (CurrentWindowState) REFERENCES WindowState(StateName)) \
                                              ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;"
            ) %(windowStateTableName)

    # Note: I wanted to add: FOREIGN KEY (WindowNo) REFERENCES '%s'(WindowNo),  but WindowNo is not unique in the
    # queue table and hence I can't do it.

    try:
        _debug(query)
        cur.execute(query)
        if doCommit: dbcon.commit()
        return 0
    except MySQLdb.Error as e:
        _debug(e)
        return 504


def __createSubscriberMissingBlockTable(dbcon, cur, queue_id, db_user, doCommit=False):
        missingBlockTableName = __getDBLevelUserMissingBlockTable(db_user, queue_id)
        '''
        query = ("CREATE TABLE IF NOT EXISTS `%s`(SequenceNo BIGINT UNSIGNED NOT NULL UNIQUE, \
                                                  RowInsertionTime TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP, \
                                                  PRIMARY KEY (SequenceNo)) \
                                                  ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;" \
                 ) %(missingBlockTableName)
        '''
        query = ("CREATE TABLE IF NOT EXISTS `%s`(DataURI VARCHAR(767) NOT NULL, \
                                                  RowInsertionTime TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP, \
                                                  PRIMARY KEY (DataURI), \
                                                  FOREIGN KEY (DataURI) REFERENCES Data(DataURI)) \
                                                  ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;"
                 ) %(missingBlockTableName)

        try:
            _debug(query)
            cur.execute(query)
            if doCommit: dbcon.commit()
            return 0
        except MySQLdb.Error as e:
            _debug(e)
            return 505


def __createSubscriberSiteInfoTable(dbcon, cur, queue_id, db_user, doCommit=False):
        siteInfoTableName = __getDBLevelUserSiteInfoTable(db_user, queue_id)
        query = ("CREATE TABLE IF NOT EXISTS `%s`(SequenceNoInstance VARCHAR(767), \
                                                  StartSequenceNo BIGINT UNSIGNED, \
                                                  RowInsertionTime TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP, \
                                                  PRIMARY KEY (SequenceNoInstance)) \
                                                  ENGINE=InnoDB DEFAULT CHARSET=latin1 ROW_FORMAT=DYNAMIC;"
                 ) % (siteInfoTableName)

        try:
            _debug(query)
            cur.execute(query)
            if doCommit: dbcon.commit()
            return 0
        except MySQLdb.Error as e:
            _debug(e)
            return 505


def __getDBLevelUserWindowStateTableName(user_id, queue_id):
    windowStateTableName = "%s_%s_%s"  %(user_id, queue_id, "WindowState")
    return windowStateTableName


def __getDBLevelUserMissingBlockTable(user_id, queue_id):
    table_name = "%s_%s_%s"  %(user_id, queue_id, "MissingBlockRecord")
    return table_name


def __getDBLevelUserSiteInfoTable(user_id, queue_id):
    table_name = "%s_%s_%s"  %(user_id, queue_id, "SiteInfo")
    return table_name


def __detectDuplicateInSubscriberQueueTable(dbcon, cur, queue_id, queue_owner, sequence_no, sequence_no_instance, newDataURI):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "SELECT DataURI FROM `%s` WHERE SequenceNo=%d AND SequenceNoInstance='%s';" %(queueTableName,
                                                                                          int(sequence_no),
                                                                                          sequence_no_instance)

    try:
        _debug(query)
        cur.execute(query)
        dataURI = cur.fetchone()

        if cur.rowcount != 0 and dataURI[0] != newDataURI:
            return 0, True
        else:
            return 0, False
    except MySQLdb.Error as e:
        _debug(e)
        return 505, False

# Suddenly we stopped using two underscores at the start of the function.  That is why because python appends stuff
# before function calls, when it is called from a class method.
# ToDo: Probably we should get rid of __ from all methods in this module.
def mark_blocks(dbcon, cur, queue_id,  date, is_before, new_state):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    operator = " < "
    if not is_before:
        operator = " > "
    query = ("UPDATE `%s` SET FileStateName='%s' \
             WHERE EventTime %s STR_TO_DATE('%s', '%s');"
             ) % (table_name, new_state, operator, date, "%Y %m %d %H %i UTC")

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def mark_blocks_as_inwindow_deprecated(dbcon, cur, queue_id, window_size_in_hours, windowing_start_date):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("UPDATE `%s` SET FileStateName = 'INWINDOW' \
             WHERE DataURI IN (SELECT DataURI \
                               FROM (SELECT DISTINCT CEILING(TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), q.EventTime)/(%d * 60 * 60)) \
                                     AS windowNo, DataURI, FileStateName FROM (`%s` AS q)) AS windows  \
                               WHERE windows.windowNo > 0    \
                               AND FileStateName = 'UNRESERVED' \
                               AND windows.windowNo IN \
                               (SELECT DISTINCT WindowNo FROM `%s` WHERE CurrentWindowState = 'UNDERFORMATION'));"
             ) % (queue_table_name, windowing_start_date, "%Y %m %d %H %i UTC", window_size_in_hours, queue_table_name,
                  window_state_table_name)

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def mark_blocks_as_inwindow(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("UPDATE `%s` q INNER JOIN `%s` w ON \
             (w.WindowNO = q.WindowNo AND w.SequenceNoInstance = q.SequenceNoInstance)  \
             SET FileStateName = 'INWINDOW' \
             WHERE w.CurrentWindowState = 'UNDERFORMATION' AND q.FileStateName = 'UNRESERVED';"
             ) % (queue_table_name, window_state_table_name)

    _debug(query)
    '''
    # for testing
    UPDATE `hlander_rsscint_1_Queue` q INNER JOIN `hlander_rsscint_1_WindowState` w ON
    (w.WindowNO = q.WindowNo AND w.SequenceNoInstance = q.SequenceNoInstance)
    SET FileStateName = 'INWINDOW' WHERE w.CurrentWindowState = 'UNDERFORMATION' AND q.FileStateName = 'UNRESERVED';
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def get_new_window_numbers_deprecated(dbcon, cur, queue_id,  window_size_in_hours, windowing_start_date):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("SELECT windowNo \
              FROM (SELECT DISTINCT CEILING(TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), \
                                          q.EventTime)/(%d * 60 * 60)) AS windowNo, FileStateName  \
                    FROM (`%s` AS q)) AS windows  \
             WHERE windows.windowNo > 0  \
             AND FileStateName = 'UNRESERVED' \
             AND windows.windowNo NOT IN (SELECT DISTINCT WindowNo FROM `%s`);"
             ) % (windowing_start_date, "%Y %m %d %H %i UTC", window_size_in_hours, queue_table_name, window_state_table_name)

    _debug(query)

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = [w[0] for w in windows]
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def get_new_window_numbers(dbcon, cur, queue_id, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    if ignore_site:
        query = ("SELECT DISTINCT WindowNo, NULL as  SequenceNoInstance FROM `%s` q WHERE \
                  q.FileStateName = 'UNRESERVED' AND \
                  q.WindowNo IS NOT NULL AND NOT EXISTS \
                  (SELECT DISTINCT WindowNo From `%s` w WHERE \
                  q.WindowNo = w.WindowNo);"
                ) % (queue_table_name, window_state_table_name)
    else:
        query = ("SELECT DISTINCT WindowNo, SequenceNoInstance FROM `%s` q WHERE \
                  q.FileStateName = 'UNRESERVED' AND \
                  q.WindowNo IS NOT NULL AND NOT EXISTS \
                  (SELECT DISTINCT WindowNo,SequenceNoInstance From `%s` w WHERE \
                  q.WindowNo = w.WindowNo AND q.SequenceNoInstance = w.SequenceNoInstance);"
                ) % (queue_table_name, window_state_table_name)

    '''
    # for testing
    SELECT DISTINCT WindowNo, SequenceNoInstance FROM `hlander_rsscint_1_Queue` q WHERE 
    q.FileStateName = 'UNRESERVED' AND 
    q.WindowNo IS NOT NULL AND NOT EXISTS 
    (SELECT DISTINCT WindowNo,SequenceNoInstance From `hlander_rsscint_1_WindowState` w 
    WHERE q.WindowNo = w.WindowNo AND 
    q.SequenceNoInstance = w.SequenceNoInstance);

    '''

    _debug(query)

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = [(w[0],w[1]) for w in windows]
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def assign_window_no_to_blocks(dbcon, cur, queue_id, window):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    site_info_table_name = __getDBLevelUserSiteInfoTable(queue_owner, queue_id)

    if window.sizeMetric == "Blocks":
        query = ("UPDATE `%s` q   INNER JOIN `%s` s ON q.SequenceNoInstance = s.SequenceNoInstance \
                 SET q.WindowNo = ((q.SequenceNo - s.StartSequenceNo + 1)/%d)   WHERE q.WindowNo IS NULL;"
                 ) % (queue_table_name, site_info_table_name, int(window.size))
    else:
        query = ("UPDATE `%s` SET WindowNo = \
                  CEILING(TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), EventTime)/(%d * 60 * 60)) \
                  WHERE WindowNo IS NULL;"
                 ) % (queue_table_name, window.startTime, "%Y %m %d %H %i UTC", int(window.size))

    _debug(query)
    '''
    # For testing:
    # for time based: UPDATE `hlander_rsscint_1_Queue` SET WindowNo = EventTime/60 WHERE WindowNo IS NULL;
    # HEADS UP: Following inner join syntax not portable.
    # for block based:
    # UPDATE `hlander_rsscint_1_Queue` q   INNER JOIN `hlander_rsscint_1_SiteInfo` s ON
      q.SequenceNoInstance = s.SequenceNoInstance SET q.WindowNo = (q.SequenceNo - s.StartSequenceNo)
      WHERE q.DataURI IS NULL;
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 202, str(e)


def insert_new_window_time_based(dbcon, cur, queue_id,  window_no, window_start_date, window_end_date, date_format
                                 , site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 509, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # ToDo: Should we use Insert ignore to allow multiple inserts of same record?  Probably not because we take care to
    # not add same entry multiple times
    if ignore_site:
        query = ("INSERT INTO `%s` (WindowNo, StartTime, EndTime, CurrentWindowState)  \
                  VALUES (%d, STR_TO_DATE('%s','%s'), \
                          STR_TO_DATE('%s','%s'), '%s');"
                 ) % (window_state_table_name, window_no, window_start_date, date_format,
                      window_end_date, date_format, "UNDERFORMATION")
    else:
        query = ("INSERT INTO `%s` (WindowNo, StartTime, EndTime, CurrentWindowState, SequenceNoInstance)  \
                  VALUES (%d, STR_TO_DATE('%s','%s'), \
                          STR_TO_DATE('%s','%s'), '%s', '%s');"
                 ) % (window_state_table_name, window_no, window_start_date, date_format,
                      window_end_date, date_format, "UNDERFORMATION", site)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 510, print(e)


def insert_new_window_block_based(dbcon, cur, queue_id,  window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 509, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # ToDo: Should we use Insert ignore to allow multiple inserts of same record?  Probably not because we take care to
    # not add same entry multiple times
    if ignore_site:
        query = ("INSERT INTO `%s` (WindowNo, CurrentWindowState)  \
                  VALUES (%d, '%s');"
                 ) % (window_state_table_name, window_no,  "UNDERFORMATION")
    else:
        query = ("INSERT INTO `%s` (WindowNo, CurrentWindowState, SequenceNoInstance)  \
                  VALUES (%d, '%s');"
                 ) % (window_state_table_name, window_no,  "UNDERFORMATION", site)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 510, print(e)


def for_time_windows_get_windows_whose_time_has_come(dbcon, cur, queue_id):
    complete_windows = []
    missing_data_windows = []
    inconclusive_windows = []
    needs_further_investigation_windows = []
    r, c, msg_or_list_of_result = for_time_windows_get_window_info(dbcon, cur, queue_id)
    if not r:
        return r, c, msg_or_list_of_result, complete_windows, missing_data_windows
    for winNo, site, s, e, elapsed_hours in msg_or_list_of_result:
        r1, c1, required_count = get_required_block_count_for_window_completion(dbcon, cur, queue_id, winNo, site)
        r2, c2, current_count = get_block_count_in_window(dbcon, cur, queue_id, winNo, site)
        if (r1 is False) or (r2 is False):
            continue
        if required_count == current_count:
            complete_windows.append((winNo, site, s, e, elapsed_hours))
        else:
            needs_further_investigation_windows.append((winNo, site, s, e, elapsed_hours))

    for (winNo, site, s, e, elapsed_hours) in needs_further_investigation_windows:
        r, c, msg_or_count = does_next_window_exists(dbcon, cur, queue_id, winNo, site)
        if r and msg_or_count > 0:
            missing_data_windows.append((winNo, site, s, e, elapsed_hours))
        else:
            inconclusive_windows.append((winNo, site, s, e, elapsed_hours))

    return True, 0, complete_windows, missing_data_windows, inconclusive_windows


def for_time_windows_get_window_info(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("SELECT WindowNo, SequenceNoInstance, StartTime, EndTime , \
             CEILING((TIMESTAMPDIFF(SECOND, EndTime, utc_timestamp())/(60 * 60))) as \
             elapsed_time_in_hours FROM `%s` WHERE CurrentWindowState = 'UNDERFORMATION' AND  \
             EndTime < utc_timestamp();"
             ) % window_state_table_name

    _debug(query)

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()  # fetchall() returns list of tuples. tuples are in ().  Array is in []

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = windows
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def for_time_windows_get_windows_whose_time_has_come_deprecated(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("SELECT WindowNo, SequenceNoInstance, StartTime, EndTime , \
             CEILING((TIMESTAMPDIFF(SECOND, EndTime, utc_timestamp())/(60 * 60))) as \
             elapsed_time_in_hours FROM `%s` WHERE CurrentWindowState = 'UNDERFORMATION' AND  \
             EndTime < utc_timestamp();"
             ) % window_state_table_name

    _debug(query)

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()  # fetchall() returns list of tuples. tuples are in ().  Array is in []

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = windows
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def for_block_windows_get_windows_whose_time_has_come(dbcon, cur, queue_id, window):
    _debug(inspect.currentframe().f_code.co_name)
    # fet 5 tuple list
    r, c, msg_or_list = get_window_info(dbcon, cur, queue_id)
    if not r:
        return r, c, msg_or_list, None, None

    complete_windows = []
    missing_data_windows = []
    inconclusive_windows = [] # we cant tell about them yet. Not sufficient information
    needs_further_investigation_windows = []

    for (window_no, site, min_block, max_block, count) in msg_or_list:
        if count == window.size:
            complete_windows.append((window_no, site, min_block, max_block, count))
        else:
            needs_further_investigation_windows.append((window_no, site, min_block, max_block, count))

    for (window_no, site, min_block, max_block, count) in needs_further_investigation_windows:
        r, c, msg_or_count = does_next_window_exists(dbcon, cur, queue_id, window_no, site)
        if r and msg_or_count > 0:
            missing_data_windows.append((window_no, site, min_block, max_block, count))
        else:
            inconclusive_windows.append((window_no, site, min_block, max_block, count))

    return True, 0, complete_windows, missing_data_windows, inconclusive_windows


def does_next_window_exists(dbcon, cur, queue_id, window_no, site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # ToDo: Fix: When first ever block of the window has EventTime == start_time, this block will never be processed!
    query = ("SELECT count(*) FROM (SELECT WindowNo, SequenceNoInstance FROM \
             `%s` Where WindowNo > %d AND SequenceNoInstance = '%s') as d;"
             ) % (window_state_table_name, int(window_no), site)

    _debug(query)
    ''' for testing only
    SELECT count(*) FROM (SELECT WindowNo, SequenceNoInstance FROM 
    `hlander_rsscint_1_WindowState` Where WindowNo > 2 AND SequenceNoInstance = 'lax') as d;
    '''

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_window_info(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("SELECT r.a, r.b, r.c, r.d, r.e as count FROM \
             (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c, max(SequenceNo) as d, \
             count(*) as e  FROM `%s` q, `%s` w WHERE w.CurrentWindowState = 'UNDERFORMATION' AND \
             q.WindowNo = w.WindowNo AND \
             q.SequenceNoInstance = w.SequenceNoInstance) as r WHERE r.a IS NOT NULL;"
             ) % queue_table_name,window_state_table_name

    _debug(query)
    ''' Query for testing purposes:
    SELECT r.a, r.b, r.c, r.d, r.e as count FROM 
    (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c, 
    max(SequenceNo) as d, count(*) as e FROM `hlander_rsscint_1_Queue` q, `hlander_rsscint_1_WindowState` w 
    WHERE w.CurrentWindowState = 'UNDERFORMATION' AND 
    q.WindowNo = w.WindowNo AND q.SequenceNoInstance = w.SequenceNoInstance) as r WHERE r.a IS NOT NULL;
    '''

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()  # fetchall() returns list of tuples. tuples are in ().  Array is in []

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = windows
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def get_a_specific_window_info(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    if ignore_site:
        query = ("SELECT r.a, r.b, r.c, r.d, r.e as count FROM \
                 (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c,  \
                  max(SequenceNo) as d, count(*) as e \
                  FROM `%s` q      WHERE q.WindowNo = %d ) \
                  as r WHERE r.a IS NOT NULL;"
                 ) % queue_table_name, window_no
    else:
        query = ("SELECT r.a, r.b, r.c, r.d, r.e as count FROM \
                 (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c,  \
                  max(SequenceNo) as d, count(*) as e \
                  FROM `%s` q      WHERE q.WindowNo = %d AND q.SequenceNoInstance = '%s') \
                  as r WHERE r.a IS NOT NULL;"
                 ) % queue_table_name, window_no, site

    _debug(query)
    ''' Query for testing purposes:
    SELECT r.a, r.b, r.c, r.d, r.e as count FROM 
    (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c, 
    max(SequenceNo) as d, count(*) as e FROM `hlander_rsscint_1_Queue` q 
    WHERE q.WindowNo = 10 AND q.SequenceNoInstance = 'lax') as r WHERE r.a IS NOT NULL;
    '''

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()  # fetchall() returns list of tuples. tuples are in ().  Array is in []

        if windows is None:
            return False, 507, "windows is None"
        elif cur.rowcount == 0:
            return True, 0, windows_array
        else:
            windows_array = windows
            return True, cur.rowcount, windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def get_required_block_count_for_window_completion(dbcon, cur, queue_id, window_no, site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    # ToDo: If last blcok of the window is missing, following will give wrong result.  Need to improve using the
    # first block of next window
    # ToDo: What if we need to ignore site?  How that will affect results of following query?
    query = ("SELECT (r.d-r.c+1) as count FROM \
             (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c, max(SequenceNo) as d \
             FROM `%s` q  WHERE q.WindowNo = %d AND q.SequenceNoInstance = '%s') as r WHERE r.a IS NOT NULL;"
             ) % (queue_table_name, int(window_no), site)

    _debug(query)
    ''' Query for testing purposes:
    SELECT r.a, r.b, r.c, r.d, (r.d-r.c+1) as count FROM 
    (SELECT q.WindowNo as a, q.SequenceNoInstance as b, min(SequenceNo) as c, 
    max(SequenceNo) as d FROM `hlander_rsscint_1_Queue` q, `hlander_rsscint_1_WindowState` w 
    WHERE w.CurrentWindowState = 'UNDERFORMATION' AND 
    q.WindowNo = w.WindowNo AND q.SequenceNoInstance = w.SequenceNoInstance) as r WHERE r.a IS NOT NULL;
    '''

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_required_block_count_for_window_completion_deprecated(dbcon, cur, queue_id, start_time, end_time):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("SELECT SUM(t.block_count) from (SELECT MAX(SequenceNo) - MIN(SequenceNo) + 1 as block_count FROM \
             `%s` WHERE FileStateName = 'INWINDOW' AND EventTime > '%s' AND EventTime <= '%s'  \
             GROUP BY SequenceNoInstance) as t;"
             ) % (queue_table_name, start_time, end_time)

    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_block_count_in_window(dbcon, cur, queue_id, window_no, site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    # ToDo: Fix: When first ever block of the window has EventTime == start_time, this block will never be processed!
    # ToDo: You assumed user asked for site based filtering as well in the following. What if he didn't?
    query = ("SELECT count(*) FROM `%s` WHERE WindowNo = %s AND SequenceNoInstance = '%s' ;"
             ) % (queue_table_name, window_no, site)

    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_block_count_in_window_deprecated(dbcon, cur, queue_id, start_time, end_time):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    # ToDo: Fix: When first ever block of the window has EventTime == start_time, this block will never be processed!
    query = ("SELECT count(*) FROM `%s` WHERE FileStateName = 'INWINDOW' AND EventTime > '%s' AND EventTime <= '%s' ;"
             ) % (queue_table_name, start_time, end_time)

    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_elapsed_time_in_hours(dbcon, cur, end_time):
    _debug(inspect.currentframe().f_code.co_name)
    query = ("SELECT  CEILING((TIMESTAMPDIFF(SECOND, '%s', utc_timestamp())/(60 * 60)));"
             ) % end_time

    _debug(query)

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def get_blocks_in_window_deprecated(dbcon, cur, queue_id, start_time, end_time):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("SELECT DataURI FROM `%s` WHERE FileStateName = 'INWINDOW' AND EventTime > '%s' AND EventTime <= '%s' \
             ORDER BY SequenceNoInstance, SequenceNo;"
             ) % (queue_table_name, start_time, end_time)

    _debug(query)

    data_array = []

    try:
        cur.execute(query)
        rows = cur.fetchall()  # fetchall() returns a list

        if rows is None:
            return False, 507, "rows is None"
        elif cur.rowcount == 0:
            return True, 0, data_array
        else:
            data_array = [w[0] for w in rows]
            return True, cur.rowcount, data_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def get_blocks_in_window(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    if ignore_site:
        query = ("SELECT DataURI FROM `%s` WHERE WindowNo = %d ORDER BY DataURI;"
                 ) % (queue_table_name, window_no)
    else:
        query = ("SELECT DataURI FROM `%s` WHERE WindowNo = %d AND SequenceNoInstance = '%s' ORDER BY DataURI;"
                 ) % (queue_table_name, window_no, site)

    _debug(query)
    ''' for testing
    SELECT DataURI FROM `hlander_rsscint_1_Queue` WHERE WindowNo = 10 AND SequenceNoInstance = 'lax' ORDER BY DataURI;
    '''

    data_array = []

    try:
        cur.execute(query)
        rows = cur.fetchall()  # fetchall() returns a list

        if rows is None:
            return False, 507, "rows is None"
        elif cur.rowcount == 0:
            return True, 0, data_array
        else:
            data_array = [w[0] for w in rows]
            return True, cur.rowcount, data_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def change_window_state(dbcon, cur, queue_id, window_no, site, old_state, new_state, app_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("UPDATE `%s` SET CurrentWindowState='%s' , AppLevelInfo='%s' WHERE WindowNo=%d AND SequenceNoInstance='%s'\
             AND CurrentWindowState='%s';"
             ) % (window_state_table_name, new_state, app_id, window_no, site, old_state)
    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 202, str(e)


def fill_site_info_table(dbcon, cur, queue_id, useQSeqNoInst, qSeqNoInstListAsString):
    _debug(inspect.currentframe().f_code.co_name)
    ret_code, count, data_array = get_new_sites(dbcon, cur, queue_id)
    if not ret_code:
        return False, 510, "get_new_sites failed for queue id %s. This failure is in %s" % (queue_id, __name__)
    q_seq_no_list = qSeqNoInstListAsString.split(",")
    if useQSeqNoInst.lower() == "yes":
        new_sites = list(set(data_array) & set(q_seq_no_list))
        # Sites user didn't ask.  Mark those blocks as unwanted
        ignore_site_list = list(set(data_array) - set(q_seq_no_list))
        # We used to mark unwanted blocks.  But now we delete them right away
        # mark_unwanted_blocks(dbcon, cur, queue_id, ignore_site_list)
        del_unwanted_blocks(dbcon, cur, queue_id, ignore_site_list)
    else: new_sites = data_array
    if len(new_sites) == 0: return True, 0, "OK"

    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 509, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    site_info_table_name = __getDBLevelUserSiteInfoTable(queue_owner, queue_id)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    for site in new_sites:
        query = ("INSERT INTO `%s` (SequenceNoInstance, StartSequenceNo) \
                  VALUES('%s' , (SELECT min(SequenceNo) FROM `%s` WHERE SequenceNoInstance = '%s' AND \
                  FileStateName = 'UNRESERVED'));"
                 ) % (site_info_table_name, site, queue_table_name, site)
        '''
        # Example query:
        INSERT INTO `hlander_rsscint_1_SiteInfo` (SequenceNoInstance, StartSequenceNo)
        VALUES('lax' , (SELECT min(SequenceNo) FROM `hlander_rsscint_1_Queue` WHERE SequenceNoInstance = 'lax' AND
        FileStateName = 'UNRESERVED'));
        '''
        _debug(query)

        try:
            cur.execute(query)
            dbcon.commit()
        except MySQLdb.Error as e:
            _debug(e)
            return False, 510, print(e)

    return True, 0, "OK"


def get_new_sites(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    site_info_table_name = __getDBLevelUserSiteInfoTable(queue_owner, queue_id)

    query = ("SELECT DISTINCT SequenceNoInstance  FROM `%s` WHERE FileStateName = 'UNRESERVED' AND \
              SequenceNoInstance NOT IN (SELECT SequenceNoInstance FROM `%s`);"
             ) % (queue_table_name, site_info_table_name)
    '''
    # Example query for testing in sql client
    SELECT DISTINCT SequenceNoInstance  FROM `hlander_rsscint_1_Queue` WHERE FileStateName = 'UNRESERVED' AND \
    SequenceNoInstance NOT IN (SELECT SequenceNoInstance FROM `hlander_rsscint_1_SiteInfo`);
    '''

    _debug(query)

    data_array = []

    try:
        cur.execute(query)
        rows = cur.fetchall()  # fetchall() returns a list

        if rows is None:
            return False, 507, "rows is None"
        elif cur.rowcount == 0:
            return True, 0, data_array
        else:
            data_array = [w[0] for w in rows]
            return True, cur.rowcount, data_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e)


def mark_unwanted_blocks(dbcon, cur, queue_id, ignore_site_list):
    _debug(inspect.currentframe().f_code.co_name)
    if len(ignore_site_list) == 0:
        return True, 0, "OK"
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("UPDATE `%s` set FileStateName='UNWANTED' WHERE SequenceNoInstance in %s;"
             ) % (table_name, str(tuple(ignore_site_list)))
    '''
    # example query
    UPDATE `hlander_rsscint_1_Queue` set FileStateName='UNWANTED' WHERE SequenceNoInstance in ('lax','mia');

    '''

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def del_unwanted_blocks(dbcon, cur, queue_id, ignore_site_list):
    _debug(inspect.currentframe().f_code.co_name)
    if len(ignore_site_list) == 0:
        return True, 0, "OK"
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("DELETE FROM `%s` WHERE SequenceNoInstance in %s;"
             ) % (table_name, str(tuple(ignore_site_list)))
    '''
    # example query
    DELETE FROM `hlander_rsscint_1_Queue`  WHERE SequenceNoInstance in ('lax','mia');

    '''

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def mark_blocks_as_late(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    missing_blocks_table_name = __getDBLevelUserMissingBlockTable(queue_owner, queue_id)

    query = ("UPDATE `%s` SET FileStateName = 'LATE' \
             WHERE FileStateName = 'UNRESERVED' AND \
             DataURI IN (SELECT DataURI FROM `%s`);"
             ) % (queue_table_name, missing_blocks_table_name)

    '''
    # Query for testing:
    UPDATE `hlander_rsscint_1_Queue` SET FileStateName = 'LATE'
    WHERE FileStateName = 'UNRESERVED' AND
    DataURI IN (SELECT DataURI FROM `hlander_rsscint_1_MissingBlockRecord`);
    '''

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def mark_blocks_as_duplicate(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (
        queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    query = ("UPDATE `%s` SET FileStateName = 'DUPLICATE' WHERE DataURI IN \
             (SELECT DataURI FROM (SELECT * FROM `%s`) AS tmp  WHERE FileStateName = 'UNRESERVED' AND \
             WindowNo IN (SELECT WindowNo FROM `%s` WHERE CurrentWindowState != 'UNDERFORMATION'));"
             ) % (queue_table_name, queue_table_name, window_state_table_name)

    '''
    # Query for testing:
     UPDATE `hlander_rsscint_1_Queue` SET FileStateName = 'DUPLICATE' WHERE DataURI IN 
     (SELECT DataURI FROM (SELECT * FROM `hlander_rsscint_1_Queue`) AS tmp  
     WHERE FileStateName = 'UNRESERVED' AND WindowNo IN 
     (SELECT WindowNo FROM `hlander_rsscint_1_WindowState` WHERE CurrentWindowState != 'UNDERFORMATION'));
    '''

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def delete_late_blocks_from_missing_table(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (
        queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    missing_blocks_table_name = __getDBLevelUserMissingBlockTable(queue_owner, queue_id)

    query = ("DELETE FROM `%s`  \
             WHERE DataURI IN \
             (SELECT DataURI from `%s` WHERE FileStateName = 'LATE');"
             ) % (missing_blocks_table_name, queue_table_name)

    '''
    # Query for testing:
    DELETE FROM `hlander_rsscint_1_MissingBlockRecord`
    WHERE DataURI IN (SELECT DataURI from `hlander_rsscint_1_Queue` WHERE FileStateName = 'LATE');
    '''

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def mark_blocks_as_duplicate_deprecated(dbcon, cur, queue_id, windowing_start_date, window_size, is_size_in_hours):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)
    missing_blocks_table_name = __getDBLevelUserMissingBlockTable(queue_owner, queue_id)
    site_info_table_name = __getDBLevelUserSiteInfoTable(queue_owner, queue_id)
    block_no_to_window_no_mapping_query = ""
    if is_size_in_hours:
        block_no_to_window_no_mapping_query = "SELECT DISTINCT CEILING( \
                    TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), q.EventTime)/(%d * 60 * 60))" \
                    % (windowing_start_date, "%Y %m %d %H %i UTC", window_size)
    else:
        block_no_to_window_no_mapping_query = "SELECT DISTINCT CEILING( \
                    TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), q.EventTime)/(%d * 60 * 60))" \
                    % (windowing_start_date, "%Y %m %d %H %i UTC", window_size)

    query = ("UPDATE `%s` SET FileStateName = 'INWINDOW' \
             WHERE DataURI IN (SELECT DataURI \
                               FROM (SELECT DISTINCT CEILING(TIMESTAMPDIFF(SECOND,STR_TO_DATE('%s','%s'), q.EventTime)/(%d * 60 * 60)) \
                                     AS windowNo, DataURI, FileStateName FROM (`%s` AS q)) AS windows  \
                               WHERE windows.windowNo > 0    \
                               AND FileStateName = 'UNRESERVED' \
                               AND windows.windowNo IN \
                               (SELECT DISTINCT WindowNo FROM `%s` WHERE CurrentWindowState = 'UNDERFORMATION'));"
             ) % (queue_table_name, windowing_start_date, "%Y %m %d %H %i UTC", window_size_in_hours, queue_table_name,
                  window_state_table_name)

    _debug(query)

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def delete_window_blocks(dbcon, cur, queue_id, event_time, operator):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("DELETE FROM `%s` \
             WHERE EventTime %s STR_TO_DATE('%s', '%s');"
             ) % (queue_table_name, operator, event_time, "%Y %m %d %H %i UTC")
    _debug(query)
    ''' for testing only:
    DELETE FROM `hlander_rsscint_1_Queue` WHERE EventTime > STR_TO_DATE('2020 10 07 13 14', "%Y %m %d %H %i UTC");
    '''

    try:
        cur.execute(query)
    except MySQLdb.Error as e:
        _debug(e)
        return False, 203, str(e)

    return True, 0, "Okay"


def init_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # Initially my test was WaitStartTime IS NOT NULL but mysql was not using Null as default value, rather a 0
    if ignore_site:
        query = ("UPDATE  `%s` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  \
                  WaitStartTime = 0 AND WindowNo = %d;"
                 ) % (window_state_table_name, window_no)
    else:
        query = ("UPDATE  `%s` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  \
                  WaitStartTime = 0 AND WindowNo = %d AND SequenceNoInstance = '%s';"
                 ) % (window_state_table_name, window_no, site)

    _debug(query)
    ''' for testing:
    UPDATE  `hlander_rsscint_2_WindowState` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  
    WaitStartTime = 0 
    WindowNo = 10 AND 
    SequenceNoInstance = 'lax';
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def update_wait_start_time(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # Initially my test was WaitStartTime IS NOT NULL but mysql was not using Null as default value, rather a 0
    if ignore_site:
        query = ("UPDATE  `%s` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  \
                  WindowNo = %d;"
                 ) % (window_state_table_name, window_no)
    else:
        query = ("UPDATE  `%s` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  \
                  WindowNo = %d AND SequenceNoInstance = '%s';"
                 ) % (window_state_table_name, window_no, site)

    _debug(query)
    ''' for testing:
    UPDATE  `hlander_rsscint_2_WindowState` SET WaitStartTime = CURRENT_TIMESTAMP WHERE  
    WindowNo = 10 AND 
    SequenceNoInstance = 'lax';
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def get_wait_time_in_hours(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)

    # Initially my test was WaitStartTime IS NOT NULL but mysql was not using Null as default value, rather a 0
    if ignore_site:
        query = ("SELECT timestampdiff(HOUR, WaitStartTime, current_timestamp) FROM `%s` WHERE \
                  WindowNo = %d  AND WaitStartTime <> 0;"
                 ) % (window_state_table_name, window_no)
    else:
        query = ("SELECT timestampdiff(HOUR, WaitStartTime, current_timestamp) FROM `%s` WHERE \
                  WindowNo = %d AND SequenceNoInstance = '%s' AND WaitStartTime <> 0;"
                 ) % (window_state_table_name, window_no, site)

    _debug(query)
    ''' for testing:
    SELECT timestampdiff(HOUR, WaitStartTime, current_timestamp) FROM `hlander_rsscint_1_WindowState` WHERE 
    WindowNo = 10 AND SequenceNoInstance = 'lax' AND WaitStartTime <> 0;
    '''

    result = True
    code = 0
    msg = "Okay"
    duration = 0
    try:
        cur.execute(query)
        count = cur.fetchone()
        if count is None or cur.rowcount == 0:
            result = False
        else:
            duration = count[0]

    except MySQLdb.Error as e:
        _debug(e)
        result = False
        code = 300
        msg = str(e)

    return result, code, msg, duration


def did_window_receive_new_block(dbcon, cur, queue_id, window_no, site, ignore_site):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)
    if ignore_site:
        query = ("SELECT count(*) FROM `%s` AS q  WHERE q.WindowNo = %d  AND \
                  q.RowEntryEpoch > (SELECT w.WaitStartTime FROM `%s` AS w WHERE w.WindowNo = %d );"
                 ) % (queue_table_name, window_no, window_state_table_name, window_no)
    else:
        query = ("SELECT count(*) FROM `%s` AS q  WHERE q.WindowNo = %d AND q.SequenceNoInstance = '%s' AND \
                  q.RowEntryEpoch > (SELECT w.WaitStartTime FROM `%s` AS w WHERE w.WindowNo = %d AND \
                  w.SequenceNoInstance = '%s');"
                 ) % (queue_table_name, window_no, site, window_state_table_name, window_no, site)
    _debug(query)

    ''' for testing:
    SELECT count(*) FROM `hlander_rsscint_1_Queue` AS q  WHERE q.WindowNo = 10 AND 
    q.SequenceNoInstance = 'lax' AND q.RowEntryEpoch > 
    (SELECT w.WaitStartTime FROM `hlander_rsscint_1_WindowState` AS w 
    WHERE w.WindowNo = 10 AND w.SequenceNoInstance = 'lax');
    '''

    result = True
    code = 0
    msg = "Okay"
    try:
        cur.execute(query)
        count = cur.fetchone()
        if count is  None or cur.rowcount == 0 or count[0] == 0:
            result = False

    except MySQLdb.Error as e:
        _debug(e)
        result = False
        code = 300
        msg = str(e)

    return result, code, msg


def get_min_window(dbcon, cur, queue_id):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    window_state_table_name = __getDBLevelUserWindowStateTableName(queue_owner, queue_id)
    query = ("SELECT WindowNo, SequenceNoInstance FROM `%s` WHERE CurrentWindowState = 'UNDERFORMATION' \
              ORDER BY WindowNo, SequenceNoInstance LIMIT 1;"
             ) % window_state_table_name

    _debug(query)
    ''' Following query didn't work.  The next to it worked.
    SELECT least(WindowNo, SequenceNoInstance) FROM `hlander_rsscint_1_WindowState` 
    WHERE CurrentWindowState = 'UNDERFORMATION';

    SELECT WindowNo, SequenceNoInstance FROM `hlander_rsscint_1_WindowState` WHERE 
    CurrentWindowState = 'UNDERFORMATION' ORDER BY WindowNo, SequenceNoInstance LIMIT 1;
    '''

    windows_array = []

    try:
        cur.execute(query)
        windows = cur.fetchall()  # fetchall() returns list of tuples. tuples are in ().  Array is in []

        if windows is None:
            return False, 507, "windows is None", windows_array
        elif cur.rowcount == 0:
            return True, 0, "OK", windows_array
        else:
            windows_array = windows
            return True, cur.rowcount, "OK", windows_array
    except MySQLdb.Error as e:
        _debug(e)
        return False, 508, print(e), windows_array


def update_window_no(dbcon, cur, queue_id, data_uri):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = ("UPDATE `%s` SET WindowNo = WindowNo+1 WHERE DataURI = '%s';"
             ) % (queue_table_name, data_uri)

    _debug(query)
    ''' for testing:
    UPDATE `hlander_rsscint_1_Queue` SET WindowNo = WindowNo+1 WHERE DataURI = "abc";
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def update_window_state(dbcon, cur, queue_id, data_uri):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = ("UPDATE `%s` SET FileStateName = 'INMULTIWINDOW'  WHERE DataURI = '%s';"
             ) % (queue_table_name, data_uri)

    _debug(query)
    ''' for testing:
    UPDATE `hlander_rsscint_1_Queue` SET FileStateName = 'INMULTIWINDOW'  WHERE DataURI = 'abc';
    '''

    try:
        cur.execute(query)
        dbcon.commit()
        return True, 0, "OK"
    except MySQLdb.Error as e:
        _debug(e)
        return False, 506, str(e)


def get_block_URI(dbcon, cur, queue_id, site_from_url, block_no_from_url):
    _debug(inspect.currentframe().f_code.co_name)
    # find queue_owner
    ret_code, queue_owner = __getQueueOwner(dbcon, cur, queue_id)
    if ret_code != 0:
        return False, 506, "Unable to find owner of the queue with id %s. This failure is in %s" % (queue_id, __name__)
    queue_table_name = __getDBLevelUserQueueTableName(queue_owner, queue_id)

    query = ("SELECT DataURI FROM `%s` WHERE SequenceNo = %d AND SequenceNoInstance = '%s';"
             ) % (queue_table_name, block_no_from_url, site_from_url)

    _debug(query)
    ''' For debugging:
    SELECT DataURI FROM `hlander_rssacint_1_Queue` WHERE SequenceNo = 500 AND SequenceNoInstance = 'lax';
    '''

    try:
        cur.execute(query)
        value = cur.fetchone()

        if value is None:
            return False, 508, "value is None"
        else:
            return True, 0, value[0]
    except MySQLdb.Error as e:
        _debug(e)
        return False, 509, print(e)


def peekNext(dbcon, cur, queue_id, queue_owner):
    queueTableName = __getDBLevelUserQueueTableName(queue_owner, queue_id)
    query = "SELECT DataURI, SequenceNoInstance, SequenceNo FROM `%s` WHERE FileStateName='UNRESERVED' \
             ORDER BY EventTime LIMIT 1;" %(queueTableName)

    ''' Test Query
    SELECT DataURI, SequenceNoInstance, SequenceNo FROM `hlander_pcap.sz_1_Queue` WHERE FileStateName='UNRESERVED' \
    ORDER BY EventTime LIMIT 1;
    '''
    _debug(query)
    dataURI = "Un-Initialized"
    site = "Un-Initialized"
    seqNo = 0
    try:
        cur.execute(query)
        (dataURI, site, seqNo) = cur.fetchone()
        if dataURI is None or site is None or seqNo is None:
            return (1, dataURI, site, seqNo)
        else:
            return (0, dataURI, site, seqNo)
    except MySQLdb.Error as e:
        _debug(e)
        return 201, dataURI, site, seqNo

def getPeekInfoFromKV(dbcon, cur, queue_id, queue_owner):
    pass


#### Some test cases ####
def __addDummyRowsInQueue(dbcon, cur, count, referredTable):
    try:
        for i in range(0, count):
            query = 'INSERT IGNORE INTO Data (DataURI, data_type) Values \
                    ("%s", "%s");' % (str(i), "raw");
            _debug(query)
            cur.execute(query)

            query = 'INSERT INTO `%s` (DataURI, FileStateName) Values \
                    ("%s", "%s" );' % (referredTable, str(i), "UNRESERVED")
            _debug(query)

            cur.execute(query)

    except MySQLdb.Error as e:
        _debug(e)
        return 1

    try:
        dbcon.commit()
        return 0

    except MySQLdb.Error as e:
        _debug(e)
        return 2

