#!/usr/bin/env python3


import subprocess
from subprocess import call
import time
import os
import sys
from top.conf import sysConfInfo


HDFS_CMD = sysConfInfo.HDFS_CMD if sysConfInfo.HDFS_CMD else 'hdfs'

_DEBUG_THIS = False
def _debug(s):
  if _DEBUG_THIS:
    sys.stderr.write("DEBUG: " + str(s) + "\n")


def makeNewHDFSDir(hdfsURI):
    cmd = "%s -mkdir -p %s" % (HDFS_CMD, hdfsURI)
    return call(cmd, shell=True)


def makeHDFSDirIfDoesNotExist(hdfsURI):
    cmd = "%s -mkdir -p %s" % (HDFS_CMD, hdfsURI)
    return call(cmd, shell=True)


def doesHDFSDirExists(hdfsURI):
    dirName = os.path.dirname(hdfsURI)
    cmd = "%s -test -d  %s" % (HDFS_CMD, dirName)
    return call(cmd, shell=True)


def setFileOrDirRights(hdfsURI, rights):
    cmd = "%s -chmod  %s  %s" % (HDFS_CMD, rights, hdfsURI)
    return call(cmd, shell=True)


def setFileOrDirGroup(hdfsURI, group):
    cmd = "%s -chgrp  %s  %s" % (HDFS_CMD, group, hdfsURI)
    return call(cmd, shell=True)


def doesHDFSFileExists(hdfsURI):
    cmd = "%s -test -e  %s" % (HDFS_CMD, hdfsURI)
    return call(cmd, shell=True)


def copyFileFromLocal(localURI, hdfsURI, shouldOverWrite):
    if shouldOverWrite == True:
        cmd = "%s -copyFromLocal -f  %s  %s" % (HDFS_CMD, localURI, hdfsURI)
    else:
        cmd = "%s -copyFromLocal     %s  %s" % (HDFS_CMD, localURI, hdfsURI)
    _debug(cmd)
    return call(cmd, shell=True)


def copyToLocal(localURI, hdfsURI):
    cmd = "%s -copyToLocal  %s  %s" % (HDFS_CMD, hdfsURI, localURI)
    _debug(cmd)
    return call(cmd, shell=True)


def delHDFSFile(hdfsURI):
    cmd = "%s -rm  %s" % (HDFS_CMD, hdfsURI)
    _debug(cmd)
    return call(cmd, shell=True)


def mvHDFSFile(srcHdfsURI, dstHdfsURI):
    cmd = "%s -mv %s  %s" % (HDFS_CMD, srcHdfsURI, dstHdfsURI)
    _debug(cmd)
    return call(cmd, shell=True)


def getHDFSFileModTime(hdfsURI):
    stat = " -stat \%Y "
    cmd = "%s %s %s" % (HDFS_CMD, stat, hdfsURI)
    _debug(cmd)

    ret = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    if ret == None:
        return ret
    
    modTime = ret.stdout.readline().strip()
    
    retCode = ret.returncode
    ret.communicate()
    retCode = ret.returncode
    #print retCode
    #print modTime
    return retCode, modTime


def getHDFSFileSizeInBytes(hdfsURI):
    cmd = "%s -du  %s | awk '{ print $1 }' " % (HDFS_CMD, hdfsURI)
    _debug(cmd)

    ret = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    if ret == None:
        return ret

    noOfBytes = ret.stdout.readline().strip()

    retCode = ret.returncode
    ret.communicate()
    retCode = ret.returncode
    #print retCode
    #print noOfBytes
    return retCode, noOfBytes

